/**
 * perseus/connector: this is an implementation of some JCA-related technologies
 * (resource adapters and managers) for the ObjectWeb consortium.
 * Copyright (C) 2001-2002 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 */

package org.objectweb.speedo.jca;

import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.speedo.api.Debug;
import org.objectweb.perseus.pool.api.PoolResource;

import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Collection;
import javax.resource.ResourceException;
import javax.resource.spi.ConnectionEvent;
import javax.resource.spi.ConnectionEventListener;
import javax.resource.spi.ConnectionRequestInfo;
import javax.resource.spi.LocalTransaction;
import javax.resource.spi.ManagedConnection;
import javax.resource.spi.ManagedConnectionMetaData;
import javax.security.auth.Subject;
import javax.transaction.xa.XAResource;
import javax.transaction.xa.XAException;
import javax.jdo.PersistenceManager;
import javax.jdo.Transaction;
import javax.jdo.Query;
import javax.jdo.Extent;
import javax.jdo.PersistenceManagerFactory;
import javax.jdo.JDOException;

/**
 * @author	P. Dechamboux
 */
public class JdoManagedConnection
		implements ManagedConnection, LocalTransaction,
		javax.resource.cci.LocalTransaction, PoolResource, PersistenceManager,
        ManagedConnectionMetaData {

	public final static String EIS_PRODUCT_NAME = "Generic Resource Adapter for JDO Drivers";
	public final static String EIS_PRODUCT_VERSION = "1.0";
	public final static String USER_NAME = "No user name needed to use a JDO driver";

	/**
	 * The logger into which traces about JdoManagedConnection are produced.
	 */
	private Logger logger;
	/**
	 * The ManagedConnectionFactory that have requested the allocation of this
	 * ManagedConnection.
	 */
	private JdoManagedConnectionFactory managedConnectionFactory;
	/**
	 * The Connections associated with this ManagedConnection. Always at least
	 * one connection.
	 */
	private ArrayList associatedConnections = new ArrayList(1);
	/**
	 * The ConnectionEventListener list associated with this ManagedConnection.
	 */
	private ArrayList associatedListeners;
	/**
	 * It is set if the ManagedConnection execute within a LocalTransaction.
	 */
	private PersistenceManager localTransactionTxContext = null;
	/**
	 * It is set if the ManagedConnection has been requested for a XAResource.
	 */
	private JdoXAResource xaResource = null;
	/**
	 * It is set if JDO operation occurs outside any explicit (LocalTransaction)
	 * or implicit (XAResource) transaction context.
	 */
	private PersistenceManager defaultTxContext = null;

	/**
	 *
	 * @return
	 */
	protected PersistenceManager getPersistenceManager() {
		if (localTransactionTxContext != null) {
			return localTransactionTxContext;
		}
		if (xaResource != null && xaResource.getTxContext() != null) {
			/**
			 * The Managed connection is associated to an XAResource which has
			 * been started
			 */
			if (!xaResource.synchronizationRegistered) {
				//Register the PersistenceManager on the first use
				try {
					xaResource.registerSynchronization();
				} catch (XAException ex) {
					logger.log(BasicLevel.ERROR,
							"registerSynchronization throws an exception", ex);
				}
			}
			return xaResource.getTxContext();
		}
		//
		if (defaultTxContext == null) {
			defaultTxContext = managedConnectionFactory.getTxContextFactory().getPersistenceManager();
		}
		return defaultTxContext;
	}

	/**
	 * Constructs a JdoManagedConnection.
	 * @param el	The logger into which to produce ManagedConnection-related
	 * 				traces.
	 * @param fmcf	The ManagedConnectionFactory that has requested the
	 * 				ManagedConnection creation.
	 */
	JdoManagedConnection(Logger el, JdoManagedConnectionFactory fmcf) {
		logger = el;
		if (Debug.ON)
			logger.log(BasicLevel.DEBUG,
					   "Constructs a new JdoManagedConnection");
		managedConnectionFactory = fmcf;
		associatedListeners = null;
	}

	/**
	 * Specifies if this ManagedConnection still has an active LocalTransaction.
	 * @return  true if a LocalTransaction is still active.
	 */
	boolean localTransactionTerminated() {
		return localTransactionTxContext == null;
	}

	/**
	 * Dissociates a Connection from the ones that are associated to this
	 * ManagedConnection.
	 * @param conn	The Connection to be dissociated.
	 */
	synchronized void dissociateConnection(Object conn) throws ResourceException {
		int ind = associatedConnections.indexOf(conn);
        if (ind == -1) {
            throw new ResourceException("JDO Connector: cannot dissociate an unknown Connection:"
                    + conn + "\n among " + associatedConnections);
        }
        if (Debug.ON && logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, "Dissociates Connection (" + ind + "): " + conn);
        }
		associatedConnections.remove(ind);
		ConnectionEvent ce = new ConnectionEvent(this, ConnectionEvent.CONNECTION_CLOSED);
        if (associatedListeners != null) {
		    for (Iterator it = associatedListeners.iterator(); it.hasNext();)
			    ((ConnectionEventListener) it.next()).connectionClosed(ce);
        }
	}

	
	// IMPLEMENTATION OF METHODS FROM THE (cci)ManagedConnectionMetaData INTERFACE

    public String getEISProductName() throws ResourceException {
        return EIS_PRODUCT_NAME;
    }

    public String getEISProductVersion() throws ResourceException {
        return EIS_PRODUCT_VERSION;
    }

    public int getMaxConnections() throws ResourceException {
		if (defaultTxContext != null) {
			return defaultTxContext.currentTransaction().getPersistenceManager()
					.getPersistenceManagerFactory().getMaxPool();
		} else {
			// TODO: analysis of ra.xml values
			return 10;
		}
    }

    public String getUserName() throws ResourceException {
        return USER_NAME;
    }

	// IMPLEMENTATION OF METHODS FROM THE (cci)ManagedConnection INTERFACE

	/**
	 * Delegates the creation of a Connection to the ConnectionFactory.
	 * "subject" and "info" parameters are ignored.
	 */
	public Object getConnection(Subject subject, ConnectionRequestInfo info)
			throws ResourceException {
		JdoConnection rac = ((JdoConnectionFactory) managedConnectionFactory.getJcaConnectionFactory()).createConnection();
		rac.initialize(this);
        associateConnection(rac);
		return rac;
	}

	/**
	 * Cleans up the connection.
	 */
	public void cleanup() throws ResourceException {
		if (Debug.ON)
			logger.log(BasicLevel.DEBUG, "Cleans up ManagedConnection");
		if (associatedConnections.size() != 0) {
			throw new ResourceException("Cannot cleanup a ManagedConnection that still has associated connections");
		}
		if (localTransactionTxContext != null) {
			localTransactionTxContext.close();
			localTransactionTxContext = null;
		}
		if (defaultTxContext != null) {
			defaultTxContext.close();
			defaultTxContext = null;
		}
		xaResource = null;
	}

	/**
	 * Called when ManagedConnection is removed.
	 */
	public void destroy() throws ResourceException {
		logger = null;
		managedConnectionFactory = null;
		associatedConnections = null;
		associatedListeners = null;
		localTransactionTxContext = null;
		xaResource = null;
	}

	/**
	 * Associates a new Connection to this ManagedConnection. Nothing is done
	 * if it has already been associated (enforces no duplicate).
	 */
	public synchronized void associateConnection(Object o) throws ResourceException {
		if (Debug.ON)
			logger.log(BasicLevel.DEBUG, "Associates Connection: " + o + " to the MC " + this);
		if (!associatedConnections.contains(o))
			associatedConnections.add(o);
		else
			logger.log(BasicLevel.WARN, "Connection: " + o
										+ "has already been associated: ignored!");
	}

	/**
	 * Adds a listener to the associatedListeners list if it has not
	 * already been done. This means that duplicates are ignored.
	 * It first creates this list if needed (creation at first add).
	 */
	public synchronized void addConnectionEventListener(
			ConnectionEventListener listener) {
		if (associatedListeners == null)
			associatedListeners = new ArrayList();
		if (Debug.ON)
			logger.log(BasicLevel.DEBUG, "Associates ConnectionEventListener: "
										 + listener);
		if (!associatedListeners.contains(listener))
			associatedListeners.add(listener);
		else
			logger.log(BasicLevel.WARN, "ConnectionEventListener: " + listener
										+ "has already been associated: ignored!");
	}

	/**
	 * Removes a listener from the associatedListeners list. If the list does
	 * not eist (null), or if this listener does not belong to this list,
	 * nothing is done.
	 */
	public synchronized void removeConnectionEventListener(
			ConnectionEventListener listener) {
		if (associatedListeners == null) {
			logger.log(BasicLevel.WARN,
					   "No associated ConnectionEventListener: ignored!");
			return;
		}
		int ind = associatedListeners.indexOf(listener);
		if (Debug.ON)
			logger.log(BasicLevel.DEBUG,
					   "Index within the associatedListeners: " + ind);
		if (ind == -1)
			logger.log(BasicLevel.WARN, "ConnectionEventListener" + listener
										+ "not associated here: ignored!");
		else
			associatedListeners.remove(ind);
	}

	/**
	 * Retrieves an XA resource. Switches running mode to XA.
	 */
	public synchronized XAResource getXAResource() throws ResourceException {
        logger.log(BasicLevel.DEBUG, "getXAResource()");
		if (localTransactionTxContext != null) {
			throw new ResourceException(
					"Try switching to XA mode while running a LocalTransaction.");
		}
		if (defaultTxContext != null) {
			defaultTxContext.close();
			defaultTxContext = null;
		}
		if (xaResource == null) {
			xaResource = managedConnectionFactory.getXAResourceFactory().createXAResource();
			if (Debug.ON)
				logger.log(BasicLevel.DEBUG, "Creates a new XAResource: " + xaResource);
		}
		return xaResource;
	}

	public LocalTransaction getLocalTransaction() throws ResourceException {
		return this;
	}

	public ManagedConnectionMetaData getMetaData() throws ResourceException {
		return this;
	}

	/**
	 * Not supported yet.
	 */
	public void setLogWriter(PrintWriter writer) throws ResourceException {
		throw new ResourceException("JDO Connector: logging to PrintWriter not supported yet.");
	}

	/**
	 * Not supported yet.
	 */
	public PrintWriter getLogWriter() throws ResourceException {
		throw new ResourceException("JDO Connector: logging to PrintWriter not supported yet.");
	}

	// IMPLEMENTATION OF METHODS FROM THE (cci)LocalTransaction INTERFACE

	/**
	 * Begins the LocalTransaction if it has not already been started.
	 */
	public synchronized void begin() throws ResourceException {
		if (localTransactionTxContext != null) {
			throw new ResourceException("JDO Connector: a LocalTransaction has already begun.");
		}
		try {
			if (defaultTxContext != null) {
				defaultTxContext.close();
				defaultTxContext = null;
			}
			localTransactionTxContext
					= managedConnectionFactory.getTxContextFactory().getPersistenceManager();
			localTransactionTxContext.currentTransaction().begin();
			if (Debug.ON)
				logger.log(BasicLevel.DEBUG, "LocalTransaction begin - txContext: "
											 + localTransactionTxContext);
			if (xaResource != null)
				logger.log(BasicLevel.WARN,
						   "Begins a LocalTransaction on a ManagedConnection that runs in XA mode!");
			ConnectionEvent ce = new ConnectionEvent(this, ConnectionEvent.LOCAL_TRANSACTION_STARTED);
			if (associatedListeners != null) {
                for (Iterator it = associatedListeners.iterator(); it.hasNext();)
				    ((ConnectionEventListener) it.next()).localTransactionStarted(ce);
            }
		} catch (JDOException fe) {
			ResourceException re = new ResourceException(
					"JDO Connector: cannot begin LocalTransaction [nested exception].");
			re.setLinkedException(fe);
			throw re;
		}
	}

	/**
	 * Commits the LocalTransaction if it is active.
	 */
	public synchronized void commit() throws ResourceException {
		if (localTransactionTxContext == null)
			throw new ResourceException("JDO Connector: no LocalTransaction has been begun.");
		if (Debug.ON)
			logger.log(BasicLevel.DEBUG, "LocalTransaction commit - txContext: "
										 + localTransactionTxContext);
		try {
			localTransactionTxContext.currentTransaction().commit();
			ConnectionEvent ce = new ConnectionEvent(this, ConnectionEvent.LOCAL_TRANSACTION_COMMITTED);
            if (associatedListeners != null) {
			    for (Iterator it = associatedListeners.iterator(); it.hasNext();)
				    ((ConnectionEventListener) it.next()).localTransactionCommitted(ce);
            }
		} catch (JDOException fe) {
			ResourceException re = new ResourceException(
					"JDO Connector: cannot commit LocalTransaction [nested exception].");
			re.setLinkedException(fe);
			throw re;
		} finally {
			localTransactionTxContext.close();
			localTransactionTxContext = null;
		}
	}

	/**
	 * Rollbacks the LocalTransaction if it is active.
	 */
	public synchronized void rollback() throws ResourceException {
		if (localTransactionTxContext == null)
			throw new ResourceException("JDO Connector: no LocalTransaction has been begun.");
		if (Debug.ON)
			logger.log(BasicLevel.DEBUG, "LocalTransaction rollback - txContext: "
										 + localTransactionTxContext);
		try {
			localTransactionTxContext.currentTransaction().rollback();
			ConnectionEvent ce = new ConnectionEvent(this, ConnectionEvent.LOCAL_TRANSACTION_ROLLEDBACK);
            if (associatedListeners != null) {
    			for (Iterator it = associatedListeners.iterator(); it.hasNext();)
	    			((ConnectionEventListener) it.next()).localTransactionRolledback(ce);
            }
		} catch (JDOException fe) {
			ResourceException re = new ResourceException(
					"JDO Connector: cannot rollback LocalTransaction [nested exception].");
			re.setLinkedException(fe);
			throw re;
		} finally {
			localTransactionTxContext.close();
			localTransactionTxContext = null;
		}
	}

	// IMPLEMENTATION OF METHODS FROM THE PersistenceManager INTERFACE

	public boolean isClosed() {
		return false;
	}

	public void close() {
	}

	public Transaction currentTransaction() {
		return getPersistenceManager().currentTransaction();
	}

	public void evict(Object o) {
		getPersistenceManager().evict(o);
	}

	public void evictAll(Object[] objects) {
		getPersistenceManager().evictAll(objects);
	}

	public void evictAll(Collection collection) {
		getPersistenceManager().evictAll(collection);
	}

	public void evictAll() {
		getPersistenceManager().evictAll();
	}

	public void refresh(Object o) {
		getPersistenceManager().refresh(o);
	}

	public void refreshAll(Object[] objects) {
		getPersistenceManager().refreshAll(objects);
	}

	public void refreshAll(Collection collection) {
		getPersistenceManager().refreshAll(collection);
	}

	public void refreshAll() {
		getPersistenceManager().refreshAll();
	}

	public Query newQuery() {
		return getPersistenceManager().newQuery();
	}

	public Query newQuery(Object o) {
		return getPersistenceManager().newQuery(o);
	}

	public Query newQuery(String s, Object o) {
		return getPersistenceManager().newQuery(s, o);
	}

	public Query newQuery(Class aClass) {
		return getPersistenceManager().newQuery(aClass);
	}

	public Query newQuery(Extent extent) {
		return getPersistenceManager().newQuery(extent);
	}

	public Query newQuery(Class aClass, Collection collection) {
		return getPersistenceManager().newQuery(aClass, collection);
	}

	public Query newQuery(Class aClass, String s) {
		return getPersistenceManager().newQuery(aClass, s);
	}

	public Query newQuery(Class aClass, Collection collection, String s) {
		return getPersistenceManager().newQuery(aClass, collection, s);
	}

	public Query newQuery(Extent extent, String s) {
		return getPersistenceManager().newQuery(extent, s);
	}

	public Extent getExtent(Class aClass, boolean b) {
		return getPersistenceManager().getExtent(aClass, b);
	}

	public Object getObjectById(Object o, boolean b) {
		return getPersistenceManager().getObjectById(o, b);
	}

	public Object getObjectId(Object o) {
		return getPersistenceManager().getObjectId(o);
	}

	public Object getTransactionalObjectId(Object o) {
		return getPersistenceManager().getTransactionalObjectId(o);
	}

	public Object newObjectIdInstance(Class aClass, String s) {
		return getPersistenceManager().newObjectIdInstance(aClass, s);
	}

	public void makePersistent(Object o) {
		getPersistenceManager().makePersistent(o);
	}

	public void makePersistentAll(Object[] objects) {
		getPersistenceManager().makePersistentAll(objects);
	}

	public void makePersistentAll(Collection collection) {
		getPersistenceManager().makePersistentAll(collection);
	}

	public void deletePersistent(Object o) {
		getPersistenceManager().deletePersistent(o);
	}

	public void deletePersistentAll(Object[] objects) {
		getPersistenceManager().deletePersistentAll(objects);
	}

	public void deletePersistentAll(Collection collection) {
		getPersistenceManager().deletePersistentAll(collection);
	}

	public void makeTransient(Object o) {
		getPersistenceManager().makeTransient(o);
	}

	public void makeTransientAll(Object[] objects) {
		getPersistenceManager().makeTransientAll(objects);
	}

	public void makeTransientAll(Collection collection) {
		getPersistenceManager().makeTransientAll(collection);
	}

	public void makeTransactional(Object o) {
		getPersistenceManager().makeTransactional(o);
	}

	public void makeTransactionalAll(Object[] objects) {
		getPersistenceManager().makeTransactionalAll(objects);
	}

	public void makeTransactionalAll(Collection collection) {
		getPersistenceManager().makeTransactionalAll(collection);
	}

	public void makeNontransactional(Object o) {
		getPersistenceManager().makeNontransactional(o);
	}

	public void makeNontransactionalAll(Object[] objects) {
		getPersistenceManager().makeNontransactionalAll(objects);
	}

	public void makeNontransactionalAll(Collection collection) {
		getPersistenceManager().makeNontransactionalAll(collection);
	}

	public void retrieve(Object o) {
		getPersistenceManager().retrieve(o);
	}

	public void retrieveAll(Collection collection) {
		getPersistenceManager().retrieveAll(collection);
	}

	public void retrieveAll(Object[] objects) {
		getPersistenceManager().retrieveAll(objects);
	}

	public void setUserObject(Object o) {
		getPersistenceManager().setUserObject(o);
	}

	public Object getUserObject() {
		return getPersistenceManager().getUserObject();
	}

	public PersistenceManagerFactory getPersistenceManagerFactory() {
		return getPersistenceManager().getPersistenceManagerFactory();
	}

	public Class getObjectIdClass(Class aClass) {
		return getPersistenceManager().getObjectIdClass(aClass);
	}

	public void setMultithreaded(boolean b) {
		getPersistenceManager().setMultithreaded(b);
	}

	public boolean getMultithreaded() {
		return getPersistenceManager().getMultithreaded();
	}

	public void setIgnoreCache(boolean b) {
		getPersistenceManager().setIgnoreCache(b);
	}

	public boolean getIgnoreCache() {
		return getPersistenceManager().getIgnoreCache();
	}
}