/**
 * perseus/connector: this is an implementation of some JCA-related technologies
 * (resource adapters and managers) for the ObjectWeb consortium.
 * Copyright (C) 2001-2002 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 */

package org.objectweb.speedo.jca;

import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.speedo.api.Debug;

import javax.transaction.xa.XAException;
import javax.transaction.xa.XAResource;
import javax.transaction.xa.Xid;
import javax.transaction.Synchronization;
import javax.transaction.TransactionManager;
import javax.transaction.RollbackException;
import javax.jdo.PersistenceManager;
import javax.jdo.JDOException;
import javax.jdo.Transaction;

/**
 * @author	P. Dechamboux
 */
public class JdoXAResource implements XAResource {
    /**
     * The logger into which traces about JdoXAResource are produced.
     */
    private Logger logger;
    /**
     * The XID of the transaction associated with this XAResource.
     * txContext and xid work together: they should be both null
     * or not at the same time.
     */
    Xid xid = null;
    /**
     * The JdoTransaction associated to the "xid" transaction.
     * txContext and xid work together: they should be both null
     * or not at the same time.
     */
    PersistenceManager txContext = null;

    /**
     * The factory that has requtested the creation of this XAResource.
     */
    JdoXAResourceFactory xaResourceFactory;

    public boolean synchronizationRegistered = false;

    /**
     * Constructs a JdoTxContext.
     * @param el	The logger to be used to produced XAResource-related traces.
     * @param fxarf	The factory that has requested the creation of this
     * 				JdoTxContext.
     */
    JdoXAResource(Logger el, JdoXAResourceFactory fxarf) {
        logger = el;
        xaResourceFactory = fxarf;
    }

    /**
     * Gives access to the PersistenceManager associated with this XA resource.
     */
    PersistenceManager getTxContext() {
        return txContext;
    }

    /**
     * Assigns the relevant TxContext to this XAResource.
     */
    private void getTxContext(Xid xid, boolean create)
            throws XAException {
        if (this.xid != null) {
            if (this.xid != xid) {
                logger.log(BasicLevel.ERROR, "We are inside a start/end section: should be the same xid !!!");
                throw new XAException(XAException.XAER_PROTO);
            } else if (Debug.ON && logger.isLoggable(BasicLevel.DEBUG)) {
                logger.log(BasicLevel.DEBUG, "XAResource already inside DTP context: "
                    + xid
                    + " - ignored.");
            }
            return;
        }
        if (txContext != null) {
            if (!create) {
                return;
            }
            throw new XAException(XAException.XAER_PROTO);
        }
        // Try getting the JdoTxContext associated to this xid
        txContext = xaResourceFactory.getTxContext(xid);
        if (txContext == null) {
            if (!create) {
                logger.log(BasicLevel.ERROR, "No transaction exist with this XID and must not create one!!");
                throw new XAException(XAException.XAER_PROTO);
            }
            try {
                txContext = xaResourceFactory.createTxContext(xid);
            } catch (JDOException fe) {
                logger.log(BasicLevel.ERROR, "Error during the creation of the XA resource with the xid: " + xid, fe);
                throw new XAException(XAException.XAER_RMERR);
            }
        }
        this.xid = xid;
    }


    public synchronized void registerSynchronization() throws XAException {
        if (!synchronizationRegistered) {
            Transaction tx = txContext.currentTransaction();
            if (tx.isActive()) {
                return;
            }
            logger.log(BasicLevel.DEBUG, "registerSynchronization");
            TransactionManager tm = xaResourceFactory
                    .getJdoManagedConnectionFactory().getTransactionManager();
            if (tm != null && txContext instanceof Synchronization) {
                try {
                    tm.getTransaction()
                            .registerSynchronization((Synchronization) txContext);
                    tx.begin();
                    synchronizationRegistered = true;
                } catch (RollbackException e) {
                    //The JDO container cannot be registered then the transaction
                    // must be rolledback
                    tx.rollback();
                } catch (Exception e) {
                    String msg = "Impossible to register the JDO " +
                            "container as synchronization on the current" +
                            " transaction: " + e.getMessage();
                    logger.log(BasicLevel.ERROR, msg, e);
                    throw new XAException(msg);
                }
            } else {
                String msg = "Impossible to register the JDO " +
                        "container as synchronization on the current transaction: " +
                        (tm == null
                        ? "The transaction manager has not found with the JNDI name: "
                        + xaResourceFactory.getJdoManagedConnectionFactory().getTransactionManagerJNDIName()
                        : "The Transaction object of the JDO implementation does not implement javax.transaction.Synchronization");
                logger.log(BasicLevel.ERROR, msg);
                throw new XAException(msg);
            }
        }
    }

    // IMPLEMENTATION OF METHODS FROM THE (jta)XAResource INTERFACE

    // Management of transaction enrolment: start, end, isSameRM

    /**
     * Assigns an actual JDO transaction context to the XAResource
     * within the give DTP context.
     */
    public void start(Xid xid, int i) throws XAException {
        switch(i) {
        case XAResource.TMRESUME:
            logger.log(BasicLevel.DEBUG, "start(" + xid + ", TMRESUME)");
            getTxContext(xid, false);
            break;
        case XAResource.TMJOIN:
            logger.log(BasicLevel.DEBUG, "start(" + xid + ", TMJOIN)");
            getTxContext(xid, false);
            break;
        case XAResource.TMNOFLAGS:
            logger.log(BasicLevel.DEBUG, "start(" + xid + ", TMNOFLAGS)");
            getTxContext(xid, true);
            break;
        default:
            String msg = "Unexpected flag: " + i;
            logger.log(BasicLevel.ERROR, msg);
            throw new XAException(msg);
        }
    }

    /**
     * Unbind this instance to the PersistenceManager.
     */
    public void end(Xid xid, int i) throws XAException {
        if (this.xid != xid)
            throw new XAException(XAException.XAER_PROTO);
        txContext = null;
    }

    /**
     * Used by JTA in order to verify that it has not already registered a
     * XAResource to manage this transaction context from this RM. If it is the
     * case, both XAResource are warned of this fact (endWithNoSameRM = false).
     * This means that these XAResource will be released at "end" time.
     * @param resource	The resource to be compared against this one wrt RM.
     */
    public boolean isSameRM(XAResource resource) throws XAException {
        boolean res = (resource instanceof JdoXAResource)
                && ((JdoXAResource) resource).xaResourceFactory == xaResourceFactory;
        if (Debug.ON && logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, "isSameRM(" + resource + "): " + res);
        }
        return res;
    }

    // Management of transaction termination: prepare, commit, rollback

    /**
     * Prepares the underlying JdoTxContext (prepare phase of the 2PC).
     */
    public int prepare(Xid xid) throws XAException {
        getTxContext(xid, false);
        return XAResource.XA_OK;
    }

    /**
     * Unbind the PersistenceManager to the xid. The real commit is done by the
     * PersistenceManager registered as a Synchronization.
     */
    public void commit(Xid xid, boolean b) throws XAException {
        logger.log(BasicLevel.DEBUG, "commit(" + xid + ", " + b + ")");
        xaResourceFactory.releaseTxContext(xid);
        this.xid = null;
        txContext = null;
        synchronizationRegistered = false;
    }

    /**
     * Unbind the PersistenceManager to the xid. The real commit is done by the
     * PersistenceManager registered as a Synchronization.
     */
    public void rollback(Xid xid) throws XAException {
        logger.log(BasicLevel.DEBUG, "rollback(" + xid + ")");
        xaResourceFactory.releaseTxContext(xid);
        this.xid = null;
        txContext = null;
        synchronizationRegistered = false;
    }

    // Management of recovery: forget, recover

    /**
     *
     */
    public void forget(Xid xid) throws XAException {
        xaResourceFactory.releaseTxContext(xid);
    }

    /**
     * Gets the Xid of distributed transactions to be recovered from the JDO
     * manager.
     */
    public Xid[] recover(int i) throws XAException {
        return new Xid[0];
    }

    // Management of transaction timeout: getTransactionTimeout,
    // setTransactionTimeout

    /**
     *
     */
    public int getTransactionTimeout() throws XAException {
        return 0;
    }

    /**
     *
     */
    public boolean setTransactionTimeout(int i) throws XAException {
        return false;
    }
}
