/**
 * perseus/connector: this is an implementation of some JCA-related technologies
 * (resource adapters and managers) for the ObjectWeb consortium.
 * Copyright (C) 2001-2002 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 */

package org.objectweb.speedo.jca;

import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.speedo.api.Debug;

import java.util.HashMap;
import javax.transaction.xa.Xid;
import javax.jdo.PersistenceManager;

/**
 * @author	P. Dechamboux
 */
public class JdoXAResourceFactory {
	/**
	 * The logger into which traces about JdoXAResourceFactory are produced.
	 */
	private Logger logger;

    private JdoManagedConnectionFactory mcf;
	/**
	 * The hashed structure that stores JdoTransaction that have been associated
	 * with a particular XID.
	 */
	private HashMap txContexts = new HashMap();

	/**
	 * Constructs a JdoXAResourceFactory.
	 */
    JdoXAResourceFactory(Logger logger, JdoManagedConnectionFactory mcf) {
		this.logger = logger;
		if (Debug.ON && logger.isLoggable(BasicLevel.DEBUG))
			logger.log(BasicLevel.DEBUG, "Building of a new JdoXAResourceFactory: " + this);
        this.mcf = mcf;
	}

	/**
	 * Creates a new XAResource for coordinating transaction in a potentially
	 * multi-RM environment.
	 */
	JdoXAResource createXAResource() {
		JdoXAResource res = new JdoXAResource(logger, this);
		if (Debug.ON && logger.isLoggable(BasicLevel.DEBUG))
			logger.log(BasicLevel.DEBUG, "Creates a XAResource: " + res);
		return res;
	}

	/**
	 * Releases a XAResource that is not used any more.
	 * @param xar	The released XAResource.
	 */
	void releaseXAResource(JdoXAResource xar) {
		if (Debug.ON && logger.isLoggable(BasicLevel.DEBUG))
			logger.log(BasicLevel.DEBUG, "Releases a XAResource: " + xar);
	}

	/**
	 * Looks for a JdoTxContext associated with the particular transaction.
	 * @param xid	The DTP transaction identifier.
	 */
	PersistenceManager getTxContext(Xid xid) {
		PersistenceManager txc = (PersistenceManager) txContexts.get(xid);
        if (Debug.ON && logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG,
                    "Looking for the TxContext associated with XID (" + xid
                    + "): " + txc);
		return txc;
	}

	/**
	 * Creates a JdoTxContext and associates it with the given DTP transaction.
	 * @param xid	The DTP transaction identifier.
	 */
	PersistenceManager createTxContext(Xid xid) {
		PersistenceManager txc = mcf.getTxContextFactory().getPersistenceManager();
        if (Debug.ON && logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, "Creates a JdoTxContext (" + txc
                    + ") to be associated with XID: " + xid);
        }
		txContexts.put(xid, txc);
		return txc;
	}

	/**
	 * Releases a DTP transaction context and its related resources.
	 * @param xid	The DTP transaction identifier.
	 */
	void releaseTxContext(Xid xid) {
		PersistenceManager txc = (PersistenceManager) txContexts.remove(xid);
		if (txc == null) {
			logger.log(BasicLevel.WARN, "Try to release unknow XID context: " + xid);
		} else if (Debug.ON && logger.isLoggable(BasicLevel.DEBUG)) {
                logger.log(BasicLevel.DEBUG, "Dissociates the JdoTxContext ("
                        + txc + ") associated with XID: " + xid);
        }
	}

    JdoManagedConnectionFactory getJdoManagedConnectionFactory() {
        return mcf;
    }
}
