/**
 * Copyright (C) 2001-2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.runtime;

import org.objectweb.fractal.api.control.UserBindingController;
import org.objectweb.jorm.api.PClassMapping;
import org.objectweb.jorm.api.PClassMappingCtrl;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PMapCluster;
import org.objectweb.jorm.api.PMapper;
import org.objectweb.jorm.metainfo.api.Package;
import org.objectweb.jorm.metainfo.lib.BasicClass;
import org.objectweb.jorm.metainfo.lib.BasicCompositeName;
import org.objectweb.jorm.metainfo.lib.JormManager;
import org.objectweb.jorm.naming.api.PBinder;
import org.objectweb.jorm.naming.api.PName;
import org.objectweb.jorm.naming.api.PNameCoder;
import org.objectweb.jorm.naming.api.PNameManager;
import org.objectweb.jorm.naming.api.PNamingContext;
import org.objectweb.jorm.lib.JormPathHelper;
import org.objectweb.perseus.cache.api.CacheManager;
import org.objectweb.speedo.api.JormFactory;
import org.objectweb.speedo.api.SpeedoGenClassProxy;
import org.objectweb.speedo.api.SpeedoProperties;
import org.objectweb.speedo.api.SpeedoProxy;
import org.objectweb.speedo.api.JormFactoryAttributes;
import org.objectweb.speedo.generation.lib.NamingRules;
import org.objectweb.speedo.naming.lib.NamingManagerFactory;
import org.objectweb.speedo.naming.api.NamingManager;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;

import javax.jdo.JDOException;
import javax.jdo.JDOFatalInternalException;
import javax.jdo.JDOUserException;
import java.io.File;
import java.io.ObjectInputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * This class manages naming used in Speedo (USER_ID and DATASTRORE_ID.
 * In JDO the binder instance can be used also as PNamingContext because all
 * classes in a class hierarchy must have the same identifier strucutre.
 * Then it is not necessary to use intermediates PNamingContext. Then a PName is
 * always valid inside a PBinder.
 *
 * @author S.Chassande-Barrioz
 */
public class BasicJormFactory
        implements JormFactory, UserBindingController,
		NamingManager.PBinderProvider, JormFactoryAttributes {

    public final static String MAPPER_BINDING = "mapper";
    public final static String LOGGER_FACTORY_BINDING = "logger-factory";
    public final static String CACHE_MANAGER_BINDING = "cache-manager";
    public final static String LOGGER_NAME
            = SpeedoProperties.LOGGER_NAME + ".rt.jormFactory";

    public final static String CID_BINDER_CLASS_NAM
            = "org.objectweb.speedo.runtime.binder.RdbSequenceBinder";

    public final static String BASID_BINDER_CLASS_NAM
            = "org.objectweb.jorm.facility.naming.basidir.BasidBinder";

    public final static boolean CODING_ONLY = false;

    /**
     * The mapper is used to map and find the Jorm classes.
     */
    protected PMapper mapper = null;

    protected LoggerFactory loggerFactory = null;

    protected CacheManager cache = null;

    /**
     * It contains the association between a class name and its binder. This map
     * contains only the binders which manage user pname structures.
     */
    protected Map binders = null;
    protected Map pnamingContexts = null;
    protected Logger logger = null;
    protected Map cn2pcm = null;
    private NamingManagerFactory nmf;

	private byte mappingStructuresRule = JormFactoryAttributes.CREATE_IF_REQUIRED;

    /**
     * builds the BasicJormFactory
     * and the hosting structure for the binder list.
     */
    public BasicJormFactory() {
        pnamingContexts = new HashMap();
        binders = new HashMap();
        cn2pcm = new HashMap();
        nmf = new NamingManagerFactory();
    }

    public PBinder getPBinder(String className) {
        return (PBinder) binders.get(className);
    }

	// IMPLEMENTATION OF THE JormFactoryAttributes INTERFACE //
	//-------------------------------------------------------//

	public byte getMappingStructureRule() {
		return mappingStructuresRule;
	}

	public void setMappingStructureRule(byte rule) {
		mappingStructuresRule = rule;
	}

    // IMPLEMENTATION OF THE UserBindingController INTERFACE //
    //-------------------------------------------------------//

    public Object getFcBindings(String s) {
        if (LOGGER_FACTORY_BINDING.equals(s))
            return loggerFactory;
        else if (MAPPER_BINDING.equals(s))
            return mapper;
        else if (CACHE_MANAGER_BINDING.equals(s))
            return cache;
        else
            return null;
    }

    public void addFcBinding(String s, Object o) {
        if (LOGGER_FACTORY_BINDING.equals(s)) {
            loggerFactory = (LoggerFactory) o;
            if (loggerFactory != null) {
                logger = loggerFactory.getLogger(LOGGER_NAME);
                nmf.setLogger(logger);
            }
        } else if (MAPPER_BINDING.equals(s)) {
            mapper = (PMapper) o;
            nmf.setMapper(mapper);
        } else if (CACHE_MANAGER_BINDING.equals(s)) {
            cache = (CacheManager) o;
        }
    }

    public void removeFcBinding(String s, Object serverItf) {
        if (LOGGER_FACTORY_BINDING.equals(s)) {
            loggerFactory = null;
        } else if (MAPPER_BINDING.equals(s)) {
            mapper = null;
            nmf.setMapper(null);
        } else if (CACHE_MANAGER_BINDING.equals(s)) {
            cache = null;
        }
    }


    // IMPLEMENTATION OF THE JormFactory INTERFACE //
    //---------------------------------------------//

    /**
     * @return the PBinder instance which manages the identifer of the
     * SpeedoProxy instance
     */
    public PBinder getPBinder(SpeedoProxy sp) throws PException {
        String clName = sp.getClass().getName();
        PClassMapping pcm = mapper.lookup(clName);
        if (pcm != null)
            return pcm.getPBinder();
        return getPClassMapping(sp).getPBinder();
    }

    /**
     * This Method is a shortcut to the getPBinder(SpeedoProxy) method. It only
     * does the instanciation of the classname
     * (Class.ForName(classsName).newInstance()) and call the
     * getPBinder(SpeedoProxy) method with the created instance. Then if an
     * instance is availlable it is better to use the other method.
     * @param clazz the Jorm class name managed by the wanted PBinder
     * @return the PBinder instance to use for the given jorm class name
     * @throws PException
     */
    public PBinder getPBinder(Class clazz) throws PException {
        String clName = clazz.getName();
        PClassMapping pcm = mapper.lookup(clName);
        if (pcm != null)
            return pcm.getPBinder();
        return getPClassMapping(getInstance(clazz)).getPBinder();
    }

    public PBinder getPBinder(String classname, ClassLoader cl) throws PException {
        PClassMapping pcm = mapper.lookup(classname);
        if (pcm != null)
            return pcm.getPBinder();
        return getPClassMapping(getInstance(classname, cl)).getPBinder();
    }

    public ClassLoader getClassLoader(String className) {
        Object o = mapper.lookup(className);
        if (o == null) {
            return null;
        }
        ClassLoader cl = o.getClass().getClassLoader();
        if (cl == null) {
            cl = ClassLoader.getSystemClassLoader();
        }
        return cl;
    }

    /**
     * Retrieves a PClassMapping instance managing a SpeedoProxy. If the
     * PClassMapping does not already exist then it is created, configured and
     * mapped in the PMapper. The PClassMapping configuration contains the
     * PBinder assignement, the PNamingContext assignement for each reference,
     * and the GenClassMapping assignement. The assigned GenClassMapping must
     * also be configured. In other hand the PClassMapping of referenced class
     * is not configured now. Only the PNamingContext is needed.
     * @param sp is the SpeedoProxy instance which the PClassMapping is required
     * @return the PClassMapping which manages a SpeedoProxy
     * @throws PException
     */
    public PClassMapping getPClassMapping(SpeedoProxy sp) throws PException {
        boolean debug = logger.isLoggable(BasicLevel.DEBUG);
        if (sp instanceof SpeedoGenClassProxy) {
            // The GenClassMapping must be already created before
            SpeedoGenClassProxy sgcp = (SpeedoGenClassProxy) sp;
            if (debug) {
                logger.log(BasicLevel.DEBUG,
                        "Looking for the PClassMapping of the generic class "
                        + sgcp.jdoGetGenClassId());
            }
            return JormPathHelper.getPClassMapping(sgcp.jdoGetGenClassId(), mapper);
        }
        String clName = sp.getClass().getName();
        PClassMapping pcm = mapper.lookup(clName);
        if (pcm == null) {
            synchronized (this) {
                pcm = (PClassMapping) cn2pcm.get(clName);
                if (pcm != null) {
                    return pcm;
                }
                pcm = mapper.lookup(clName);
                if (pcm == null) {
                    // Creates the PClassMapping, its PBinder and the PNC
                    if (debug) {
                        logger.log(BasicLevel.DEBUG,
                                "Looking for the PClassMapping of the class " + clName);
                    }
                    ClassLoader classLoader = sp.getClass().getClassLoader();
                    if (classLoader == null) {
                        classLoader = ClassLoader.getSystemClassLoader();
                    }

                    String pcmcn = NamingRules.fqMappingName(
                            clName, sp.jdoGetMapperName());
                    try {
                        pcm = (PClassMapping) classLoader.loadClass(pcmcn).newInstance();
                    } catch (Exception e) {
                        throw new PException(e,
                                "Impossible to instanciate the PClassMapping of the class "
                                + clName + ": " + pcmcn);
                    }
                    if (debug) logger.log(BasicLevel.DEBUG, "Jorm config:");
                    String[][] jormConf = (String[][]) sp.getJormConfig();
                    Map cn2bcn = new HashMap(jormConf.length);
                    for (int i = 0; i < jormConf.length; i++) {
                        logger.log(BasicLevel.DEBUG, jormConf[i][0] + " ==> " + jormConf[i][1]);
                        cn2bcn.put(jormConf[i][0], jormConf[i][1]);
                    }

                    // Fetch the Binder by the DataStoreNaming
                    PBinder binder = findPBinder(
                            (String) cn2bcn.get(JormPathHelper.getPath(clName)),
                            clName, classLoader);
                    pcm.setPBinder(binder);
                    binder.setPClassMapping(pcm);
                    //Assign the PNamingContext of the reference
                    RefConfig refConfig = new RefConfig(cn2bcn, sp.jdoGetMapperName(), classLoader);
                    cn2pcm.put(clName, pcm);
                    pcm.configureRefFields(refConfig);
                    cn2pcm.remove(clName);
                    if (debug) {
                        logger.log(BasicLevel.DEBUG, "PClassMapping/PBinder/PNC created for the class " + clName);
                        logger.log(BasicLevel.DEBUG, "- pcm: " + pcm);
                        logger.log(BasicLevel.DEBUG, "- binder: " + binder);
                    }

                    // do the 'Map' operation on the mapper and register the pcm
                    // in the local map
                    ClassLoader cl = sp.getClass().getClassLoader();
                    if (cl == null) {
                        cl = ClassLoader.getSystemClassLoader();
                    }
                    loadJormMI(sp, cl);
                    mapper.map(null, pcm);
                    logger.log(BasicLevel.INFO, "Class " + clName + " mapped");
                    PMapCluster cluster = mapper.getPMappingStructuresManager().getPMapCluster(clName);
                    if (cluster.isDefined()) {
						if (mappingStructuresRule >= CREATE_IF_REQUIRED) {
                        	cluster.createMappingStructures(
									mappingStructuresRule == FORCE_CREATE);
							if (mappingStructuresRule == DELETE_DATA) {
								cluster.deleteData();
							}
						}
                        if (logger.isLoggable(BasicLevel.INFO)) {
                            logger.log(BasicLevel.INFO, "Cluster of Class "
                                    + cluster.getClusterClasses()
									+ " initialized ("
									+ mappingStructuresRule + ").");
                        }
                    } else {
                        getPClassMapping((String) cluster
                                .getUnResolvedDependencies().iterator().next(),
                                classLoader);
                    }
                } else if (logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "PClassMapping for class " + clName
                            + " already exist:" + pcm);
            }
        }
        return pcm;
    }

    /**
     * This Method is a shortcut to the getPClassMapping(SpeedoProxy) method. It only
     * does the instanciation of the classname
     * (Class.ForName(classsName).newInstance()) and call the
     * getPClassMapping(SpeedoProxy) method with the created instance. Then if an
     * instance is availlable it is better to use the other method.
     *
     * Retrieves a PClassMapping instance managing a jorm class. If the
     * PClassMapping does not already exist then it is created, configured and
     * mapped in the PMapper. The PClassMapping configuration contains the
     * PBinder assignement, the PNamingContext assignement for each reference,
     * and the GenClassMapping assignement. The assigned GenClassMapping must
     * also be configured. In other hand the PClassMapping of referenced class
     * is not configured now. Only the PNamingContext is needed.
     * @param classname is the class name of the jorm class which the
     * PClassMapping is required
     * @return the PClassMapping which manages a SpeedoProxy
     * @throws PException
     */
    public PClassMapping getPClassMapping(String classname, ClassLoader cl) throws PException {
        PClassMapping pcm = mapper.lookup(classname);
        if (pcm == null)
            pcm = getPClassMapping(getInstance(classname, cl));
        return pcm;
    }

    public PClassMapping getPClassMapping(Class clazz) throws PException {
        PClassMapping pcm = mapper.lookup(clazz.getName());
        if (pcm == null)
            pcm = getPClassMapping(getInstance(clazz));
        return pcm;
    }

    /**
     * @return the PNamingContext instance which manages a reference to the
     * SpeedoProxy instance
     */
    public PNamingContext getPNamingContext(SpeedoProxy sp) throws PException {
        String clName = sp.getClass().getName();
        String path = JormPathHelper.getPath(clName);
        String[][] jormConf = (String[][]) sp.getJormConfig();
        int i;
        for (i = 0; i < jormConf.length && !path.equals(jormConf[i][0]); i++);
        if (i < jormConf.length) {
            ClassLoader cl = sp.getClass().getClassLoader();
            if (cl == null) {
                cl = getClass().getClassLoader();
            }
            if (cl == null) {
                cl = ClassLoader.getSystemClassLoader();
            }
            return (PNamingContext) findPNameManager(jormConf[i][1], clName, cl);
        } else {
            throw new PException(
                    "Impossible to find the PNamingContext for the class "
                    + clName);
        }
    }

    /**
     * This Method is a shortcut to the getPNamingContext(SpeedoProxy) method.
     * It only does the instanciation of the classname
     * (Class.ForName(classsName).newInstance()) and call the
     * getPNamingContext(SpeedoProxy) method with the created instance. Then if
     * an instance is availlable it is better to use the other method.
     * @param classname the Jorm class name managed by the wanted PNamingContext
     * @return the PNamingContext instance to use for the given jorm class name
     * @throws PException
     */
    public PNamingContext getPNamingContext(String classname, ClassLoader cl) throws PException {
        return getPNamingContext(getInstance(classname, cl));
    }

    public PNamingContext getPNamingContext(Class clazz) throws PException {
        return getPNamingContext(getInstance(clazz));
    }

    public PName decode(Class clazz, Object oid) throws PException {
        if (oid instanceof PName) {
            return (PName) oid;
        }
        PBinder binder = null;
        if (clazz != null) {
            binder = getPBinder(clazz);
        }
        PName pn = nmf.decode(binder, oid, clazz, this);
        if (pn != null) {
            return pn;
        }
        if (binder != null) {
            return binder.decodeAbstract(oid, null);
        }
        throw new JDOUserException("Impossible to decode the identifier "
                + oid  + (clazz == null
                ? "" : " of the class " + clazz.getName()));
    }

    /**
     * It retrieves a representation of an identifier since a PName.
     *
     * @param pn is the pname to encode
     * @return an identifier representation
     * @throws PException if the pname is malformed (null for example).
     */
    public Object encode(PName pn) throws PException {
        return nmf.encode(pn);
    }

    // PRIVATE METHODS //
    //-----------------//

    private SpeedoProxy getInstance(Class clazz) throws PException {
        try {
            return (SpeedoProxy) clazz.newInstance();
        } catch (ClassCastException e) {
            throw new PException("The class '" + clazz.getName()
                    + "' must be enhanced by Speedo before the runtime");
        } catch (Exception e) {
            throw new PException(e, "Impossible to instanciate the class "
                    + clazz.getName());
        }
    }

    private SpeedoProxy getInstance(String classname, ClassLoader cl) throws PException {
        if (cl == null) {
            throw new PException("Impossible to instanciate the class "
                    + classname + " without a classLoader(" + cl + ")");
        }
        try {
            return (SpeedoProxy) getClass(classname, cl).newInstance();
        } catch (PException e) {
			throw e;
        } catch (Exception e) {
            throw new PException(e, "Impossible to instanciate the class '"
                    + classname + "' with the class loader :" + cl);
        }
    }

    private Class getClass(String classname, ClassLoader cl) throws PException {
        if (cl == null) {
            throw new PException("Impossible to load the class "
                    + classname + " without a classLoader(" + cl + ")");
        }
        try {
            return cl.loadClass(classname);
        } catch (Exception e) {
            throw new PException(e, "Impossible to load the class '"
                    + classname + "' with the class loader :" + cl);
        }
    }

    public synchronized PBinder findPBinder(String bcn,
                                String clName,
                                ClassLoader classLoader) throws PException {
        if (bcn == null || bcn.length() == 0)
            throw new JDOFatalInternalException(
                    "Impossible to get the PNamingContext of the class " + clName
                    + ": Specified binder class is not valid: " + bcn);
        PBinder binder = (PBinder) binders.get(clName);
        if (binder != null)
            return binder;
        binder = nmf.getNamingManager(bcn, classLoader)
                .getPBinder(clName, bcn, classLoader, mappingStructuresRule);
        binder.setCacheManager(cache);
        binders.put(clName, binder);
        return binder;
    }

    protected synchronized PNameManager findPNameManager(String pnmcn,
                                            String clName,
                                            ClassLoader classLoader) throws PException {
        if (pnmcn == null || pnmcn.length() == 0)
            throw new JDOFatalInternalException(
                    "Impossible to get the PNamingContext of the class " + clName
                    + ": Specified PNameManager class is not valid: " + pnmcn);
        PNameManager pnm = (PNameManager) pnamingContexts.get(clName);
        if (pnm != null)
            return pnm;
        NamingManager nm = nmf.getNamingManager(pnmcn, classLoader);
        if (nm.supportPNamingcontext()) {
            pnm = nm.getPNamingContext(clName, pnmcn, classLoader, binders);
        } else {
            pnm = findPBinder(pnmcn, clName, classLoader);
        }
        pnamingContexts.put(clName, pnm);
        return pnm;
    }

    private class RefConfig implements PClassMapping.ReferenceConfigurator {

        Map cn2bcn = null;
        String mapperName = null;
        List gcms = null;
        ClassLoader classLoader = null;

        public RefConfig(Map cn2bcn, String mapperName, ClassLoader cl) {
            this.cn2bcn = cn2bcn;
            this.mapperName = mapperName;
            this.gcms = new ArrayList();
            this.classLoader = cl;
        }

        public List getGenClassMappings() {
            return gcms;
        }

        // IMPLEMENTATION OF THE PClassMapping.PNCGetter INTERFACE //
        //---------------------------------------------------------//


        public PNameCoder getPNameCoder(String sourceclassName,
                                        String refFieldName,
                                        String destclassName) {
            try {
                PNamingContext pnc = getPNamingContext(destclassName, classLoader);
                if (logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG,
                            "Assign the PNamingContext for the field " + sourceclassName
                            + "." + refFieldName + " which references the class "
                            + destclassName + ": " + pnc);
                logger.log(BasicLevel.DEBUG,
                        "getPNameManager(" + destclassName + ") ==>" + pnc);
                return pnc;
            } catch (PException e) {
                throw new JDOFatalInternalException(
                        "ERROR during the assignation of the PNameManager of the reference "
                        + refFieldName + " of the class " + sourceclassName, e);
            }
        }

        public PNameCoder getPNameCoder(String sourceclassName,
                                        String refFieldName,
                                        String[] genClassNames) {
            String cn = JormPathHelper.getPath(
                    sourceclassName, refFieldName, genClassNames);
            String pnmcn = (String) cn2bcn.get(cn);
            try {
                PNamingContext pnc = (PNamingContext) findPNameManager(pnmcn, cn, classLoader);
                logger.log(BasicLevel.DEBUG,
                        "getPNameManager(" + cn + ") ==>" + pnc);
                return pnc;
            } catch (PException e) {
                throw new JDOFatalInternalException(e.getMessage(), e);
            }
        }

        public PClassMapping getGenClassMapping(String sourceclassName,
                                                String refFieldName,
                                                String[] genClassNames) {

            String cn = JormPathHelper.getPath(
                    sourceclassName, refFieldName, genClassNames);
            String pnmcn = (String) cn2bcn.get(cn);
            logger.log(BasicLevel.DEBUG, "getGenClassMapping(" + cn + ")");
            logger.log(BasicLevel.DEBUG, "\tpnmcn=" + pnmcn);
            try {
                PClassMapping gcm = (PClassMapping)
                        Class.forName(getGCMClassName(mapperName)).newInstance();
                logger.log(BasicLevel.DEBUG, "\tgcm=" + gcm);
                PBinder pb = findPBinder(pnmcn, cn, classLoader);
                logger.log(BasicLevel.DEBUG, "\tpb=" + pb);
                gcm.setPBinder(pb);
                pb.setPClassMapping(gcm);
                gcms.add(gcm);
                return gcm;
            } catch (Exception e) {
                throw new JDOFatalInternalException(e.getMessage(), e);
            }
        }

        public PClassMapping getGenClassMapping(String sourceclassName,
                                                String refFieldName,
                                                String[] genClassNames,
                                                String destclassName) {
            String cn = JormPathHelper.getPath(
                    sourceclassName, refFieldName, genClassNames);
            String pnmcn = (String) cn2bcn.get(cn);
            try {
                PClassMapping gcm = (PClassMapping)
                        Class.forName(getGCMClassName(mapperName)).newInstance();
                logger.log(BasicLevel.DEBUG, "getGenClassMapping(" + cn + ")");
                logger.log(BasicLevel.DEBUG, "\tgcm=" + gcm);
                PBinder pb = findPBinder(pnmcn, cn, classLoader);
                logger.log(BasicLevel.DEBUG, "\tpb=" + pb);
                gcm.setPBinder(pb);
                pb.setPClassMapping(gcm);
                PNamingContext pnc = getPNamingContext(destclassName, classLoader);
                logger.log(BasicLevel.DEBUG, "\tpnc=" + pnc);
                ((PClassMappingCtrl) gcm).setPNameCoder(pnc);
                gcms.add(gcm);
                return gcm;
            } catch (Exception e) {
                throw new JDOFatalInternalException(e.getMessage(), e);
            }
        }
    }

    private void loadJormMI(SpeedoProxy sp, ClassLoader cl) {
        String clName = sp.getClass().getName();
        JormManager m = (JormManager) mapper.getMetaInfoManager();
        synchronized(m) {
            if (m.getClass(clName) != null) {
                return;
            }
            Set mos = null;
            String fn = sp.jdoGetJdoFileName();
            try {
                fn = fn.substring(0, fn.length() - 4);
                fn = fn.replace('.', File.separatorChar);
                fn += ".jmi";
                logger.log(BasicLevel.DEBUG,
                        "Loading the file " + fn + " from the classpath");
                ObjectInputStream ois = new ObjectInputStream(cl.getResourceAsStream(fn));
                mos = (Set) ois.readObject();
                ois.close();
            } catch (Exception e) {
                throw new JDOException("Impossible to load the jorm meta " +
                        "information for the class '" + clName + "' from the file '"
                        + fn + "' (Must be availlable in the classloader of the application).", e);
            }

            for (Iterator it = mos.iterator(); it.hasNext();) {
                Object o = it.next();
                if (o instanceof BasicClass) {
                    BasicClass c = (BasicClass) o;
                    Package s = m.createPackage(((Package) c.getParent()).getName());
                    s.addClass(c);
                    c.setParent(s);
                    c.setLogger(m.getLogger());
                    logger.log(BasicLevel.DEBUG,
                            "Jorm Meta Object Class " + c.getFQName() + " loaded");
                } else if (o instanceof BasicCompositeName) {
                    BasicCompositeName cn = (BasicCompositeName) o;
                    Package s = m.createPackage(((Package) cn.getParent()).getName());
                    s.addCompositeName(cn);
                    cn.setLogger(m.getLogger());
                    logger.log(BasicLevel.DEBUG,
                            "Jorm Meta Object CompositeName " + cn.getFQName() + " loaded");
                } else
                    throw new JDOException("Umanaged Jorm Meta Object " + o);
            }
            if (m.getClass(clName) == null) {
                throw new JDOException(
                    "Internal ERROR: No meta information found about the persistent class '"
                    + clName + "' in the file '" + fn + "'.");
            }
        }
    }

    private String getGCMClassName(String mapperName) {
        if (mapperName.startsWith("rdb")) {
            return "org.objectweb.jorm.mapper.rdb.genclass.RdbGenClassMapping";
        } else {
            throw new JDOFatalInternalException("Umanaged mapper: " + mapperName);
        }
    }
}
