/**
 * Copyright (C) 2001-2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package org.objectweb.speedo.runtime.persistence;

import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PMapper;
import org.objectweb.jorm.naming.api.PName;
import org.objectweb.perseus.cache.api.CacheEntry;
import org.objectweb.perseus.cache.api.CacheException;
import org.objectweb.perseus.concurrency.api.ConcurrencyManager;
import org.objectweb.perseus.concurrency.lib.Semaphore;
import org.objectweb.perseus.persistence.api.PersistenceException;
import org.objectweb.perseus.persistence.api.TransactionalPersistenceManager;
import org.objectweb.perseus.persistence.api.WorkingSet;
import org.objectweb.perseus.persistence.lib.AbstractPersistenceManager;
import org.objectweb.speedo.api.LifeCycle;
import org.objectweb.speedo.api.ProxyManager;
import org.objectweb.speedo.api.SpeedoAccessor;
import org.objectweb.speedo.api.SpeedoProxy;
import org.objectweb.speedo.api.Transaction;
import org.objectweb.speedo.api.JormFactory;
import org.objectweb.speedo.api.WorkingSetLifeCycle;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.medor.eval.prefetch.api.PrefetchCache;

import java.util.Iterator;
import java.util.List;

public class SpeedoTransactionalPersistenceManager
		extends AbstractPersistenceManager implements TransactionalPersistenceManager {

	public final static String MAPPER_BINDING = "mapper";
	public final static String JORM_FACTORY_BINDING = "jorm-factory";

	protected PMapper mapper = null;
	protected JormFactory jf = null;


	// IMPLEMENTATION OF THE UserBindingController INTERFACE //
	//-------------------------------------------------------//

	public Object getFcBindings(String s) {
		if (MAPPER_BINDING.equals(s))
			return mapper;
		else if (JORM_FACTORY_BINDING.equals(s))
			return jf;
		else
			return super.getFcBindings(s);
	}

	public void addFcBinding(String s, Object o) {
		if (MAPPER_BINDING.equals(s))
			mapper = (PMapper) o;
		else if (JORM_FACTORY_BINDING.equals(s))
			jf = (JormFactory) o;
		else
			super.addFcBinding(s, o);
	}

	public void removeFcBinding(String s, Object o) {
		if (MAPPER_BINDING.equals(s))
			mapper = null;
		else if (JORM_FACTORY_BINDING.equals(s))
			jf = null;
		else
			super.removeFcBinding(s, o);

	}

	/**
	 * Prepares the end of a context.
	 * @param tx is can be simple working set or a transaction
	 * @param cm is the ConcurrencyManager managing the context
	 * @return true if the context CAN be commit, false if the context MUST be
	 * rolled back
	 * @throws PersistenceException
	 */
	private boolean prepareWS(Transaction tx,
							  ConcurrencyManager cm,
							  boolean validate)
			throws PersistenceException {

		List entries = tx.entries();
		// prepare to write (relations consistency management)
		// iterator can not be used (prepareWrite can make
		// new objects dirty, which can modify cacheEntries)
		for (int i = 0; i < entries.size(); i++) {
			SpeedoAccessor sa = (SpeedoAccessor) entries.get(i);
			sa.prepareWrite();
		}

		// close the prefetch buffers associated to the context
		PrefetchCache pc = mapper.getPrefetchCache();
		if (pc != null) {
			pc.invalidatePrefetchBuffer(tx);
		}
		// check if the context can be validate
		boolean valid = cm.validate(tx) & validate;
		if (valid) {
			// flush dirty objects and write them to data support,
			Object da = getDataAccess(tx);
			valid = false;
			dom.flush(da, tx.entries());
			valid = true;
		}
		entries = tx.entries();
		//inform each persitent object that the working set is finished
		for (int i = 0; i < entries.size(); i++) {
			((SpeedoAccessor) entries.get(i)).workingSetClosed();
		}
		return valid;
	}

	/**
	 * commit a context.
	 * @param tx is can be simple working set or a transaction
	 * @param cm is the ConcurrencyManager managing the context
	 * @param newStatus is the new status of the context after the commit
	 * @throws PersistenceException
	 */
	private void commitWS(Transaction tx,
						  ConcurrencyManager cm,
						  byte newStatus) throws PersistenceException {
		try {
			try {
				tx.commitCH();
			} catch (PException e) {
				throw new PersistenceException("Impossible to rollback", e);
			}
			List entries = tx.entries();
			boolean opt = tx.isActive() && tx.getOptimistic();
			// copies accessor states into proxies,
			for (int i = 0; i < entries.size(); i++) {
				SpeedoAccessor sa = (SpeedoAccessor) entries.get(i);
				SpeedoProxy sp = sa.jdoGetSpeedoProxy();
				cache.unfix(sp);
				if (LifeCycle.isDeleted(sa.jdoGetStatus())) {
					replacementManager.unbind(sp);
				}
				if (LifeCycle.isDirty(sa.jdoGetStatus())) {
					sa.jdoChangeStatus(LifeCycle.ACTION_COMMIT);
					if (opt) {
						sp.jdoLoadFieldsFromAccessor(sa);
					}
				}
				if (opt) {
					sp.removeSpeedoAccessor(sa);
				}
			}
		} catch (CacheException e) {
			throw new PersistenceException(e);
		} finally {
			try {
				tx.closeCH();
			} catch (PException e) {
				throw new PersistenceException(
						"Impossible to release a connection", e);
			} finally {
				cm.finalize(tx);
				tx.clear();
				tx.setStatus(newStatus);
			}
		}
	}

	/**
	 * cancel a context.
	 * @param tx is can be simple working set or a transaction
	 * @param cm is the ConcurrencyManager managing the context
	 * @param newStatus is the new status of the context after the commit
	 * @throws PersistenceException
	 */
	private void rollbackWS(Transaction tx,
							ConcurrencyManager cm,
							byte newStatus) throws PersistenceException {
		try {
			try {
				tx.rollbackCH();
			} catch (PException e) {
				throw new PersistenceException("Impossible to rollback - ", e);
			} finally {
				tx.setStatus(newStatus);
			}
			//reload value from the support
			for (Iterator i = tx.entries().iterator(); i.hasNext();) {
				SpeedoAccessor sa = (SpeedoAccessor) i.next();
				SpeedoProxy sp = sa.jdoGetSpeedoProxy();
				cache.unfix(sp);
				if ((sp == sa)
						&& LifeCycle.isDirty(sa.jdoGetStatus())
						&& !LifeCycle.isNew(sa.jdoGetStatus())
				) {
					sm.read(tx, sa.getCeIdentifier(), sa.getCeObject(), tx);
				}
				sa.jdoChangeStatus(LifeCycle.ACTION_ROLLBACK);
			}
			dom.makeClean(tx.entries());
		} catch (CacheException ce) {
			throw new PersistenceException("Impossible to rollback - ", ce);
		} finally {
			try {
				tx.closeCH();
			} catch (PException e) {
				throw new PersistenceException(
						"Impossible to release a connection", e);
			} finally {
				cm.abort(tx);
				tx.clear();
				tx.setStatus(newStatus);
			}
		}
	}

	/**
	 * In case of the optimistic transaction this is the SpeedoAccessor which is
	 * registered into the workging set.
	 * @param cacheEntry
	 * @param workingSet
	 * @return
	 * @throws PersistenceException
	 */
	protected CacheEntry bindCeInWS(CacheEntry cacheEntry, WorkingSet workingSet)
			throws PersistenceException {
		//find sa and sp
		SpeedoProxy sp = null;
		SpeedoAccessor sa = null;
		if (cacheEntry instanceof SpeedoProxy) {
			sp = (SpeedoProxy) cacheEntry;
			sa = sp.getSpeedoAccessor();
		} else if (cacheEntry instanceof SpeedoAccessor) {
			sa = (SpeedoAccessor) cacheEntry;
			sp = sa.jdoGetSpeedoProxy();
		} else {
			logger.log(BasicLevel.DEBUG, "unmanaged entry " + cacheEntry);
			throw new PersistenceException(
					"Unmanaged CacheEntry implementation: " + cacheEntry);
		}
		Transaction tx = (Transaction) workingSet;
		if (tx.lookup(cacheEntry.getCeIdentifier()) == null) {

			Semaphore s = ((ProxyManager) tx.getPersistenceManager()).getSemaphore();
			try {
				s.P();
				if (tx.lookup(cacheEntry.getCeIdentifier()) == null) {
					if (tx.isActive() && tx.getOptimistic()) {
						logger.log(BasicLevel.DEBUG, "Load the value from the proxy into the accessor");
						sp.jdoLoadFieldsInAccessor(sa);
					}
					if (logger.isLoggable(BasicLevel.DEBUG)) {
						logger.log(BasicLevel.DEBUG,
								   "bind an entry " + sa + " into a context: " + tx.entries());
						logger.log(BasicLevel.DEBUG,
								   "contains: " + tx.entries().contains(sa));
					}
					super.bindCeInWS(sa, workingSet);
					//link the transaction to the speedo accessor
					sa.jdoSetTransaction((Transaction) workingSet);
				}
			} finally {
				s.V();
			}
		}
		return sa;
	}

	protected CacheEntry initMemoryInstance(Object o, Object oid, Object context)
			throws PersistenceException {
		SpeedoProxy sp = (SpeedoProxy) o;
		if (!sp.jdoIsActive()) {
			Semaphore s = ((ProxyManager) ((Transaction) context)
					.getPersistenceManager()).getSemaphore();
			try {
				if (!s.P() || !sp.jdoIsActive()) {
					try {
						((ProxyManager) ((Transaction) context).getPersistenceManager())
								.activeProxy(sp, (PName) oid);
						sp.jdoSetStatus(LifeCycle.PERSISTENT_CLEAN);
						sp.bind((PName) oid);
					} catch (Exception e) {
						throw new PersistenceException(
								"Impossible to initialize a new instance", e);
					}
				}
			} finally {
				s.V();
			}
		}
		return sp.getSpeedoAccessor();
	}

	protected Object getDataAccess(Object context) throws PersistenceException {
		return context;
	}

	protected void releaseDataAccess(Object context, Object da) throws PersistenceException {
		try {
			((Transaction) context).releaseCH();
		} catch (PException e) {
			throw new PersistenceException(e);
		}
	}

	protected ConcurrencyManager getConcurrencyManager(WorkingSet workingSet) {
		return ((Transaction) workingSet).getOptimistic() ? ocm : cm;
	}

	// IMPLEMENTATION OF THE TransactionalPersistenceManager INTERFACE //
	//-----------------------------------------------------------------//

	public Object createContext() throws PersistenceException {
		throw new PersistenceException(
				"The context is already created with the jdo PersistenceManager");
	}

	public void begin(Object context) throws PersistenceException {
		Transaction tx = (Transaction) context;
		// 1st step: flush objects in working set, outside transaction
		if (tx.entries().size() > 0) {
			logger.log(BasicLevel.DEBUG, "Flush modification before the transaction " + tx);
			close(context);
		}
		// 2nd step: change the transaction's status
		tx.setStatus(WorkingSetLifeCycle.getNextStatus(
				tx.getStatus(), WorkingSetLifeCycle.BEGIN_TX_ACTION));
	}

	public boolean prepare(Object context, boolean validate) throws PersistenceException {
		Transaction tx = (Transaction) context;
		tx.setStatus(WorkingSetLifeCycle.getNextStatus(
				tx.getStatus(), WorkingSetLifeCycle.PREPARE_TX_ACTION));
		boolean valid = false;
		try {
			valid = prepareWS(tx, getConcurrencyManager(tx), validate);
		} finally {
			if (valid) {
				tx.setStatus(WorkingSetLifeCycle.getNextStatus(
						tx.getStatus(), WorkingSetLifeCycle.PREPARE_OK_TX_ACTION));
			} else {
				tx.setStatus(WorkingSetLifeCycle.getNextStatus(
						tx.getStatus(), WorkingSetLifeCycle.PREPARE_FAIL_TX_ACTION));
			}
		}
		return valid;
	}

	public void commit(Object context) throws PersistenceException {
		Transaction tx = (Transaction) context;
		byte newStatus = WorkingSetLifeCycle.getNextStatus(
				tx.getStatus(), WorkingSetLifeCycle.COMMIT_TX_ACTION);
		commitWS(tx, getConcurrencyManager(tx), newStatus);
	}

	public void rollback(Object context) throws PersistenceException {
		Transaction tx = (Transaction) context;
		byte newStatus = WorkingSetLifeCycle.getNextStatus(
				tx.getStatus(), WorkingSetLifeCycle.ROLLBACK_TX_ACTION);
		rollbackWS(tx, getConcurrencyManager(tx), newStatus);
	}

	public void close(Object context) throws PersistenceException {
		Transaction tx = (Transaction) context;
		byte newStatus = WorkingSetLifeCycle.getNextStatus(
				tx.getStatus(), WorkingSetLifeCycle.CLOSE_ACTION);
		ConcurrencyManager cm = getConcurrencyManager(tx);
		if (prepareWS(tx, cm, true)) {
			commitWS(tx, cm, newStatus);
		} else {
			rollbackWS(tx, cm, newStatus);
		}
	}

	public byte getStatus(Object context) {
		return ((Transaction) context).getStatus();
	}

	protected void unfixEntry(CacheEntry ce) throws PersistenceException {
		try {
			if (ce instanceof SpeedoProxy) {
				cache.unfix(ce);
			} else if (ce instanceof SpeedoAccessor) {
				cache.unfix(((SpeedoAccessor) ce).jdoGetSpeedoProxy());
			} else {
				logger.log(BasicLevel.DEBUG, "unmanaged entry " + ce);
				throw new PersistenceException(
						"Unmanaged CacheEntry implementation: " + ce);
			}
		} catch (CacheException e) {
			throw new PersistenceException("Impossible to unfix - ", e);
		}
	}
}
