/**
 * Copyright (C) 2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.bank.server.jdo;

import org.objectweb.speedo.bank.server.jdo.Account;
import org.objectweb.speedo.bank.server.jdo.Bank;
import org.objectweb.speedo.bank.server.jdo.Client;
import org.objectweb.speedo.bank.server.BankService;
import org.objectweb.speedo.bank.common.AccountInfo;
import org.objectweb.util.monolog.api.BasicLevel;

import javax.jdo.PersistenceManagerFactory;
import javax.jdo.PersistenceManager;
import javax.jdo.Query;
import javax.ejb.SessionContext;
import java.rmi.RemoteException;
import java.util.List;
import java.util.Iterator;
import java.util.ArrayList;
import java.util.Map;
import java.util.HashMap;
import java.util.Collection;

/**
 *
 * @author S.Chassande-Barrioz
 */
public class BankServiceImpl implements BankService {
	private static final String SOCIETE_GENERALE = "SG";

	/**
	 * The JDO driver registered in JNDI
	 */
	public static PersistenceManagerFactory pmf = null;

    protected SessionContext sessionContext = null;


	/**
	 * A shortcut to the unique bank instance.
	 */
	protected transient Bank bank;

    /**
	 * find a bank.
	 * @param pm is the persistencemanager permiting to find the bank
	 * @return the bank instance if it exists, otherwise a null value
	 */
	protected Bank getBank(PersistenceManager pm) {
		if (bank == null) {
			bank = (Bank) pm.getObjectById(
				pm.newObjectIdInstance(Bank.class, SOCIETE_GENERALE), true);
		}
		return bank;
    }

	/**
	 * find a client by its identifier
	 * @param pm is the persistencemanager permiting to find the client
	 * @return the client instance if it exists, otherwise a null value
	 */
	protected Client getClient(long cid, PersistenceManager pm) {
		return (Client) pm.getObjectById(
				pm.newObjectIdInstance(Client.class, "" + cid), false);
	}

	/**
	 * find an account by its number
	 * @param pm is the persistencemanager permiting to find the account
	 * @return the account instance if it exists, otherwise a null value
	 */
	protected Account getAccount(long number, PersistenceManager pm) {
		return (Account) pm.getObjectById(
				pm.newObjectIdInstance(Account.class, "" + number), false);
	}


    // IMPLEMENTATION OF THE BankService INTERFACE //
    //---------------------------------------------//

	public boolean createBank() {
		if (bank == null) {
			PersistenceManager pm = pmf.getPersistenceManager();
			Query q = pm.newQuery(Bank.class);
			try {
				q.setFilter("(name == n)");
				q.declareParameters("String n");
				Collection c = (Collection) q.execute(SOCIETE_GENERALE);
				Iterator it = c.iterator();
				if (it.hasNext()) {
					bank = (Bank) it.next();
					Log.logger.log(BasicLevel.DEBUG, "BS.createBank(): found by a query");
					return false;
				} else {
					bank = new Bank(SOCIETE_GENERALE);
					pm.makePersistent(bank);
					Log.logger.log(BasicLevel.DEBUG, "BS.createBank(): creation of a new instance");
					return true;
				}
			} finally {
				q.closeAll();
				pm.close();
			}
		} else {
			Log.logger.log(BasicLevel.DEBUG, "BS.createBank(): already Loaded");
			return false;
		}
	}

	public long createClient(String lastName,
							 String firstName,
							 String address) throws RemoteException {
		Log.logger.log(BasicLevel.DEBUG, "BS.createClient(" + lastName + ", "
				+ firstName + ", " + address + ")");
		PersistenceManager pm = pmf.getPersistenceManager();
		Log.logger.log(BasicLevel.DEBUG, "In transaction: " + pm.currentTransaction().isActive());
		long id = getBank(pm).createClient(firstName, lastName, address).id;
		pm.close();
		return id;
    }

    public List getClients() throws RemoteException {
        PersistenceManager pm = pmf.getPersistenceManager();
        Bank bank = getBank(pm);
		Iterator it = bank.clients.iterator();
        ArrayList res = new ArrayList(bank.clients.size());
        while (it.hasNext()) {
            res.add(new Long(((Client) it.next()).id));
        }
        pm.close();
        return res;
    }

	public boolean removeClient(long cid) throws RemoteException {
		PersistenceManager pm = pmf.getPersistenceManager();
		Client c = getClient(cid, pm);
		try {
			if (c != null && c.accounts.isEmpty()) {
				pm.deletePersistent(c);
				return true;
			} else {
				return false;
			}
		} finally {
			pm.close();
		}
	}

	public List getClient(String lastName,
						  String firstName,
						  String address) throws RemoteException {
		PersistenceManager pm = pmf.getPersistenceManager();
		String filter = "";
		String param = "";
		Map paramValues = new HashMap(3);
		if (lastName != null) {
			if (paramValues.size() > 0) {
				filter += " && ";
				param += ", ";
			}
			filter += "(lastName == ln)";
			param += "String ln";
			paramValues.put("ln", lastName);
		}
		if (firstName != null) {
			if (paramValues.size() > 0) {
				filter += " && ";
				param += ", ";
			}
			filter += " && (firstName == fn)";
			param += "String fn";
			paramValues.put("fn", firstName);
		}
		if (address != null) {
			if (paramValues.size() > 0) {
				filter += " && ";
				param += ", ";
			}
			filter += " && (address == ad)";
			param += "String ad";
			paramValues.put("ad", address);
		}
		Query q = pm.newQuery(Client.class);
		int nbParam = paramValues.size();
		if (nbParam > 1) {
			filter = '(' + filter + ')';
		}
		Collection c;
		if (nbParam > 0) {
			q.setFilter(filter);
			q.declareParameters(param);
			c = (Collection) q.execute(paramValues);
		} else {
			c = (Collection) q.execute();
		}
		ArrayList cids = new ArrayList();
		Iterator it = c.iterator();
		while(it.hasNext()) {
			cids.add(new Long(((Client) it.next()).id));
		}
		q.closeAll();
		pm.close();
		Log.logger.log(BasicLevel.DEBUG, "BS.getClient("  + lastName + ", "
				+ firstName + ", " + address + "): result=" + cids);
		return cids;
	}

    public long createAccount(long cid) throws RemoteException {
        PersistenceManager pm = pmf.getPersistenceManager();
		Client c = getClient(cid, pm);
        if (c == null) {
            pm.close();
            throw new RemoteException("No client '" + cid + "' found.");
        }

        long number = new Account(c, getBank(pm).getNextAccountNumber()).number;
        pm.close();
        return number;
    }

	public AccountInfo getAccountInfo(long number) throws RemoteException {
        PersistenceManager pm = pmf.getPersistenceManager();
		Account ac = getAccount(number, pm);
		try {
			if (ac == null) {
				return null;
			} else {
				return new AccountInfo(ac.number, ac.value);
			}
		} finally {
			pm.close();
		}
    }

    public List getAccounts(long cid) throws RemoteException {
        PersistenceManager pm = pmf.getPersistenceManager();
        Client c = getClient(cid, pm);
        if (c == null) {
            pm.close();
            throw new RemoteException("No client '" + cid + "' found.");
        }
        Iterator it = c.accounts.iterator();
        ArrayList res = new ArrayList(c.accounts.size());
        while (it.hasNext()) {
            Account ac = (Account) it.next();
            res.add(new AccountInfo(ac.number, ac.value));
        }
        pm.close();
        return res;
    }

    public AccountInfo closeAccount(long number) throws RemoteException {
        PersistenceManager pm = pmf.getPersistenceManager();
		Account ac = getAccount(number, pm);
		try {
			if (ac == null) {
				return null;
			} else {
				pm.deletePersistent(ac);
				AccountInfo ai = new AccountInfo(ac.number, ac.value);
				return ai;
			}
		} finally {
			pm.close();
		}
    }

	public void debit(long accountNumber, double value) throws RemoteException {
		PersistenceManager pm = pmf.getPersistenceManager();
		try {
			Account a = getAccount(accountNumber, pm);
			a.value -= value;
		} finally {
			pm.close();
		}
	}

	public void credit(long accountNumber, double value) throws RemoteException {
		PersistenceManager pm = pmf.getPersistenceManager();
		try {
			Account a = getAccount(accountNumber, pm);
			a.value += value;
		} finally {
			pm.close();
		}
	}

	public void debitCredit(long todebit, long tocredit, double value) throws RemoteException {
		PersistenceManager pm = pmf.getPersistenceManager();
		try {
			Account atd = getAccount(todebit, pm);
			Account atc = getAccount(tocredit, pm);
			atd.value -= value;
			atc.value += value;
		} finally {
			pm.close();
		}
	}
}
