/**
 * Copyright (C) 2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.bank.server.mdb;

import org.objectweb.speedo.bank.server.jdo.BankServiceImpl;
import org.objectweb.speedo.bank.server.jdo.Log;
import org.objectweb.speedo.bank.server.session.BankApplicationHome;
import org.objectweb.speedo.bank.common.BankMessage;
import org.objectweb.speedo.bank.common.BankMessageOperation;
import org.objectweb.util.monolog.api.BasicLevel;

import javax.ejb.MessageDrivenBean;
import javax.ejb.MessageDrivenContext;
import javax.ejb.EJBException;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.jdo.PersistenceManagerFactory;
import javax.jms.Message;
import javax.jms.JMSException;
import javax.jms.MessageListener;
import javax.jms.ObjectMessage;
import javax.rmi.PortableRemoteObject;
import java.rmi.RemoteException;
import java.util.Stack;
import java.util.List;

/**
 *
 * @author S.Chassande-Barrioz
 */
public class GenericMdb
		extends BankServiceImpl
		implements MessageDrivenBean, MessageListener {

	private static final String PMF_CTX_NAME = "java:comp/env/jdo/pmf";
	private static final String SESSION_CTX_NAME = "java:comp/env/ejb/BankApplication";

	protected transient MessageDrivenContext mdbContext;
	protected BankApplicationHome bah;

	public void setMessageDrivenContext(MessageDrivenContext ctx) throws EJBException {
		mdbContext = ctx;
	}

	public GenericMdb() {
	}

	/**
	 * A container invokes this method before it ends the life of the message-driven object.
	 * This happens when a container decides to terminate the message-driven object.
	 *
	 * This method is called with no transaction context.
	 *
	 * @throws javax.ejb.EJBException Thrown by the method to indicate a failure caused by
	 * a system-level error.
	 */
	public void ejbRemove() {
		bah = null;
		mdbContext = null;
	}

	/**
	 * The Message driven  bean must define an ejbCreate methods with no args.
	 *
	 */
	public void ejbCreate() {
	}

	public void init() {
		InitialContext initialContext = null;
		if (pmf == null) {
			synchronized (BankServiceImpl.class) {
				if (pmf == null) {
					try {
						initialContext = new InitialContext();
						pmf = (PersistenceManagerFactory)
							initialContext.lookup(PMF_CTX_NAME);
					} catch (NamingException e) {
						e.printStackTrace();
						throw new EJBException("No PersistenceMangerFactory registered"
							+ " in JNDI with the name '" + PMF_CTX_NAME + "'", e);
					}
				}
			}
		}
		if (bah == null) {
			try {
				if (initialContext == null) {
					initialContext = new InitialContext();
				}
				bah = (BankApplicationHome) PortableRemoteObject.narrow(
						initialContext.lookup(SESSION_CTX_NAME),
									 BankApplicationHome.class);
			} catch (Exception e) {
				e.printStackTrace();
				throw new EJBException("No BankApplicationHome availlable", e);
			}
		}
	}

	/**
	 *  onMessage method
	 */
	public void onMessage(Message message) {
		init();
		if (!(message instanceof ObjectMessage)) {
			Log.logger.log(BasicLevel.ERROR, "Bad message, found:" + message);
		}
		Object o = null;
		try {
			o = ((ObjectMessage)message).getObject();
			Log.logger.log(BasicLevel.ERROR, "Message:" + o);
			treatMessage((BankMessage) o);
		} catch (ClassCastException e) {
			Log.logger.log(BasicLevel.ERROR, "Bad message, found:" + o, e);
		} catch (Exception e) {
			Log.logger.log(BasicLevel.ERROR, "Message treatment failed: ", e);
		}
	}

	protected void treatMessage(BankMessage message)
			throws JMSException, RemoteException {
		Stack cids = new Stack();
		Stack numbers = new Stack();
		for(int i=0; i<message.operations.size(); i++) {
			BankMessageOperation op = (BankMessageOperation)
					message.operations.get(i);
			Log.logger.log(BasicLevel.DEBUG, "operation: " + op);
			long l = -1;
			switch(op.operation) {
			case BankMessage.OP_CREATE_BANK:
				createBank();
				break;

			case BankMessage.OP_CREATE_CLIENT:
				l = createClient(op.lastName, op.firstName, op.address);
				cids.push(new Long(l));
				break;
			case BankMessage.OP_REMOVE_CLIENT:
				l = ((Long) cids.peek()).longValue();
				removeClient(l);
				break;
			case BankMessage.OP_GET_CLIENT:
				List cs = getClient(op.lastName, op.firstName, op.address);
				for(int j=0; j<cs.size(); j++) {
					cids.push(cs.get(j));
				}
				break;
			case BankMessage.OP_CLIENT_POP:
				cids.pop();
				break;

			case BankMessage.OP_CREATE_ACCOUNT:
				l = ((Long) cids.peek()).longValue();
				numbers.push(new Long(createAccount(l)));
				break;
			case BankMessage.OP_REMOVE_ACCOUNT:
				l = ((Long) numbers.peek()).longValue();
				closeAccount(l);
				break;
			case BankMessage.OP_CREDIT_ACCOUNT:
				l = ((Long) numbers.peek()).longValue();
				credit(l, op.value);
				break;
			case BankMessage.OP_DEBIT_ACCOUNT:
				l = ((Long) numbers.peek()).longValue();
				debit(l, op.value);
				break;
			case BankMessage.OP_ACCOUNT_POP:
				numbers.pop();
				break;
			}
		}
	}

}
