/**
 * Copyright (C) 2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.runtime.memload;

import org.objectweb.speedo.SpeedoTestHelper;
import org.objectweb.speedo.runtime.WorkingSetStatistic;
import org.objectweb.speedo.api.ExceptionHelper;
import org.objectweb.speedo.api.SpeedoProperties;
import org.objectweb.speedo.pobjects.userid.IntUserId;
import org.objectweb.util.monolog.api.BasicLevel;

import javax.jdo.PersistenceManager;
import javax.jdo.JDOException;
import javax.jdo.JDOFatalException;
import javax.jdo.Extent;

import java.util.ArrayList;
import java.util.Vector;
import java.util.Iterator;
import java.util.Properties;

/**
 * Stresses the creation load.
 * @author P. Dechamboux
 */
public class TestCreation extends SpeedoTestHelper {
	public static Vector rollbackExceptions = new Vector();
	public static Vector errors = new Vector();


	public TestCreation(String s) {
		super(s);
	}

	public Properties getPMFProperties() {
		Properties p = super.getPMFProperties();
		p.setProperty(SpeedoProperties.MAPPING_STRUCTURE,
				SpeedoProperties.MAPPING_STRUCTURE_DD);
		return p;
	}

	protected String getLoggerName() {
		return LOG_NAME + ".rt.memload.TestCreation";
	}

	/**
	 * Tests the concurrency of several writer on a same persistent object.
	 * (500 users)
	 */
	public void testCreation() {
		final int nbThread = Integer.getInteger(
				"conform.org.objectweb.speedo.rt.memload.TestCreation.testCreation.thread",
				10).intValue();
		final int nbTx = Integer.getInteger(
				"conform.org.objectweb.speedo.rt.memload.TestCreation.testCreation.tx",
				100).intValue();
		final int nbCreation = Integer.getInteger(
				"conform.org.objectweb.speedo.rt.memload.TestCreation.testCreation.creation",
				1000).intValue();
		final int threadTimeout = Integer.getInteger(
				"conform.org.objectweb.speedo.rt.memload.TestCreation.testCreation.timeout",
				2000).intValue();
		logger.log(BasicLevel.INFO, "testCreation: nbThread=" + nbThread + ", nbTx=" + nbTx + ", nbCreation=" + nbCreation + ", thread timeout=" + threadTimeout + "ms");
		rollbackExceptions.clear();
		errors.clear();
		Thread[] ts = new Thread[nbThread];
		final boolean[] txToExecute = new boolean[nbTx];
		final long[] txExecTime = new long[nbTx];
		for (int i = 0; i < nbTx; i++) {
			txToExecute[i] = true;
			txExecTime[i] = -1;
		}
		// creation of data structures
		PersistenceManager pm = pmf.getPersistenceManager();
		pm.getObjectIdClass(IntUserId.class);
		pm.close();

		for (int i = 0; i < nbThread; i++) {
			final int _i = i;
			ts[i] = new Thread(
					new Runnable() {
						public void run() {
							while (true) {
								int currentTx;
								synchronized (txToExecute) {
									for (currentTx = 0; currentTx < nbTx; currentTx++) {
										if (txToExecute[currentTx]) {
											txToExecute[currentTx] = false;
											break;
										}
									}
								}
								if (currentTx == nbTx) {
									break;
								}
								boolean committed = false;
								while (!committed) {
									PersistenceManager pm = pmf.getPersistenceManager();
									logger.log(BasicLevel.DEBUG, _i + " got a PM");
									boolean rollback = false;
									try {
										long bt = System.currentTimeMillis();
										pm.currentTransaction().begin();
										logger.log(BasicLevel.DEBUG, _i + " begin tx");
										for (int no = 0; no < nbCreation; no++) {
											int oid = (currentTx * nbCreation) + no;
											IntUserId iui = new IntUserId(oid, "Obj No " + oid);
											pm.makePersistent(iui);
										}
										pm.currentTransaction().commit();
										txExecTime[currentTx] = System.currentTimeMillis() - bt;
										committed = true;
										logger.log(BasicLevel.DEBUG, _i + " finished");
									} catch (JDOFatalException e) {
										rollbackExceptions.add(e);
										rollback = true;
										logger.log(BasicLevel.INFO, "Tx " + _i + " has been rolledback: try again !", e);
									} catch (Exception e) {
										Exception ie = ExceptionHelper.getNested(e);
										errors.add(ie);
										rollback = true;
										logger.log(BasicLevel.ERROR, _i + " has a problem", ie);
										pm.currentTransaction().rollback();
										break;
									} finally {
										try {
											pm.close();
										} catch (JDOException e) {
											logger.log(BasicLevel.ERROR, "tx " + _i
													+ " has been "
													+ (rollback ? "rolledback" : "committed")
													+ " and the close occurs an error", e);
											throw e;
										}
									}
								}
							}
						}
					}
			);
		}
		long exectime = System.currentTimeMillis();
		for (int i = 0; i < nbThread; i++) {
			logger.log(BasicLevel.DEBUG, "Start " + i);
			ts[i].start();
		}
		try {
			ArrayList al = new ArrayList(nbThread);
			for (int i = 0; i < nbThread; i++) {
				logger.log(BasicLevel.DEBUG, "Join " + i);
				ts[i].join(threadTimeout);
				if (ts[i].isAlive()) {
					al.add(new Integer(i));
					logger.log(BasicLevel.DEBUG, i + " is not finished after" +
							" the delay, it could be blocked");
				}
			}
			exectime = System.currentTimeMillis() - exectime;
			if (al.size() > 0) {
				for (Iterator it = al.iterator(); it.hasNext();) {
					int th = ((Integer) it.next()).intValue();
					if (!ts[th].isAlive()) {
						logger.log(BasicLevel.DEBUG, th + " is late but ok.");
						it.remove();
					}
				}
				if (al.size() > 0) {
					fail("Thread " + al + " blocked!");
				}
			}
		} catch (InterruptedException e) {
			//fail(e.getMessage());
		} finally {

			int nbCommittedTx = 0;
			double averageTxTime = 0;
			for (int i = 0; i < nbTx; i++) {
				if (txExecTime[i] == -1) {
					continue;
				}
				nbCommittedTx++;
				averageTxTime += txExecTime[i];
			}
			averageTxTime /= nbTx;
			logger.log(BasicLevel.INFO, "Total execution time= " + exectime + ", committedTx= " + nbCommittedTx);
			logger.log(BasicLevel.INFO, "Committed transaction rate: "
					+ ((double) nbCommittedTx * 100) / nbTx + "%");
			logger.log(BasicLevel.INFO, "Average execution time per transaction: "
					+ averageTxTime + " ms)");
			logger.log(BasicLevel.INFO, "Transaction throughput: "
					+ ((double) nbCommittedTx / (double) exectime) * 1000 + " tx/sec");
			logger.log(BasicLevel.INFO, "Creation throughput: "
					+ (((double) nbCommittedTx * nbCreation) / (double) exectime) * 1000 + " creation/sec");
			if (WorkingSetStatistic.instance != null) {
				WorkingSetStatistic.instance.printStatistic(logger);
			}
		}
		if (errors.size() > 0) {
			fail("There are " + errors.size() + "/" + nbThread + " errors during the run");
		}
	}
}
