/**
 * Copyright (C) 2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.runtime.relations;

import org.objectweb.speedo.SpeedoTestHelper;
import org.objectweb.speedo.pobjects.relations.C;
import org.objectweb.speedo.pobjects.relations.D;

import javax.jdo.PersistenceManager;
import javax.jdo.Query;
import java.util.Collection;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;

/**
 *
 * @author S.Chassande-Barrioz
 */
public class TestRelationsSeqId extends SpeedoTestHelper {

	public TestRelationsSeqId(String s) {
		super(s);
	}

	protected String getLoggerName() {
		return LOG_NAME + "rt.relations.Relations";
	}

	public void testOneOne() {
        C c1 = new C("testOneOne_c1");
        C c2 = new C("testOneOne_c2");
        D d1 = new D("testOneOne_d1");
        D d2 = new D("testOneOne_d2");

        c1.d = d1;
        assertTrue(d1.c == c1);

        c2.d = d2;
        assertTrue(d2.c == c2);

        c1.d = d2;
        assertTrue(d1.c == null);
        assertTrue(d2.c == c1);
        assertTrue(c2.d == null);

        d1.c = c2;
        assertTrue(c2.d == d1);

        d1.c = c1;
        assertTrue(c1.d == d1);
        assertTrue(d2.c == null);
        assertTrue(c2.d == null);
	}

    public void testOneMany() { // tests on active persistent objects
        C c1 = new C("testOneMany_c1");
        C c2 = new C("testOneMany_c2");

        D d1 = new D("testOneMany_d1");
        D d2 = new D("testOneMany_d2");
        D d3 = new D("testOneMany_d3");
        D d4 = new D("testOneMany_d4");

        PersistenceManager pm = pmf.getPersistenceManager();
        pm.makePersistent(c1);
        pm.makePersistent(c2);
        pm.makePersistent(d1);
        pm.makePersistent(d2);
        pm.makePersistent(d3);
        pm.makePersistent(d4);
        pm.close();

        try {
            pm = pmf.getPersistenceManager();

            c1.nds.add(d1);
            assertTrue(d1.nc == c1);
            c1.nds.add(d2);
            assertTrue(d2.nc == c1);

            c2.nds.add(d3);
            assertTrue(d3.nc == c2);
            c2.nds.add(d4);
            assertTrue(d4.nc == c2);

            c1.nds.add(d3);
            assertTrue(d3.nc == c1);
            assertTrue(!c2.nds.contains(d3));

            c1.nds.remove(d1);
            assertTrue(d1.nc == null);

            d3.nc = c2;
            assertTrue(!c1.nds.contains(d3));
            assertTrue(c2.nds.contains(d3));

            c1.nds = c2.nds;
            // reload objects before testing effects of the above assignement
            pm.close();
            pm = pmf.getPersistenceManager();

            assertTrue(!c1.nds.contains(d1));
            assertTrue(!c1.nds.contains(d2));
            assertTrue(c1.nds.contains(d3));
            assertTrue(c1.nds.contains(d4));

            assertTrue(!c2.nds.contains(d1));
            assertTrue(!c2.nds.contains(d2));
            assertTrue(!c2.nds.contains(d3));
            assertTrue(!c2.nds.contains(d4));

            assertTrue(d1.nc == null);
            assertTrue(d2.nc == null);
            assertTrue(d3.nc == c1);
            assertTrue(d4.nc == c1);

            pm.close();
        } finally {
            pm = pmf.getPersistenceManager();
			pm.deletePersistent(c1);
			pm.deletePersistent(c2);
			pm.deletePersistent(d1);
			pm.deletePersistent(d2);
			pm.deletePersistent(d3);
			pm.deletePersistent(d4);
            pm.close();
        }
    }

    public void testOneMany2() { // tests on non persistent objects
        C c1 = new C("testOneMany2_c1");
        C c2 = new C("testOneMany2_c2");

        D d1 = new D("testOneMany2_d1");
        D d2 = new D("testOneMany2_d2");
        D d3 = new D("testOneMany2_d3");
        D d4 = new D("testOneMany2_d4");

        PersistenceManager pm = pmf.getPersistenceManager();

        d1.nc = c1;
        // tests that c1.nds, initially null, is created and contains d1
        assertTrue(c1.nds != null);
        assertTrue(c1.nds.contains(d1));

        d2.nc = c1;
        assertTrue(c1.nds.contains(d2));

        d3.nc = c2;
        assertTrue(c2.nds != null);
        assertTrue(c2.nds.contains(d3));

        d4.nc = c2;
        assertTrue(c2.nds.contains(d4));

        pm.makePersistent(c1);
        pm.makePersistent(c2);

        pm.close();

        try {
            pm = pmf.getPersistenceManager();

            Collection c = new ArrayList();
            c.addAll(c2.nds);
            c1.nds = c;

            pm.close();

            pm = pmf.getPersistenceManager();

            assertTrue(!c1.nds.contains(d1));
            assertTrue(!c1.nds.contains(d2));
            assertTrue(c1.nds.contains(d3));
            assertTrue(c1.nds.contains(d4));

            assertTrue(c2.nds.size() == 0);

            assertTrue(d1.nc == null);
            assertTrue(d2.nc == null);
            assertTrue(d3.nc == c1);
            assertTrue(d4.nc == c1);

            pm.close();
        } finally {
            pm = pmf.getPersistenceManager();
            pm.deletePersistent(c1);
            pm.deletePersistent(c2);
            pm.deletePersistent(d1);
            pm.deletePersistent(d2);
            pm.deletePersistent(d3);
            pm.deletePersistent(d4);
            pm.close();
        }
    }

    public void testOneMany3() { // tests with manually managed consistency
        C c1 = new C("testOneMany3_c1");

        D d1 = new D("testOneMany3_d1");
        D d2 = new D("testOneMany3_d2");

        PersistenceManager pm = pmf.getPersistenceManager();

        ArrayList l = new ArrayList();
        l.add(d1);
        l.add(d2);
        c1.nds = l;

        d1.nc = c1;
        d2.nc = c1;

        assertTrue(c1.nds.size() == 2);

        pm.makePersistent(c1);
        pm.close();

        try {
            pm = pmf.getPersistenceManager();
            assertTrue(c1.nds.size() == 2);
            pm.close();
        } finally {
            pm = pmf.getPersistenceManager();
            pm.deletePersistent(c1);
            pm.deletePersistent(d1);
            pm.deletePersistent(d2);
            pm.close();
        }
    }

    public void testOneMany4() { // tests with manually managed consistency
		PersistenceManager pm = pmf.getPersistenceManager();
        C c1 = new C("testOneMany4_c1");
        D d1 = new D("testOneMany4_d1");
        ArrayList l = new ArrayList();
        l.add(d1);
        c1.nds = l;
        d1.nc = c1;
        assertTrue(c1.nds.size() == 1);

		pm.currentTransaction().begin();
        pm.makePersistent(c1);
		pm.currentTransaction().commit();
		try {
			assertNotNull("c1 not written", found(C.class, "testOneMany4_c1", pm));
			assertNotNull("d1 not written", found(D.class, "testOneMany4_d1", pm));
		} finally {
			pm.close();
		}

		pm = pmf.getPersistenceManager();
		D d2 = new D("testOneMany4_d2");
		d2.nc = c1;
		pm.close();

		pm = pmf.getPersistenceManager();
		pm.deletePersistent(c1);
		pm.deletePersistent(d1);
		pm.deletePersistent(d2);
		pm.close();
		pm = pmf.getPersistenceManager();
		try {
			assertNull("c1 was not removed", found(C.class, "testOneMany4_c1", pm));
			assertNull("d1 was not removed", found(D.class, "testOneMany4_d1", pm));
			assertNull("d2 was not removed", found(D.class, "testOneMany4_d2", pm));
		} finally {
			pm.close();
		}
	}

    public void testManyMany() { // tests on active persistent objects
        C c1 = new C("testManyMany_c1");
        C c2 = new C("testManyMany_c2");
        C c3 = new C("testManyMany_c3");

        D d1 = new D("testManyMany_d1");
        D d2 = new D("testManyMany_d2");
        D d3 = new D("testManyMany_d3");

        PersistenceManager pm = pmf.getPersistenceManager();
        pm.makePersistent(c1);
        pm.makePersistent(c2);
        pm.makePersistent(c3);
        pm.makePersistent(d1);
        pm.makePersistent(d2);
        pm.makePersistent(d3);
        pm.close();

        try {
            pm = pmf.getPersistenceManager();
            c1.ds.add(d1);
            assertTrue(d1.cs.contains(c1));
            c1.ds.add(d2);
            assertTrue(d2.cs.contains(c1));

            c2.ds.add(d1);
            assertTrue(d1.cs.contains(c2));
            c2.ds.add(d2);
            assertTrue(d2.cs.contains(c2));
            c2.ds.add(d3);
            assertTrue(d3.cs.contains(c2));

            c3.ds.add(d2);
            assertTrue(d2.cs.contains(c3));
            c3.ds.add(d3);
            assertTrue(d3.cs.contains(c3));

            c1.ds.add(d3);
            assertTrue(d3.cs.contains(c1));

            c1.ds.remove(d1);
            assertTrue(!d1.cs.contains(c1));

            d2.cs.remove(c2);
            assertTrue(!c2.ds.contains(d2));

            c1.ds = c2.ds;
            // reload objects before testing effects of the above assignement
            pm.close();
            pm = pmf.getPersistenceManager();
            assertTrue(c1.ds.contains(d1));
            assertTrue(!c1.ds.contains(d2));
            assertTrue(c1.ds.contains(d3));

            assertTrue(c2.ds.contains(d1));
            assertTrue(!c2.ds.contains(d2));
            assertTrue(c2.ds.contains(d3));

            assertTrue(!c3.ds.contains(d1));
            assertTrue(c3.ds.contains(d2));
            assertTrue(c3.ds.contains(d3));

            assertTrue(d1.cs.contains(c1));
            assertTrue(d1.cs.contains(c2));
            assertTrue(!d1.cs.contains(c3));

            assertTrue(!d2.cs.contains(c1));
            assertTrue(!d2.cs.contains(c2));
            assertTrue(d2.cs.contains(c3));

            assertTrue(d3.cs.contains(c1));
            assertTrue(d3.cs.contains(c2));
            assertTrue(d3.cs.contains(c3));

            pm.close();
        } finally {
            pm = pmf.getPersistenceManager();
            pm.deletePersistent(c1);
            pm.deletePersistent(c2);
            pm.deletePersistent(c3);
            pm.deletePersistent(d1);
            pm.deletePersistent(d2);
            pm.deletePersistent(d3);
            pm.close();
        }
    }

    public void testManyMany2() { // tests on non persistent objects
        C c1 = new C("testManyMany2_c1");
        C c2 = new C("testManyMany2_c2");
        C c3 = new C("testManyMany2_c3");

        D d1 = new D("testManyMany2_d1");
        D d2 = new D("testManyMany2_d2");
        D d3 = new D("testManyMany2_d3");

        PersistenceManager pm = pmf.getPersistenceManager();

        c1.ds.add(d1);
        c1.ds.add(d2);

        c2.ds.add(d1);
        c2.ds.add(d2);
        c2.ds.add(d3);

        c3.ds.add(d2);
        c3.ds.add(d3);

        pm.makePersistent(c1);
        pm.makePersistent(c2);
        pm.makePersistent(c3);
        pm.close();

        try {
            pm = pmf.getPersistenceManager();
            assertTrue(d1.cs.contains(c1));
            assertTrue(d2.cs.contains(c1));

            assertTrue(d1.cs.contains(c2));
            assertTrue(d2.cs.contains(c2));
            assertTrue(d3.cs.contains(c2));

            assertTrue(d2.cs.contains(c3));
            assertTrue(d3.cs.contains(c3));

            Collection c = new ArrayList();
            c.addAll(c2.ds);
            c1.ds = c;

            pm.close();

            pm = pmf.getPersistenceManager();
            assertTrue(c1.ds.contains(d1));
            assertTrue(c1.ds.contains(d2));
            assertTrue(c1.ds.contains(d3));

            assertTrue(c2.ds.contains(d1));
            assertTrue(c2.ds.contains(d2));
            assertTrue(c2.ds.contains(d3));

            assertTrue(!c3.ds.contains(d1));
            assertTrue(c3.ds.contains(d2));
            assertTrue(c3.ds.contains(d3));

            assertTrue(d1.cs.contains(c1));
            assertTrue(d1.cs.contains(c2));
            assertTrue(!d1.cs.contains(c3));

            assertTrue(d2.cs.contains(c1));
            assertTrue(d2.cs.contains(c2));
            assertTrue(d2.cs.contains(c3));

            assertTrue(d3.cs.contains(c1));
            assertTrue(d3.cs.contains(c2));
            assertTrue(d3.cs.contains(c3));

            pm.close();
        } finally {
            pm = pmf.getPersistenceManager();
            pm.deletePersistent(c1);
            pm.deletePersistent(c2);
            pm.deletePersistent(c3);
            pm.deletePersistent(d1);
            pm.deletePersistent(d2);
            pm.deletePersistent(d3);
            pm.close();
        }
    }

    public void testManyMany3() { // tests with manually managed consistency
        C c1 = new C("testManyMany3_c1");
        C c2 = new C("testManyMany3_c2");
        C c3 = new C("testManyMany3_c3");

        D d1 = new D("testManyMany3_d1");
        D d2 = new D("testManyMany3_d2");
        D d3 = new D("testManyMany3_d3");

        PersistenceManager pm = pmf.getPersistenceManager();

        c1.ds = Arrays.asList(new Object[] { d1, d2 });
        c2.ds = Arrays.asList(new Object[] { d1, d2, d3 });
        c3.ds = Arrays.asList(new Object[] { d2, d3 });

        d1.cs = Arrays.asList(new Object[] { c1, c2 });
        d2.cs = Arrays.asList(new Object[] { c1, c2, c3 });
        d3.cs = Arrays.asList(new Object[] { c2, c3 });

        pm.makePersistent(c1);
        pm.makePersistent(c2);
        pm.makePersistent(c3);
        pm.close();

        try {
            pm = pmf.getPersistenceManager();
            assertTrue(c1.ds.size() == 2);
            assertTrue(c1.ds.contains(d1));
            assertTrue(c1.ds.contains(d2));

            assertTrue(c2.ds.size() == 3);
            assertTrue(c2.ds.contains(d1));
            assertTrue(c2.ds.contains(d2));
            assertTrue(c2.ds.contains(d3));

            assertTrue(c3.ds.size() == 2);
            assertTrue(c3.ds.contains(d2));
            assertTrue(c3.ds.contains(d3));

            assertTrue(d1.cs.size() == 2);
            assertTrue(d1.cs.contains(c1));
            assertTrue(d1.cs.contains(c2));

            assertTrue(d2.cs.size() == 3);
            assertTrue(d2.cs.contains(c1));
            assertTrue(d2.cs.contains(c2));
            assertTrue(d2.cs.contains(c3));

            assertTrue(d3.cs.size() == 2);
            assertTrue(d3.cs.contains(c2));
            assertTrue(d3.cs.contains(c3));

            pm.close();
        } finally {
            pm = pmf.getPersistenceManager();
			pm.deletePersistent(c1);
			pm.deletePersistent(c2);
			pm.deletePersistent(c3);
			pm.deletePersistent(d1);
			pm.deletePersistent(d2);
			pm.deletePersistent(d3);
            pm.close();
        }
    }

	private Object found(Class clazz, String name, PersistenceManager pm) {
		Query q = pm.newQuery(clazz);
		q.setFilter("(name == p1)");
		q.declareParameters("String p1");
		Collection res = (Collection) q.execute(name);
		Iterator it = res.iterator();
		try {
			if (it.hasNext()) {
				return it.next();
			} else {
				return null;
			}
		} finally {
			q.closeAll();
		}
	}
}
