/*
 * Copyright 2001 Sun Microsystems, Inc. All rights reserved.
 * Use is subject to license terms.
 *
 */


package com.sun.jdotck;

import com.sun.javatest.Test;

import java.io.FileInputStream;
import java.io.InputStream;
import java.io.PrintWriter;
import java.util.Arrays;
import java.util.Properties;

import java.lang.reflect.Field;

import javax.jdo.JDOHelper;
import javax.jdo.PersistenceManagerFactory;
import javax.jdo.PersistenceManager;

public abstract class JDO_Test implements Test {
	public static final int TRANSIENT = 0;
	public static final int PERSISTENT_NEW = 1;
	public static final int PERSISTENT_CLEAN = 2;
	public static final int PERSISTENT_DIRTY = 3;
	public static final int HOLLOW = 4;
	public static final int TRANSIENT_CLEAN = 5;
	public static final int TRANSIENT_DIRTY = 6;
	public static final int PERSISTENT_NEW_DELETED = 7;
	public static final int PERSISTENT_DELETED = 8;
	public static final int PERSISTENT_NONTRANSACTIONAL = 9;
	public static final int NUM_STATES = 10;
	public static final int ILLEGAL_STATE = 10;

	public static final String[] states = {
		"transient",
		"persistent-new",
		"persistent-clean",
		"persistent-dirty",
		"hollow",
		"transient-clean",
		"transient-dirty",
		"persistent-new-deleted",
		"persistent-deleted",
		"persistent-nontransactional",
		"illegal"
	};
	private static final int IS_PERSISTENT = 0;
	private static final int IS_TRANSACTIONAL = 1;
	private static final int IS_DIRTY = 2;
	private static final int IS_NEW = 3;
	private static final int IS_DELETED = 4;
	private static final int NUM_STATUSES = 5;

/*
 * This table indicates the values returned by the status interrogation
 * methods for each state. This is used to determine the current lifecycle
 * state of an object.
 */
	private static final boolean state_statuses[][] = {
//      IS_PERSISTENT   IS_TRANSACTIONAL    IS_DIRTY    IS_NEW      IS_DELETED
// transient
		{false, false, false, false, false},

// persistent-new
		{true, true, true, true, false},

// persistent-clean
		{true, true, false, false, false},

// persistent-dirty
		{true, true, true, false, false},

// hollow
		{true, false, false, false, false},

// transient-clean
		{false, true, false, false, false},

// transient-dirty
		{false, true, true, false, false},

// persistent-new-deleted
		{true, true, true, true, true},

// persistent-deleted
		{true, true, true, false, true},

// persistent-nontransactional
		{true, false, false, false, false}
	};

	private String[] arguments;
	/**
	 *This member is the <code>out</code> third parameter passed to the <code>run</code> method.
	 *It is saved here for use by any methods of any subclasses.
	 */
	protected PrintWriter out;
	/**
	 *This member is the <code>err</code> second parameter passed to the <code>run</code> method.
	 *It is saved here for use by any methods of any subclasses.
	 */
	protected PrintWriter err;

	/**
	 * This member is the number of errors detected.
	 */
	private int errorCount;

	/**
	 * This member is the number of tests detected.
	 */
	private int testCount;

	/**
	 * Name of the file contaninig the properties for the PersistenceManagerFactory.
	 */
	protected static String PMFProperties;

	/**
	 * The Properties object for the PersistenceManagerFactory.
	 */
	protected static Properties PMFPropertiesObject;

	/**
	 * The PersistenceManagerFactory.
	 */
	protected PersistenceManagerFactory pmf;

	/**
	 * The PersistenceManager.
	 */
	protected PersistenceManager pm;

// These values MUST correspond to index in sorted array below.
	private static final int OPTION_APPLICATIONIDENTITY = 0;
	private static final int OPTION_ARRAY = 1;
	private static final int OPTION_ARRAYLIST = 2;
	private static final int OPTION_CHANGEAPPLICATIONIDENTITY = 3;
	private static final int OPTION_DATASTOREIDENTITY = 4;
	private static final int OPTION_HASHMAP = 5;
	private static final int OPTION_HASHTABLE = 6;
	private static final int OPTION_LINKEDLIST = 7;
	private static final int OPTION_LIST = 8;
	private static final int OPTION_MAP = 9;
	private static final int OPTION_NONDURABLEIDENTITY = 10;
	private static final int OPTION_NONTRANSACTIONALREAD = 11;
	private static final int OPTION_NONTRANSACTIONALWRITE = 12;
	private static final int OPTION_NULLCOLLECTION = 13;
	private static final int OPTION_OPTIMISTIC = 14;
	private static final int OPTION_RETAINVALUES = 15;
	private static final int OPTION_TRANSIENTTRANSACTIONAL = 16;
	private static final int OPTION_TREEMAP = 17;
	private static final int OPTION_TREESET = 18;
	private static final int OPTION_VECTOR = 19;

	private static boolean[] options = new boolean[OPTION_VECTOR + 1];
	private static Boolean syncObj = new Boolean(false);
	private static boolean syncFlag = false;

	private static String[] optionNames = {// MUST be in "natural sorted order"
		"-ApplicationIdentity",
		"-Array",
		"-ArrayList",
		"-ChangeApplicationIdentity",
		"-DatastoreIdentity",
		"-HashMap",
		"-Hashtable",
		"-LinkedList",
		"-List",
		"-Map",
		"-NonDurableIdentity",
		"-NontransactionalRead",
		"-NontransactionalWrite",
		"-NullCollection",
		"-Optimistic",
		"-RetainValues",
		"-TransientTransactional",
		"-TreeMap",
		"-TreeSet",
		"-Vector"
	};

	protected JDO_Test() {
	}

	/**
	 * Get the <code>PersistenceManagerFactory</code> instance
	 * for the implementation under test.
	 */
	protected PersistenceManagerFactory getPMF() {
		if (pmf == null) {
			PMFPropertiesObject = loadProperties(PMFProperties); // will exit here if no properties
			pmf = JDOHelper.getPersistenceManagerFactory(PMFPropertiesObject);
		}
		return pmf;
	}

	/**
	 * Get the <code>PersistenceManager</code> instance
	 * for the implementation under test.
	 */
	protected PersistenceManager getPM() {
		if (pm == null) {
			pm = getPMF().getPersistenceManager();
		}
		return pm;
	}

	/**
	 *This method takes the args passed to the test <code>run</code> method,
	 *and initializes some internal data structures to provide method access
	 *to the arguments.
	 */
	protected void initializeArgs(String[] args, PrintWriter err, PrintWriter out) {
		pmf = null;
		pm = null;
		int i;
		arguments = args;
		this.err = err;
		this.out = out;
		errorCount = 0;
		if (syncFlag == true) return; // data has already been initialized
		while (true) {
			synchronized (syncObj) { // only need to set these once in static variables
				if (syncFlag == true) break; // data has been initialized already
				for (i = 0; i < args.length; i += 2) {
					String paramName = args[i];
					String paramValue = args[i + 1];
					if (paramName.equals("-PMFProperties")) { // NOI18N
						PMFProperties = paramValue;
					} else {
						int index = Arrays.binarySearch(optionNames, paramName);
						if (index < 0) {
							out.print("Illegal parameter ");
							out.println(paramName);
						} else {
							options[index] = paramValue.equals("true"); //NOI18N
						}
					}
				}
				syncFlag = true;
			}
			try {
				Thread.sleep(50);
			} catch (Exception e) {
			}
		}
	}

	/** This method cleans up the environment: closes the PersistenceManager.  This
	 * should avoid leaving multiple PersistenceManager instances around, in case
	 * the PersistenceManagerFactory performs PersistenceManager pooling.
	 */

	protected void cleanup() {
		if (pm != null) {
			if (pm.currentTransaction().isActive()) {
				pm.currentTransaction().rollback();
			}
			pm.close();
			pm = null;
		}

		out.flush();
		err.flush();
	}

	/**
	 *This method load Properties from a given file.
	 */
	protected Properties loadProperties(String fileName) {
		if (fileName == null) {
			fileName = System.getProperty("user.home") + "/.jdo/PMFProperties.properties";
		}
		Properties props = new Properties();
		InputStream propStream = null;
		try {
			propStream = new FileInputStream(fileName);
		} catch (java.io.IOException ex) {
			System.out.println("Could not open properties file \"" + fileName + "\"");
			System.out.println("Please specify PMF properties file name" +
					" -PMFProperties argument (defaults to {user.home}/.jdo/PMFProperties.properties)");
			System.exit(1);
		}
		try {
			props.load(propStream);
		} catch (java.io.IOException ex) {
			System.out.println("Error loading properties file \"" + fileName + "\"");
			ex.printStackTrace();
			System.exit(1);
		}
		return props;
	}

	/**
	 *This method is called to increment the error counter by 1.
	 */
	public void incrementErrorCount() {
		errorCount++;
	}

	/**
	 *This method is called to increment the error counter by some number.
	 */
	public void incrementErrorCount(int numberOfErrors) {
		errorCount += numberOfErrors;
	}

	/**
	 *This method is called to get the number of errors that have been detected.
	 */
	public int getErrorCount() {
		return errorCount;
	}

	/**
	 *This method is called to increment the test counter by 1.
	 */
	public void incrementTestCount() {
		testCount++;
	}

	/**
	 *This method is called to increment the test counter by some number.
	 */
	public void incrementTestCount(int numberOfErrors) {
		testCount += numberOfErrors;
	}

	/**
	 *This method is called to get the number of tests that have been detected.
	 */
	public int getTestCount() {
		return testCount;
	}

	/**
	 *Reports whether TransientTransactional is supported.
	 */
	public static boolean isTransientTransactionalSupported() {
		return options[OPTION_TRANSIENTTRANSACTIONAL];
	}

	/**
	 *Reports whether NontransactionalRead is supported.
	 */
	public static boolean isNontransactionalReadSupported() {
		return options[OPTION_NONTRANSACTIONALREAD];
	}

	/**
	 *Reports whether NontransactionalWrite is supported.
	 */
	public static boolean isNontransactionalWriteSupported() {
		return options[OPTION_NONTRANSACTIONALWRITE];
	}

	/**
	 *Reports whether RetainValues is supported.
	 */
	public static boolean isRetainValuesSupported() {
		return options[OPTION_RETAINVALUES];
	}

	/**
	 *Reports whether Optimistic is supported.
	 */
	public static boolean isOptimisticSupported() {
		return options[OPTION_OPTIMISTIC];
	}

	/**
	 *Reports whether Application Identity is supported.
	 */
	public static boolean isApplicationIdentitySupported() {
		return options[OPTION_APPLICATIONIDENTITY];
	}

	/**
	 * Reports whether Changing Application Identity is supported.
	 */
	public static boolean isChangeApplicationIdentitySupported() {
		return options[OPTION_CHANGEAPPLICATIONIDENTITY];
	}

	/**
	 *Reports whether Datastore Identity is supported.
	 */
	public static boolean isDatastoreIdentitySupported() {
		return options[OPTION_DATASTOREIDENTITY];
	}

	/**
	 *Reports whether Non-Durable Identity is supported.
	 */
	public static boolean isNonDurableIdentitySupported() {
		return options[OPTION_NONDURABLEIDENTITY];
	}

	/**
	 *Reports whether an <code>ArrayList</code> collection is supported.
	 */
	public static boolean isArrayListSupported() {
		return options[OPTION_ARRAYLIST];
	}

	/**
	 *Reports whether a <code>HashMap</code> collection is supported.
	 */
	public static boolean isHashMapSupported() {
		return options[OPTION_HASHMAP];
	}

	/**
	 *Reports whether a <code>Hashtable</code> collection is supported.
	 */
	public static boolean isHashtableSupported() {
		return options[OPTION_HASHTABLE];
	}

	/**
	 *Reports whether a <code>LinkedList</code> collection is supported.
	 */
	public static boolean isLinkedListSupported() {
		return options[OPTION_LINKEDLIST];
	}

	/**
	 *Reports whether a <code>TreeMap</code> collection is supported.
	 */
	public static boolean isTreeMapSupported() {
		return options[OPTION_TREEMAP];
	}

	/**
	 *Reports whether a <code>TreeSet</code> collection is supported.
	 */
	public static boolean isTreeSetSupported() {
		return options[OPTION_TREESET];
	}

	/**
	 *Reports whether a <code>Vector</code> collection is supported.
	 */
	public static boolean isVectorSupported() {
		return options[OPTION_VECTOR];
	}

	/**
	 *Reports whether a <code>Map</code> collection is supported.
	 */
	public static boolean isMapSupported() {
		return options[OPTION_MAP];
	}

	/**
	 *Reports whether a <code>List</code> collection is supported.
	 */
	public static boolean isListSupported() {
		return options[OPTION_LIST];
	}

	/**
	 *Reports whether arrays are supported.
	 */
	public static boolean isArraySupported() {
		return options[OPTION_ARRAY];
	}

	/**
	 *Reports whether a null collection is supported.
	 */
	public static boolean isNullCollectionSupported() {
		return options[OPTION_NULLCOLLECTION];
	}

	/**
	 *This utility method returns a <code>String</code> that indicates the current state of an instance.
	 *@param o The object.
	 *@return The current state of the instance, by using the <code>JDOHelper</code> state
	 *interrogation methods.
	 */
	public static String getStateOfInstance(Object o) {
		boolean existingEntries = false;
		StringBuffer buff = new StringBuffer("{");
		if (JDOHelper.isPersistent(o)) {
			buff.append("persistent");
			existingEntries = true;
		}
		if (JDOHelper.isTransactional(o)) {
			if (existingEntries) buff.append(", ");
			buff.append("transactional");
			existingEntries = true;
		}
		if (JDOHelper.isDirty(o)) {
			if (existingEntries) buff.append(", ");
			buff.append("dirty");
			existingEntries = true;
		}
		if (JDOHelper.isNew(o)) {
			if (existingEntries) buff.append(", ");
			buff.append("new");
			existingEntries = true;
		}
		if (JDOHelper.isDeleted(o)) {
			if (existingEntries) buff.append(", ");
			buff.append("deleted");
		}
		buff.append("}");
		return buff.toString();
	}

	/**
	 * This method will return the current lifecycle state of an instance.
	 */
	public static int currentState(Object o) {
		boolean[] status = new boolean[5];
		status[IS_PERSISTENT] = JDOHelper.isPersistent(o);
		status[IS_TRANSACTIONAL] = JDOHelper.isTransactional(o);
		status[IS_DIRTY] = JDOHelper.isDirty(o);
		status[IS_NEW] = JDOHelper.isNew(o);
		status[IS_DELETED] = JDOHelper.isDeleted(o);
		int i, j;
		outerloop:
			for (i = 0; i < NUM_STATES; ++i) {
				for (j = 0; j < NUM_STATUSES; ++j) {
					if (status[j] != state_statuses[i][j])
						continue outerloop;
				}
				return i;
			}
		return NUM_STATES;
	}

	/** This method mangles an object by changing all its public fields
	 */
	protected void mangleObject(Object oid)
			throws Exception {
		Class oidClass = oid.getClass();
		Field[] fields = oidClass.getFields();
		for (int i = 0; i < fields.length; ++i) {
			Field field = fields[i];
			field.setAccessible(true);
			out.println("field" + i + " has name: " + field.getName() + " type: " + field.getType());
			Class fieldType = field.getType();
			if (fieldType == long.class) {
				field.setLong(oid, 10000L);
			}
			if (fieldType == int.class) {
				field.setInt(oid, 10000);
			}
			if (fieldType == short.class) {
				field.setShort(oid, (short) 10000);
			}
			if (fieldType == byte.class) {
				field.setByte(oid, (byte) 100);
			}
			if (fieldType == char.class) {
				field.setChar(oid, '0');
			}
			if (fieldType == String.class) {
				field.set(oid, "This is certainly a challenge");
			}
			if (fieldType == Integer.class) {
				field.set(oid, new Integer(10000));
			}
			if (fieldType == Long.class) {
				field.set(oid, new Long(10000L));
			}
			if (fieldType == Short.class) {
				field.set(oid, new Short((short) 10000));
			}
			if (fieldType == Byte.class) {
				field.set(oid, new Byte((byte) 100));
			}
			if (fieldType == Character.class) {
				field.set(oid, new Character('0'));
			}
		}
	}

}
