/*
 * Copyright 2001 Sun Microsystems, Inc. All rights reserved.
 * Use is subject to license terms.
 *
 */


package com.sun.jdotck.api.instancecallbacks;

import java.io.PrintWriter;

import com.sun.javatest.Test;
import com.sun.javatest.Status;
import com.sun.jdotck.JDO_Test;

import java.io.*;
import javax.jdo.*;

/**
 *<B>JavaTest Title:</B> Calling InstanceCallbacks Postload Method
 *<BR>
 *<B>JavaTest Keywords:</B> instancecallbacks
 *<BR>
 *<B>Assertion ID:</B> A10.1-1.
 *<BR>
 *<B>Assertion Description: </B>
 The <code>jdoPostLoad()</code> method of a class implementing
 <code>InstanceCallbacks</code> is called after the default fetch
 group values are loaded from the <code>StateManager</code> into the instance.

 *@test
 *@title Calling InstanceCallbacks Postload Method
 *@executeClass com.sun.jdotck.api.instancecallbacks.CallingJdoPostload
 *@executeArgs -PMFProperties $PMFProperties
 * -TransientTransactional $TransientTransactional
 * -NontransactionalRead $NontransactionalRead
 * -NontransactionalWrite $NontransactionalWrite
 * -RetainValues $RetainValues
 * -Optimistic $Optimistic
 * -ApplicationIdentity $ApplicationIdentity
 * -ChangeApplicationIdentity $ChangeApplicationIdentity
 * -DatastoreIdentity $DatastoreIdentity
 * -NonDurableIdentity $NonDurableIdentity
 * -ArrayList $ArrayList
 * -HashMap $HashMap
 * -Hashtable $Hashtable
 * -LinkedList $LinkedList
 * -TreeMap $TreeMap
 * -TreeSet $TreeSet
 * -Vector $Vector
 * -Map $Map
 * -List $List
 * -Array $Array
 * -NullCollection $NullCollection
 *@source CallingJdoPostload.java
 *@keywords instancecallbacks
 */


/*
 *  Create a persistent object and commit().
 *
 *  Retrieve object in the next transaction and verify its state is hollow.
 *
 *  Reference a persistent attribute and verify:
 *     the call to jdoPostLoad() occurred just before this reference--and after the object was retrieved (via getObjectById()),
 *     jdoPostLoad() was called exactly once and
 *     once the attribute has been referenced, the state of the object is persistent-clean
 *
 *  Verify that the values of the persistent attributes available in jdoPostLoad() are the values actually in the object (the
 *    values have already been restored be the time jdoPostLoad() is called).
 */

public class CallingJdoPostload extends JDO_Test {

	public CallingJdoPostload() {
	}

	/**
	 *The <code>main</code> is called when the class
	 *is directly executed from the command line.
	 *@param args The arguments passed to the program.
	 */
	public static void main(String[] args) {
		CallingJdoPostload testInstance = new CallingJdoPostload();
		Status s = testInstance.run(args, new PrintWriter(System.err, true), new PrintWriter(System.out, true));
		s.exit();
	}

	/**
	 *The <code>run</code> method is invoked by JavaTest to
	 *execute the test defined by this class.
	 *@param args The arguments provided by the test script,
	 *providing configuration information for a test.
	 *@param err A stream to which to report errors.
	 *@param out A stream to report messages.
	 *@return A <code>Status</code> object representing the outcome of the test.
	 */
	public Status run(String[] args, PrintWriter errPrt, PrintWriter outPrt) {
		try {
			return run2(args, errPrt, outPrt);
		} catch (Exception e) {
			outPrt.println("Assertion A10.1-1 (CallingJdoPostload) unexpected exception");
			outPrt.println(e);
			incrementErrorCount();
			e.printStackTrace(outPrt);
			outPrt.flush();
			return Status.failed("Assertion A10.1-1 (CallingJdoPostload) unexpected exception");
		}
	}

	Status run2(String[] args, PrintWriter errPrt, PrintWriter outPrt) {
		initializeArgs(args, errPrt, outPrt);

		PersistenceManagerFactory fact = getPMF();
		PersistenceManager pm = fact.getPersistenceManager();
		Transaction t = pm.currentTransaction();

		InstanceCallbackNonPersistFdsClass.initializeStaticsForTest();
		t.begin();
		InstanceCallbackNonPersistFdsClass.removeAllInstances(pm); // always start fresh with no instances
		t.commit();

		t.begin();
		// create instance
		InstanceCallbackNonPersistFdsClass obj1 = new InstanceCallbackNonPersistFdsClass(2.2f, 13);
		pm.makePersistent(obj1);
		Object objPtr1 = pm.getObjectId(obj1);
		obj1.setNonPersist(1, (char) 2, 3.3, (short) 4);
		t.commit();

		t.begin();
		// set applicationStep before and after getObjectById() to be sure when jdoPostLoad() gets called.
		InstanceCallbackNonPersistFdsClass.applicationStep = InstanceCallbackNonPersistFdsClass.beforeGetObjectById;
		try {
			obj1 = (InstanceCallbackNonPersistFdsClass) pm.getObjectById(objPtr1, true);  // jdoPreLoad() called
		} catch (JDOUserException e) {
			// could not locate persistent object created in previous transaction
			errPrt.println("JDOUserException " + e + " could not locate obj1.");
			incrementErrorCount();
			return Status.failed("CallingJdoPostload:  Could not locate persistent object created in previous transaction");
		} catch (JDODataStoreException e) {
			// could not locate persistent object created in previous transaction
			errPrt.println("JDODataStoreException " + e + " could not locate obj1.");
			incrementErrorCount();
			return Status.failed("CallingJdoPostload:  Could not locate persistent object created in previous transaction");
		}

		int objectState = currentState(obj1);
		// expect state to be hollow or persistent-clean
		if (objectState != HOLLOW && objectState != PERSISTENT_CLEAN) {
			outPrt.print("State of object was not hollow or persistent clean after accessed via getObjectById((), state is ");
			outPrt.println(states[objectState]);
			incrementErrorCount();
		}

		InstanceCallbackNonPersistFdsClass.applicationStep = InstanceCallbackNonPersistFdsClass.afterGetObjectById;
		int getIntValue = obj1.intValue;  // reference attribute in default fetch group
		objectState = currentState(obj1);

		// expect jdoPostLoad() was called
		if (!InstanceCallbackNonPersistFdsClass.postloadCalled) {
			outPrt.println("jdoPostLoad() never called");
			incrementErrorCount();
		}

		// only check other information if jdoPostLoad() called
		if (InstanceCallbackNonPersistFdsClass.postloadCalled) {
			// expect jdoPostLoad() was called only once
			if (InstanceCallbackNonPersistFdsClass.postloadCalledMultipleTimes) {
				outPrt.println("jdoPostLoad was called more than once for the same loaded object");
				incrementErrorCount();
			}

			// verify the value of intValue was available in jdoPostLoad()
			if (InstanceCallbackNonPersistFdsClass.savedIntValue != getIntValue) {
				outPrt.println("intValue available in jdoPostLoad is incorrect.  It was " + InstanceCallbackNonPersistFdsClass.savedIntValue +
						", it should be " + getIntValue);
				incrementErrorCount();
			}

			// verify the value of floatValue was available in jdoPostLoad()
			if (InstanceCallbackNonPersistFdsClass.savedFloatValue != obj1.floatValue) {
				outPrt.println("floatValue available in jdoPostLoad is incorrect.  It was " + InstanceCallbackNonPersistFdsClass.savedFloatValue +
						", it should be " + obj1.floatValue);
				incrementErrorCount();
			}
		}

		t.rollback();
		pm.close();
		if (getErrorCount() > 0)
			return Status.failed("Assertion A10.1-1 (CallingJdoPostload) failed");
		else
			return Status.passed("Assertion A10.1-1 (CallingJdoPostload) passed");
	}
}
