/*
 * Copyright 2001 Sun Microsystems, Inc. All rights reserved.
 * Use is subject to license terms.
 *
 */


package com.sun.jdotck.api.instancecallbacks;


import com.sun.javatest.Test;
import com.sun.javatest.Status;
import com.sun.jdotck.JDO_Test;

import javax.jdo.*;
import java.io.*;


/**
 *<B>JavaTest Title:</B> Calling InstanceCallbacks Prestore Method
 *<BR>
 *<B>JavaTest Keywords:</B> instancecallbacks
 *<BR>
 *<B>Assertion ID:</B> A10.2-1.
 *<BR>
 *<B>Assertion Description: </B>
 The <code>jdoPreStore()</code> method of a class implementing
 <code>InstanceCallbacks</code> is called before the values are
 stored from the instance to the <code>StateManager</code>.

 *@test
 *@title Calling InstanceCallbacks Prestore Method
 *@executeClass com.sun.jdotck.api.instancecallbacks.CallingJdoPrestore
 *@executeArgs -PMFProperties $PMFProperties
 * -TransientTransactional $TransientTransactional
 * -NontransactionalRead $NontransactionalRead
 * -NontransactionalWrite $NontransactionalWrite
 * -RetainValues $RetainValues
 * -Optimistic $Optimistic
 * -ApplicationIdentity $ApplicationIdentity
 * -ChangeApplicationIdentity $ChangeApplicationIdentity
 * -DatastoreIdentity $DatastoreIdentity
 * -NonDurableIdentity $NonDurableIdentity
 * -ArrayList $ArrayList
 * -HashMap $HashMap
 * -Hashtable $Hashtable
 * -LinkedList $LinkedList
 * -TreeMap $TreeMap
 * -TreeSet $TreeSet
 * -Vector $Vector
 * -Map $Map
 * -List $List
 * -Array $Array
 * -NullCollection $NullCollection
 *@source CallingJdoPrestore.java
 *@keywords instancecallbacks
 */

/*
 *  Use non-managed and transactional values to construct a persistent attribute value to store in database.
 *   Check to see this constructed persistent attribute got stored.
 *
 *  jdoPreStore() should have been called when obj1 and obj2 were created.
 *  jdoPreStore() should have been called again when obj2 was modified.
 *  jdoPreStore() should not be called a second time on obj1 because it was not modified.

 /*
 This test is accomplished in three transactions.
     1)  object1 and object2 are created.  At commit, we expect jdoPreStore()
         to be called on each and calculate values of persistent attributes from non-presistent attributes.
     2)  Retrieve object1 and object2 from the database and verify the persistent attributes
         have values calculated in the jdoPreStore()s called in the previous transaction.
         Change non-persistent fields in each but only change a persistent field in one of them.
     3)  Retrieve object1 and object2 from the database and verify that again jdoPreStore()
         was called for each in the previous transaction.
*/

public class CallingJdoPrestore extends JDO_Test {

	public CallingJdoPrestore() {
	}

	/**
	 *The <code>main</code> is called when the class
	 *is directly executed from the command line.
	 *@param args The arguments passed to the program.
	 */
	public static void main(String[] args) {
		CallingJdoPrestore testInstance = new CallingJdoPrestore();
		Status s = testInstance.run(args, new PrintWriter(System.err, true), new PrintWriter(System.out, true));
		s.exit();
	}

	/**
	 *The <code>run</code> method is invoked by JavaTest to
	 *execute the test defined by this class.
	 *@param args The arguments provided by the test script,
	 *providing configuration information for a test.
	 *@param err A stream to which to report errors.
	 *@param out A stream to report messages.
	 *@return A <code>Status</code> object representing the outcome of the test.
	 */
	public Status run(java.lang.String[] args, java.io.PrintWriter errPrt, java.io.PrintWriter outPrt) {
		try {
			return run2(args, errPrt, outPrt);
		} catch (Exception e) {
			outPrt.println("Assertion A10.2-1 (CallingJdoPrestore) unexpected exception");
			outPrt.println(e);
			incrementErrorCount();
			e.printStackTrace(outPrt);
			return Status.failed("Assertion A10.2-1 (CallingJdoPrestore) unexpected exception");
		}
	}

	Status run2(String[] args, PrintWriter errPrt, PrintWriter outPrt) {
		initializeArgs(args, errPrt, outPrt);

		PersistenceManagerFactory fact = getPMF();
		PersistenceManager pm = fact.getPersistenceManager();
		Transaction t = pm.currentTransaction();

		InstanceCallbackClass.initializeStaticsForTest();
		t.begin();
		InstanceCallbackNonPersistFdsClass.removeAllInstances(pm); // always start fresh with no instances
		t.commit();

		InstanceCallbackClass.performPreStoreTests = true;
		t.begin();
		int expectedIntValue1, expectedIntValue2;
		float expectedFloatValue1, expectedFloatValue2;

		int origIntValue1 = 10, origIntValue2 = 12;
		float origFloatValue1 = 2.5f, origFloatValue2 = -3.0f;

		// create two instances

		InstanceCallbackNonPersistFdsClass obj1 = new InstanceCallbackNonPersistFdsClass(origFloatValue1, origIntValue1);
		pm.makePersistent(obj1);
		Object objPtr1 = pm.getObjectId(obj1);
		obj1.setNonPersist(5, (char) 10, .25, (short) 100);
		expectedIntValue1 = obj1.calcIntValue();
		expectedFloatValue1 = obj1.calcFloatValue();

		InstanceCallbackNonPersistFdsClass obj2 = new InstanceCallbackNonPersistFdsClass(origFloatValue2, origIntValue2);
		pm.makePersistent(obj2);
		Object objPtr2 = pm.getObjectId(obj2);
		obj2.setNonPersist(2, (char) 125, .1, (short) 750);
		expectedIntValue2 = obj2.calcIntValue();
		expectedFloatValue2 = obj2.calcFloatValue();

		t.commit();

		t.begin();
		// read both objects back in to determine persistent state is as expected.
		try {
			obj1 = (InstanceCallbackNonPersistFdsClass) pm.getObjectById(objPtr1, true);
			checkValues("Object1 created and read back in:  ", origIntValue1, expectedIntValue1, obj1.intValue);
			checkValues("Object1 created and read back in:  ", origFloatValue1, expectedFloatValue1, obj1.floatValue);
			origIntValue1 = obj1.intValue;
			origFloatValue1 = obj1.floatValue;
			obj1.setNonManaged(-1, (char) 62);
			expectedIntValue1 = obj1.calcIntValue();
			expectedFloatValue1 = obj1.calcFloatValue();
		} catch (JDOUserException e) {
			// could not locate persistent object created in previous transaction
			errPrt.println("JDOUserException " + e + " could not reference obj1.");
			incrementErrorCount();
			return Status.failed("CallingJdoPrestore: Failed to find object obj1 created in previous transaction.");
		} catch (JDODataStoreException e) {
			// could not locate persistent object created in previous transaction
			errPrt.println("JDODataStoreException " + e + " could not reference obj1.");
			incrementErrorCount();
			return Status.failed("CallingJdoPrestore: Failed to find object obj1 created in previous transaction.");
		}

		try {
			obj2 = (InstanceCallbackNonPersistFdsClass) pm.getObjectById(objPtr2, true);
			checkValues("Object2 created and read back in:  ", origIntValue2, expectedIntValue2, obj2.intValue);
			checkValues("Object2 created and read back in:  ", origFloatValue2, expectedFloatValue2, obj2.floatValue);
			origIntValue2 = obj2.intValue;
			origFloatValue2 = obj2.floatValue;
			obj2.setNonPersist(12, (char) 30, 5.0, (short) 137);
			expectedIntValue2 = obj2.calcIntValue();
			expectedFloatValue2 = obj2.calcFloatValue();

			// change obj2 making it persistent dirty (leave obj1 persistent clean).
			obj2.incrementIntValue();
		} catch (JDOUserException e) {
			// could not locate persistent object created in previous transaction
			errPrt.println("JDOUserException " + e + " could not reference obj2.");
			incrementErrorCount();
			return Status.failed("CallingJdoPrestore: Failed to find object obj2 created in previous transaction.");
		} catch (JDODataStoreException e) {
			// could not locate persistent object created in previous transaction
			errPrt.println("JDODataStoreException " + e + " could not reference obj2.");
			incrementErrorCount();
			return Status.failed("CallingJdoPrestore: Failed to find object obj2 created in previous transaction.");
		}

		t.commit();

		t.begin();
		// read both objects back in to determine persistent state is as expected.
		try {
			// obj1 not made dirty, so jdoPreStore() should not have been called.
			obj1 = (InstanceCallbackNonPersistFdsClass) pm.getObjectById(objPtr1, true);
			checkValues("Object1, did not expect jdoPreStore to be called:  ", origIntValue1, origIntValue1, obj1.intValue);
			checkValues("Object1, did not expect jdoPreStore to be called:  ", origFloatValue1, origFloatValue1, obj1.floatValue);
		} catch (JDOUserException e) {
			// could not locate persistent object created in previous transaction
			errPrt.println("JDOUserException " + e + " could not reference obj1.");
			incrementErrorCount();
			return Status.failed("CallingJdoPrestore: Failed to find object obj1 created in previous transaction.");
		} catch (JDODataStoreException e) {
			// could not locate persistent object created in previous transaction
			errPrt.println("JDODataStoreException " + e + " could not reference obj1.");
			incrementErrorCount();
			return Status.failed("CallingJdoPrestore: Failed to find object obj1 created in previous transaction.");
		}

		try {
			obj2 = (InstanceCallbackNonPersistFdsClass) pm.getObjectById(objPtr2, true);
			checkValues("Object2, expected jdoPreStore() to be called:  ", origIntValue2, expectedIntValue2, obj2.intValue);
			checkValues("Object2, expected jdoPreStore() to be called:  ", origFloatValue2, expectedFloatValue2, obj2.floatValue);
		} catch (JDOUserException e) {
			// could not locate persistent object created in previous transaction
			errPrt.println("JDOUserException " + e + " could not reference obj2.");
			incrementErrorCount();
			return Status.failed("CallingJdoPrestore: Failed to find object obj2 created in previous transaction.");
		} catch (JDODataStoreException e) {
			// could not locate persistent object created in previous transaction
			errPrt.println("JDODataStoreException " + e + " could not reference obj2.");
			incrementErrorCount();
			return Status.failed("CallingJdoPrestore: Failed to find object obj2 created in previous transaction.");
		}
		t.rollback();
		pm.close();
		if (getErrorCount() > 0)
			return Status.failed("Assertion A10.2-1 (CallingJdoPrestore) failed");
		else
			return Status.passed("Assertion A10.2-1 (CallingJdoPrestore) passed");
	}

	void checkValues(String title, int origIntValue, int expectedIntValue, int intValue) {
		if (expectedIntValue != intValue) {
			out.println(title + "Original persistent intValue is " + origIntValue + ", expected " + expectedIntValue +
					".  Instead intValue is " + intValue);
			incrementErrorCount();
		}
	}

	void checkValues(String title, float origFloatValue, float expectedFloatValue, float floatValue) {
		if (expectedFloatValue != floatValue) {
			out.println(title + "Original persistent floatValue is " + origFloatValue + ", expected " + expectedFloatValue +
					".  Instead floatVal is " + floatValue);
			incrementErrorCount();
		}
	}
}
