/*
 * Copyright 2001 Sun Microsystems, Inc. All rights reserved.
 * Use is subject to license terms.
 *
 */


package com.sun.jdotck.api.persistencemanager;

import java.io.*;

import com.sun.javatest.Test;
import com.sun.javatest.Status;
import com.sun.jdotck.JDO_Test;
import com.sun.jdotck.models.mylib.*;

import java.util.*;

import javax.jdo.*;


/**
 *<B>JavaTest Title:</B> EvictingCollectionOfInstancesSideEffects
 *<BR>
 *<B>JavaTest Keywords:</B> cache
 *<BR>
 *<B>Assertion IDs:</B> A12.5.1-4
 *<BR>
 *<B>Assertion Description: </B>
 If PersistenceManager.evict is called with no parameter, then all referenced instances are evicted. For each instance evicted, it:
 *<UL>
 *<LI> calls the jdoPreClearmethod on each instance, if the class of the instance implements InstanceCallbacks </LI>
 *<LI> clears persistent fields on each instance after the call to jdoPreClear()</LI>
 *<LI> changes the state of instances to hollow or persistent-nontransactional (cannot distinguish between these two states) this is not directly testable. </LI>
 *</UL>

 *@test
 *@title Delete Persistent
 *@executeClass com.sun.jdotck.api.persistencemanager.EvictingCollectionOfInstancesSideEffects
 *@executeArgs -PMFProperties $PMFProperties
 * -TransientTransactional $TransientTransactional
 * -NontransactionalRead $NontransactionalRead
 * -NontransactionalWrite $NontransactionalWrite
 * -RetainValues $RetainValues
 * -Optimistic $Optimistic
 * -ApplicationIdentity $ApplicationIdentity
 * -DatastoreIdentity $DatastoreIdentity
 * -NonDurableIdentity $NonDurableIdentity
 * -ArrayList $ArrayList
 * -HashMap $HashMap
 * -Hashtable $Hashtable
 * -LinkedList $LinkedList
 * -TreeMap $TreeMap
 * -TreeSet $TreeSet
 * -Vector $Vector
 * -Map $Map
 * -List $List
 * -Array $Array
 * -NullCollection $NullCollection

 *@source EvictingCollectionOfInstancesSideEffects.java
 *@keywords
 */

public class EvictingCollectionOfInstancesSideEffects extends PersistenceManagerTest {
	private PersistenceManagerFactory pmf;
	private PersistenceManager pm;
	private Transaction tx;

	PCPoint2 pnt1 = null;
	PCPoint2 pnt2 = null;
	PCPoint2 p1 = null;
	PCPoint2 p2 = null;


	/* standalone interface */
	public static void main(String[] args) throws Exception {

		EvictingCollectionOfInstancesSideEffects t = new EvictingCollectionOfInstancesSideEffects();

		Status s = t.run(args, new PrintWriter(System.err),
				new PrintWriter(System.out));
		s.exit();
	}

	public Status runTest(String[] args, PrintWriter err, PrintWriter out) {

		initializeArgs(args, err, out);
		PersistenceManagerFactory pmf = getPMF();
		pm = pmf.getPersistenceManager();

		testEvictingCollectionOfInstancesSideEffects();

		pm.close();

		int errCnt = getErrorCount();
		if (errCnt == 0) {
			return Status.passed("EvictingCollectionOfInstancesSideEffects tests passed");
		} else {
			return Status.failed("EvictingCollectionOfInstancesSideEffects: number of errors detected is " + errCnt);
		}

	}

	public void testEvictingCollectionOfInstancesSideEffects() {
		createObjects();
		testEvictAllCollection();
		testEvictAllArray();

	}


	private void createObjects() {
		try {
			tx = pm.currentTransaction();
			tx.begin();
			pnt1 = new PCPoint2(1, 2);
			pnt2 = new PCPoint2(2, 3);
			p1 = new PCPoint2(3, 1);
			p2 = new PCPoint2(4, 2);
			pm.makePersistent(pnt1);
			pm.makePersistent(pnt2);
			pm.makePersistent(p1);
			pm.makePersistent(p2);
			tx.commit();

			/* P-nontransactional instance*/
			tx.setOptimistic(true);
			tx.begin();
			pnt1.getX();
			pnt2.getX();
			tx.commit();

			/* P-clean instance */
			tx.setOptimistic(false);
			tx.begin();
			p1.getX();
			p2.getX();
			tx.commit();

		} catch (Exception ex) {
			out.println(" - createObjects() FAILED");
		}
	}


	public void testEvictAllCollection() {
		out.println("**  in testEvictAllCollection() ");
		try {
			tx = pm.currentTransaction();
			tx.begin();

			Collection col1 = new java.util.HashSet();
			col1.add(pnt1);
			col1.add(p1);

			pm.evictAll(col1);

			if (!p1.wasClearCalled()) {
				out.println(" p1 jdoPreClear not called ");
				incrementErrorCount();
			}
			if (!pnt1.wasClearCalled()) {
				out.println(" pnt1 jdoPreClear not called ");
				incrementErrorCount();
			}

			if (testState(p1, HOLLOW, "hollow") ||
					testState(p1, PERSISTENT_NONTRANSACTIONAL, "persistent_nontransactional")) {
				; // expected result
			} else {
				incrementErrorCount();
			}

			if (testState(pnt1, HOLLOW, "hollow") ||
					testState(pnt1, PERSISTENT_NONTRANSACTIONAL, "persistent_nontransactional")) {
				; // expected result
			} else {
				incrementErrorCount();
			}

			tx.commit();
			out.println(" \nPASSED in testEvictAllCollection()");
		} catch (Exception ex) {
			assertUnexpectedException(ex, "testEvictAllCollection()");
			tx.rollback();
		}

	}


	public void testEvictAllArray() {
		out.println("**  in testEvictAllArray() ");
		try {
			tx = pm.currentTransaction();
			tx.begin();

			Collection col1 = new java.util.HashSet();
			col1.add(pnt2);
			col1.add(p2);

			pm.evictAll(col1.toArray());

			if (!p2.wasClearCalled()) {
				out.println(" p2 jdoPreClear not called ");
				incrementErrorCount();
			}
			if (!pnt2.wasClearCalled()) {
				out.println(" pnt2 jdoPreClear not called ");
				incrementErrorCount();
			}

			if (testState(p2, HOLLOW, "hollow") ||
					testState(p2, PERSISTENT_NONTRANSACTIONAL, "persistent_nontransactional")) {
				; // expected result
			} else {
				incrementErrorCount();
			}

			if (testState(pnt2, HOLLOW, "hollow") ||
					testState(pnt2, PERSISTENT_NONTRANSACTIONAL, "persistent_nontransactional")) {
				; // expected result
			} else {
				incrementErrorCount();
			}

			tx.commit();
			out.println(" \nPASSED in testEvictAllArray()");
		} catch (Exception ex) {
			assertUnexpectedException(ex, "testEvictAllArray()");
			tx.rollback();
		}

	}

	/***
	 public boolean testState (PCPoint obj) {

	 int curr = currentState(obj);
	 if ( curr == HOLLOW ||
	 curr == PERSISTENT_NONTRANSACTIONAL) {
	 return true;
	 } else {
	 out.println (" obj is not hollow or P_nontransactional" +
	 "x= " + obj.getX() + " y= " + obj.getY());
	 return false;
	 }

	 }

	 ***/

}
