/*
 * Copyright 2001 Sun Microsystems, Inc. All rights reserved.
 * Use is subject to license terms.
 *
 */

package com.sun.jdotck.extents;

import com.sun.javatest.Test;
import com.sun.javatest.Status;
import com.sun.jdotck.JDO_Test;
import com.sun.jdotck.models.company.*;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.Collection;
import java.io.PrintWriter;

import java.util.Date;
import java.util.GregorianCalendar;
import java.util.TimeZone;
import java.util.HashSet;
import java.math.BigDecimal;

import java.lang.reflect.Constructor;

import javax.jdo.*;

/**
 *
 * @author  Craig Russell
 * @version 1.0
 */
abstract class ExtentTest extends JDO_Test {

	protected Class extentClass = Employee.class;
	protected Company company;
	protected Object companyOID;

	/** Creates new ExtentTest */
	public ExtentTest() {
		if (PMFProperties == null) {
			PMFProperties = System.getProperty("user.home") + "/.jdo/PMFProperties.properties";
			if (PMFProperties == null)
				System.out.println("Please specify PMF properties in {user.home}/.jdo/PMFProperties.properties");
		}
	}

	protected void checkPM() {
		try {
			Extent ex = getPM().getExtent(Company.class, false);
			int count = countIterator(ex.iterator());
			if (count == 1) {
				out.println("Found company");
				return;
			}
		} catch (Exception e) {
			e.printStackTrace(System.err);
			System.err.println("Failed to find Company; initializing DB");
		}
		initDB();
	}

	protected void initDB() {
		HashSet h;
		Address addr1 = new Address(7001L, "18 Network Circle",
				"Santa Clara", "CA", "94455");
		company = new Company("Sun Microsystems", new Date(), addr1);
		GregorianCalendar born =
				new GregorianCalendar(TimeZone.getTimeZone("America/Los_Angeles"));
		GregorianCalendar hired =
				new GregorianCalendar(TimeZone.getTimeZone("America/Los_Angeles"));

		born.set(1969, 7, 20);
		hired.set(1982, 5, 5);
		Address addr2 = new Address(7002L, "43 Sematery Drive",
				"Woodside", "CA", "94320");
		Employee scott =
				new FullTimeEmployee(3001L,
						"Scott", "McNealy", addr2,
						born.getTime(), hired.getTime(),
						200000.0);
		born.set(1960, 4, 8);
		hired.set(1985, 2, 3);
		Address addr3 = new Address(7003L, "1298 Wanderlust Road",
				"Pleasanton", "CA", "95560");
		Employee ed =
				new PartTimeEmployee(3002L,
						"Ed", "Zander", addr3,
						born.getTime(), hired.getTime(),
						400000.0);
		scott.addToTeam(ed);

		// Set up their departments.
		Department board =
				new Department(2001L, "board", company);
		h = new HashSet();
		h.add(scott);
		board.setEmployees(h);
		scott.setDepartment(board);

		company.addDepartment(board);

		Department emg =
				new Department(2002L, "emg", company);
		h = new HashSet();
		h.add(ed);
		emg.setEmployees(h);
		ed.setDepartment(emg);

		company.addDepartment(emg);

		// Insure these guys
		Insurance scottIns = new Insurance(5001L, "Aetna", scott);
		Insurance edIns = new Insurance(5002L, "BlueCross", ed);
		scott.setInsurance(scottIns);
		ed.setInsurance(edIns);

		// Give them some projects to work on.  Scott works on both; Ed only
		// on one.
		Project solaris = new Project(4001L, "Solaris", new BigDecimal(100.375));
		Project sparc = new Project(4002L, "Sparc", new BigDecimal(200.500));
		h = new HashSet();
		h.add(scott);
		h.add(ed);
		solaris.setMembers(h); // Solaris is worked on by Scott and Ed

		h = new HashSet();
		h.add(scott);
		sparc.setMembers(h); // Sparc is worked on by Scott

		h = new HashSet();
		h.add(solaris);
		h.add(sparc);
		scott.setProjects(h); // Scott works on Solaris and Sparc

		h = new HashSet();
		h.add(solaris);
		ed.setProjects(h); // Ed works on Solaris

		/* Now put all of these into the database
		 */
		pm.currentTransaction().begin();
		pm.makePersistent(company);
		pm.currentTransaction().commit();
		// System.out.println ("Company OID: " + pm.getObjectId(company));
	}

	protected Employee addEmployee() {
		Address addr1 = new Address(7004L, "456 Chelsey Lane", "Mountain View", "CA", "94040");
		Employee emp1 = new Employee(3003L, "First3003", "Last3003", addr1, new Date(), new Date());
		getPM().makePersistent(emp1);
		return emp1;
	}

	protected void deleteEmployee(Employee e) {
		getPM().deletePersistent(e);
	}

	protected int countIterator(Iterator it) {
		int count = 0;
		for (; it.hasNext(); count++, it.next()) ;
		return count;
	}

	protected int printIterator(Iterator it) {
		int count = 0;
		for (; it.hasNext(); count++) {
			System.out.println(it.next().toString());
		}
		return count;
	}

	protected Extent getExtent() {
		return getPM().getExtent(extentClass, true);
	}

	protected Object getCompanyOID() {
		String companyOIDString = (String)
				PMFPropertiesObject.get("com.sun.jdotck.extents.CompanyOID");
		String companyOIDClassName = (String)
				PMFPropertiesObject.get("com.sun.jdotck.extents.CompanyOIDClass");
		if (companyOIDClassName == null) {
			companyOIDClassName = "com.sun.jdori.fostore.OID";
		}
		try {
			Class companyOIDClass = Class.forName(companyOIDClassName);
			Constructor companyOIDConstructor = companyOIDClass.getConstructor(new Class[]{String.class});
			Object companyOID = companyOIDConstructor.newInstance(new Object[]{companyOIDString});
			return companyOID;
		} catch (Exception ex) {
			throw new JDOFatalInternalException("PMFProperties must be configured with the following properties\n" +
					"\tcom.sun.jdotck.extents.CompanyOID = <string result of oid.toString()>\n" +
					"\tcom.sun.jdotck.extents.CompanyOIDClass = <name of companyOID class>\n", ex);
		}
	}

	protected PersistenceManager getPM() {
		if (pm == null) {
			pm = getPMF().getPersistenceManager();
			checkPM();
		}
		return pm;
	}

	public PersistenceManagerFactory getPMF() {
		if (pmf == null) {
			pmf = super.getPMF();
		}
		return pmf;
	}

	protected void beginTransaction() {
		getPM().currentTransaction().begin();
	}

	protected void commitTransaction() {
		getPM().currentTransaction().commit();
	}

	protected void rollbackTransaction() {
		getPM().currentTransaction().rollback();
	}
}
