/*
 * Copyright 2001 Sun Microsystems, Inc. All rights reserved.
 * Use is subject to license terms.
 *
 */

package com.sun.jdotck.lifecycle;

import java.io.PrintWriter;

import com.sun.javatest.Test;
import com.sun.javatest.Status;
import com.sun.jdotck.JDO_Test;

import java.util.Hashtable;
import java.util.Iterator;
import javax.jdo.*;

/**
 *<B>JavaTest Title:</B> Test State Transitions
 *<BR>
 *<B>JavaTest Keywords:</B> lifecycle
 *<BR>
 *<B>Assertion IDs:</B> A5.9-1 through A5.9-170
 A5.5.1-7, A5.5.2-2, A5.5.3-1, A5.5.3-2, A5.5.3-3, A5.5.3-4, A5.5.3-6
 A5.5.4-4, A5.5.4-5, A5.5.4-6, A5.5.5-2, A5.5.5-3, A5.5.6-1, A5.5.6-4
 A5.5.7-1, A5.5.7-3, A5.5.7-4, A5.6-8, A5.6-13, A5.7-1, A5.7.1-2
 A5.7.2-1, A5.7.2-3, A5.8-1, A5.8-3, A5.8-5, A5.8-7

 *@test
 *@title Test State Transitions
 *@executeClass com.sun.jdotck.lifecycle.StateTransitions
 *@executeArgs -PMFProperties $PMFProperties
 * -TransientTransactional $TransientTransactional
 * -NontransactionalRead $NontransactionalRead
 * -NontransactionalWrite $NontransactionalWrite
 * -RetainValues $RetainValues
 * -Optimistic $Optimistic
 * -ApplicationIdentity $ApplicationIdentity
 * -DatastoreIdentity $DatastoreIdentity
 * -NonDurableIdentity $NonDurableIdentity
 * -ArrayList $ArrayList
 * -HashMap $HashMap
 * -Hashtable $Hashtable
 * -LinkedList $LinkedList
 * -TreeMap $TreeMap
 * -TreeSet $TreeSet
 * -Vector $Vector
 * -Map $Map
 * -List $List
 * -Array $Array
 * -NullCollection $NullCollection
 *@source StateTransitions.java
 *@keywords lifecycle
 */

public class StateTransitions extends JDO_Test {
	private PersistenceManagerFactory pmf;
	private PersistenceManager pm;
	private Transaction transaction;
	private int scenario;
	private int operation;
	private int current_state;
	private int expected_state;
	private int new_state;
	private String pmfName;
	private String contextFactory;
	private String providerURL;
	private String connectionUserName;
	private String connectionPassword;
	private String connectionURL;

	private static final String PMFNAME = "pmfName";
	private static final String JNDICONTEXTFACTORY = "PMFProperties";

	/**
	 * Operations that cause state changes
	 */
	private static final int MAKEPERSISTENT = 0;
	private static final int DELETEPERSISTENT = 1;
	private static final int MAKETRANSACTIONAL = 2;
	private static final int MAKENONTRANSACTIONAL = 3;
	private static final int MAKETRANSIENT = 4;
	private static final int COMMITNORETAINVALUES = 5;
	private static final int COMMITRETAINVALUES = 6;
	private static final int ROLLBACKNORETAINVALUES = 7;
	private static final int ROLLBACKRETAINVALUES = 8;
	private static final int REFRESHDATASTORE = 9;
	private static final int REFRESHOPTIMISTIC = 10;
	private static final int EVICT = 11;
	private static final int READOUTSIDETX = 12;
	private static final int READOPTIMISTIC = 13;
	private static final int READDATASTORE = 14;
	private static final int WRITEOUTSIDETX = 15;
	private static final int WRITEINSIDETX = 16;

	private static final String[] operations = {
		"makePersistent",
		"deletePersistent",
		"makeTransactional",
		"makeNontransactional",
		"makeTransient",
		"commit, retainValues=false",
		"commit, retainValues=true",
		"rollback, retainValues=false",
		"rollback, retainValues=true",
		"refresh with active datastore tx",
		"refresh with active optimistic tx",
		"evict",
		"read field outside tx",
		"read field with active optimistic tx",
		"read field with active datastore tx",
		"write field outside tx",
		"write field with active tx"
	};
	private static final int NUM_OPERATIONS = 17;

	private static final boolean[] closes_transaction =
			{false, false, false, false, false, true, true, true, true, false,
			 false, false, false, false, false, false, false};

	/**
	 * Illegal state transitions
	 */
	private static final int UNCHANGED = -1;
	private static final int ERROR = -2;
	private static final int IMPOSSIBLE = -3;
	private static final int NOT_APPLICABLE = -4;

	/**
	 * State transitions
	 */

	public static final int[][] transitions = {// [operation] [current state] = new state
// makePersistent
		{PERSISTENT_NEW, UNCHANGED, UNCHANGED, UNCHANGED, UNCHANGED, PERSISTENT_NEW,
		 PERSISTENT_NEW, UNCHANGED, UNCHANGED, UNCHANGED},

// deletePersistent
		{ERROR, PERSISTENT_NEW_DELETED, PERSISTENT_DELETED, PERSISTENT_DELETED,
		 PERSISTENT_DELETED, ERROR, ERROR, UNCHANGED, UNCHANGED, PERSISTENT_DELETED},

// makeTransactional
		{TRANSIENT_CLEAN, UNCHANGED, UNCHANGED, UNCHANGED, PERSISTENT_CLEAN,
		 UNCHANGED, UNCHANGED, UNCHANGED, UNCHANGED, PERSISTENT_CLEAN},

// makeNontransactional
		{ERROR, ERROR, PERSISTENT_NONTRANSACTIONAL, ERROR, UNCHANGED,
		 TRANSIENT, ERROR, ERROR, ERROR, UNCHANGED},

// makeTransient
		{UNCHANGED, ERROR, TRANSIENT, ERROR, TRANSIENT,
		 UNCHANGED, UNCHANGED, ERROR, ERROR, TRANSIENT},

// commit, retainValues = false
		{UNCHANGED, HOLLOW, HOLLOW, HOLLOW, UNCHANGED, UNCHANGED,
		 TRANSIENT_CLEAN, TRANSIENT, TRANSIENT, UNCHANGED},

// commit, retainValues = true
		{UNCHANGED, PERSISTENT_NONTRANSACTIONAL, PERSISTENT_NONTRANSACTIONAL,
		 PERSISTENT_NONTRANSACTIONAL, UNCHANGED, UNCHANGED, TRANSIENT_CLEAN,
		 TRANSIENT, TRANSIENT, UNCHANGED},

// rollback, retainValues = false
		{UNCHANGED, TRANSIENT, HOLLOW, HOLLOW, UNCHANGED, UNCHANGED,
		 TRANSIENT_CLEAN, TRANSIENT, HOLLOW, UNCHANGED},

// rollback, retainValues = true
		{UNCHANGED, TRANSIENT, PERSISTENT_NONTRANSACTIONAL, PERSISTENT_NONTRANSACTIONAL,
		 UNCHANGED, UNCHANGED, TRANSIENT_CLEAN, TRANSIENT, PERSISTENT_NONTRANSACTIONAL, UNCHANGED},

// refresh with active datastore transaction
		{UNCHANGED, UNCHANGED, UNCHANGED, PERSISTENT_CLEAN, UNCHANGED,
		 UNCHANGED, UNCHANGED, UNCHANGED, UNCHANGED, UNCHANGED},

// refresh with active optimistic transaction
		{UNCHANGED, UNCHANGED, UNCHANGED, PERSISTENT_NONTRANSACTIONAL, UNCHANGED,
		 UNCHANGED, UNCHANGED, UNCHANGED, UNCHANGED, UNCHANGED},

// evict
		{UNCHANGED, UNCHANGED, HOLLOW, UNCHANGED, UNCHANGED,
		 UNCHANGED, UNCHANGED, UNCHANGED, UNCHANGED, HOLLOW},

// read field outside transaction
		{UNCHANGED, IMPOSSIBLE, IMPOSSIBLE, IMPOSSIBLE, PERSISTENT_NONTRANSACTIONAL,
		 UNCHANGED, IMPOSSIBLE, IMPOSSIBLE, IMPOSSIBLE, UNCHANGED},

// read field with active optimistic transaction
		{UNCHANGED, UNCHANGED, UNCHANGED, UNCHANGED, PERSISTENT_NONTRANSACTIONAL,
		 UNCHANGED, UNCHANGED, ERROR, ERROR, UNCHANGED},

// read field with active datastore transaction
		{UNCHANGED, UNCHANGED, UNCHANGED, UNCHANGED, PERSISTENT_CLEAN,
		 UNCHANGED, UNCHANGED, ERROR, ERROR, PERSISTENT_CLEAN},

// write field outside transaction
		{UNCHANGED, IMPOSSIBLE, IMPOSSIBLE, IMPOSSIBLE, PERSISTENT_NONTRANSACTIONAL,
		 UNCHANGED, IMPOSSIBLE, IMPOSSIBLE, IMPOSSIBLE, UNCHANGED},

// write field with active transaction
		{UNCHANGED, UNCHANGED, PERSISTENT_DIRTY, UNCHANGED, PERSISTENT_DIRTY,
		 TRANSIENT_DIRTY, UNCHANGED, ERROR, ERROR, PERSISTENT_DIRTY}
	};

	private static final int DATASTORE_TX = 0;
	private static final int OPTIMISTIC_TX = 1;
	private static final int NO_TX = 2;

	private static final String[] scenario_string = {
		"datastore transaction", "optimistic transaction", "no transaction"
	};

	private static final boolean[][] applies_to_scenario = {
//      Datastore   Optimistic      No tx
		{true, true, false}, // makePersistent
		{true, true, false}, // deletePersistent
		{true, true, false}, // makeTransactional
		{true, true, false}, // makeNontransactional
		{true, true, false}, // makeTransient
		{true, true, false}, // commit RetainValues = false
		{true, true, false}, // commit RetainValues = true
		{true, true, false}, // rollback RetainValues = false
		{true, true, false}, // rollback RetainValues = true
		{true, false, false}, // refresh with active datastore
		{false, true, false}, // refresh with active optimistic
		{true, true, false}, // evict
		{false, false, true}, // read field outside transaction
		{false, true, false}, // read field with optimistic
		{true, false, false}, // read field with datastore
		{false, false, true}, // write field outside transaction
		{true, true, false}  // write field with active transaction
	};

	public StateTransitions() {
	}


	public static void main(String[] args) throws Exception {
		StateTransitions state_transitions = new StateTransitions();
		Status status = state_transitions.run(args, new PrintWriter(System.err, true),
				new PrintWriter(System.out, true));
		status.exit();
	}

	public Status run(String[] args, PrintWriter err, PrintWriter out) {
		initializeArgs(args, err, out);

		try {
			PersistenceManagerFactory pmf = getPMF();
			if (pmf == null) {
				return Status.failed("StateTransitions: Unable to acquire a PMF");
			}
			pm = pmf.getPersistenceManager();
			generatePersistentInstances();

			scenario = DATASTORE_TX;
			checkTransitions();
			if (isOptimisticSupported()) {
				scenario = OPTIMISTIC_TX;
				checkTransitions();
			}
			if (isNontransactionalReadSupported() || isNontransactionalWriteSupported()) {
				scenario = NO_TX;
				checkTransitions();
			}
		} catch (Exception e) {
			out.println("StateTransitions: unexpected exception:");
			out.println(e);
			e.printStackTrace(out);
			return Status.failed("StateTransitions: unexpected exception");
		} finally {
			Transaction tx = pm.currentTransaction();
			if (tx.isActive()) {
				tx.rollback();
			}
			pm.close();
		}

		int errCnt = getErrorCount();
		if (errCnt == 0) {
			return Status.passed("StateTransitions passed");
		} else {
			return Status.failed("StateTransitions failed, number of errors detected is " + errCnt);
		}
	}


	private void generatePersistentInstances() {
		if (doPersistentInstancesExist()) return;
		int i;
		Transaction t = pm.currentTransaction();
		t.begin();
		for (i = 0; i < 50; ++i) {
			StateTransitionObj sto = new StateTransitionObj(i);
			sto.writeField(i);
			pm.makePersistent(sto);
		}
		t.commit();
		if (!doPersistentInstancesExist())
			out.println("StateTransitions unable to create instances of StateTransitionsObj");
	}

	private boolean doPersistentInstancesExist() {
		boolean ret;
		Transaction t = pm.currentTransaction();
		t.begin();
		Extent e = pm.getExtent(StateTransitionObj.class, false);
		Iterator iter = e.iterator();
		ret = iter.hasNext();
		t.rollback();
		return ret;
	}

	void checkTransitions() {
		for (operation = 0; operation < NUM_OPERATIONS; ++operation) {
// rule out situations that do not apply
			if (!applies_to_scenario[operation][scenario]) continue;
			if (operation == READOUTSIDETX && !isNontransactionalReadSupported()) continue;
			if (operation == WRITEOUTSIDETX && !isNontransactionalWriteSupported()) continue;
			if ((operation == COMMITRETAINVALUES || operation == ROLLBACKRETAINVALUES) &&
					!isRetainValuesSupported())
				continue;
			if (operation == MAKENONTRANSACTIONAL &&
					!(isNontransactionalReadSupported() || isNontransactionalWriteSupported()))
				continue;
			for (current_state = 0; current_state < NUM_STATES; ++current_state) {
				if (scenario == OPTIMISTIC_TX && current_state == PERSISTENT_CLEAN) continue;
				if ((current_state == TRANSIENT_CLEAN || current_state == TRANSIENT_DIRTY) &&
						!isTransientTransactionalSupported())
					continue;   // this state is not supported by implementation
				if (current_state == PERSISTENT_NONTRANSACTIONAL &&
						!(isNontransactionalReadSupported() || isNontransactionalWriteSupported()))
					continue;   // this state is not supported by implementation
				expected_state = transitions[operation][current_state];
				if (expected_state == IMPOSSIBLE) continue;
				if (expected_state == UNCHANGED) expected_state = current_state;
				try {
					transaction = pm.currentTransaction();
					if (transaction.isActive()) {
						out.println("Transaction is active (but should not be), rolling back");
						transaction.rollback();
					}
					if (scenario != NO_TX) {

						if (operation == COMMITNORETAINVALUES || operation == ROLLBACKNORETAINVALUES)
							transaction.setRetainValues(false);
						if (operation == COMMITRETAINVALUES || operation == ROLLBACKRETAINVALUES)
							transaction.setRetainValues(true);
/* possibly switch above 4 lines to the following line
                    transaction.setRetainValues( operation == COMMITRETAINVALUES || operation == ROLLBACKRETAINVALUES );
 */
						transaction.setOptimistic(scenario == OPTIMISTIC_TX);
						transaction.begin();
						if (!transaction.isActive())
							out.println("StateTransitions: Transaction should be active, but it is not");
					}

// get and verify object is in right initial state
					StateTransitionObj obj = getInstanceInState(current_state);
					if (obj == null) {  // could not get object in state
						if (transaction.isActive()) transaction.rollback();
						continue;
					}

// Apply operation, catching possible exception
					Exception e = null;
					try {
						applyOperation(operation, obj);
					} catch (Exception excep) {
						if (excep instanceof javax.jdo.JDOUserException) {
							e = excep;
						} else {
							out.print("StateTransitions: Unexpected exception:");
							out.print(excep);
							printSituation();
							out.println("Stack trace is");
							excep.printStackTrace(out);
							incrementErrorCount();
							continue;
						}
					}

// Get new state, verify correct transition and exceptions occurred
					new_state = currentState(obj);
					if (expected_state == ERROR || expected_state == NOT_APPLICABLE) {
						if (e == null) {
							incrementErrorCount();
							out.print("StateTransitions: JDOUserException should have been thrown");
							printSituation();
						} else {
							if (new_state != current_state) {
								incrementErrorCount();
								out.print("StateTransitions: ");
								out.print("JDOUserException properly thrown, but instance should remain in current state,");
								out.print("instance changed state to ");
								out.print(states[new_state]);
								printSituation();
							}
						}
					}
					if (expected_state >= 0 && new_state != expected_state &&
							!((new_state == HOLLOW && expected_state == PERSISTENT_NONTRANSACTIONAL) ||
							(new_state == PERSISTENT_NONTRANSACTIONAL && expected_state == HOLLOW))) {
						incrementErrorCount();
						out.print("StateTransitions: Invalid state transition to ");
						out.print(states[new_state]);
						out.print(", new state should be ");
						out.print(states[expected_state]);
						printSituation();
					}
					if (transaction.isActive()) transaction.rollback();
				} catch (Exception unexpected_exception) {
					out.println("Unexpected exception caught in StateTransitions ");
					printSituation();
					unexpected_exception.printStackTrace(out);
					incrementErrorCount();
					if (transaction.isActive()) transaction.rollback();
				}
			}
		}
		out.flush();
	}

	void printSituation() {
		out.print(" (");
		out.print(scenario_string[scenario]);
		out.print(", initial state=");
		out.print(states[current_state]);
		out.print(", operation=");
		out.print(operations[operation]);
		out.println(")");
	}

	void applyOperation(int operation, StateTransitionObj stobj) {
		StateTransitionObj obj = (StateTransitionObj) stobj;
		switch (operation) {
		case MAKEPERSISTENT:
			{
				pm.makePersistent(obj);
				break;
			}
		case DELETEPERSISTENT:
			{
				pm.deletePersistent(obj);
				break;
			}
		case MAKETRANSACTIONAL:
			{
				pm.makeTransactional(obj);
				break;
			}
		case MAKENONTRANSACTIONAL:
			{
				pm.makeNontransactional(obj);
				break;
			}
		case MAKETRANSIENT:
			{
				pm.makeTransient(obj);
				break;
			}
		case COMMITNORETAINVALUES:
			{
				pm.currentTransaction().commit();
				break;
			}
		case COMMITRETAINVALUES:
			{
				pm.currentTransaction().commit();
				break;
			}
		case ROLLBACKNORETAINVALUES:
			{
				pm.currentTransaction().rollback();
				break;
			}
		case ROLLBACKRETAINVALUES:
			{
				pm.currentTransaction().rollback();
				break;
			}
		case REFRESHDATASTORE:
			{
				pm.refresh(obj);
				break;
			}
		case REFRESHOPTIMISTIC:
			{
				pm.refresh(obj);
				break;
			}
		case EVICT:
			{
				pm.evict(obj);
				break;
			}
		case READOUTSIDETX:
			{
				int val = obj.readField();
				break;
			}
		case READOPTIMISTIC:
			{
				int val = obj.readField();
				break;
			}
		case READDATASTORE:
			{
				int val = obj.readField();
				break;
			}
		case WRITEOUTSIDETX:
			{
				obj.writeField(42);
				break;
			}
		case WRITEINSIDETX:
			{
				obj.writeField(42);
				break;
			}
		default:
			{
				err.print("StateTransitions internal error, illegal operation ");
				err.println(operation);
			}
		}
	}


	/**
	 * Get an instance in the specified state.
	 */
	private StateTransitionObj getInstanceInState(int state) {
		switch (state) {
		case TRANSIENT:
			return getTransientInstance();
		case PERSISTENT_NEW:
			return getPersistentNewInstance();
		case PERSISTENT_CLEAN:
			return getPersistentCleanInstance();
		case PERSISTENT_DIRTY:
			return getPersistentDirtyInstance();
		case HOLLOW:
			return getHollowInstance();
		case TRANSIENT_CLEAN:
			return getTransientCleanInstance();
		case TRANSIENT_DIRTY:
			return getTransientDirtyInstance();
		case PERSISTENT_NEW_DELETED:
			return getPersistentNewDeletedInstance();
		case PERSISTENT_DELETED:
			return getPersistentDeletedInstance();
		case PERSISTENT_NONTRANSACTIONAL:
			return getPersistentNontransactionalInstance();
		default:
			{
				return null;
			}
		}
	}

	private StateTransitionObj getTransientInstance() {
		StateTransitionObj obj = new StateTransitionObj(23);
		int curr = currentState(obj);
		if (curr != TRANSIENT) {
			out.print("StateTransitions: Unable to create transient instance, state is ");
			out.println(states[curr]);
			printSituation();
			return null;
		}
		return obj;
	}

	private StateTransitionObj getPersistentNewInstance() {
		StateTransitionObj obj = getTransientInstance();
		if (obj == null) return null;
		pm.makePersistent(obj); // should transition to persistent-new
		int curr = currentState(obj);
		if (curr != PERSISTENT_NEW) {
			out.print("StateTransitions: Unable to create persistent-new instance");
			out.print(" from transient instance via makePersistent(), state is ");
			out.println(states[curr]);
			printSituation();
			return null;
		}
		return obj;
	}

	public StateTransitionObj getPersistentCleanInstance() {
		StateTransitionObj obj = getHollowInstance();
		if (obj == null) return null;
		StateTransitionObj sto = (StateTransitionObj) obj;
		int val = sto.readField();
		int curr = currentState(sto);
		if (curr != PERSISTENT_CLEAN) {
			out.print("StateTransition: Unable to create persistent-clean instance");
			out.print(" from a hollow instance by reading a field, state is ");
			out.println(states[curr]);
			printSituation();
			return null;
		}
		return obj;
	}

	public StateTransitionObj getPersistentDirtyInstance() {
		StateTransitionObj obj = getHollowInstance();
		if (obj == null) return null;
		StateTransitionObj pcobj = (StateTransitionObj) obj;
		pcobj.writeField(23);
		int curr = currentState(obj);
		if (curr != PERSISTENT_DIRTY) {
			out.print("StateTransition: Unable to create persistent-dirty instance");
			out.print(" from a hollow instance by writing a field, state is ");
			out.println(states[curr]);
			printSituation();
			return null;
		}
		return obj;
	}

	public StateTransitionObj getHollowInstance() {
		Extent extent = pm.getExtent(StateTransitionObj.class, false);
		Iterator iter = extent.iterator();
		if (!iter.hasNext()) {
			out.println("Extent for StateTransitionObj should not be empty");
			return null;
		}
		StateTransitionObj obj = (StateTransitionObj) iter.next();
		int curr = currentState(obj);
		if (curr != HOLLOW && curr != PERSISTENT_NONTRANSACTIONAL) {
			out.print("StateTransition: Attempt to get hollow instance via accessing extent failed, state is ");
			out.println(states[curr]);
			printSituation();
			return null;
		}
		return obj;
	}

	public StateTransitionObj getTransientCleanInstance() {
		StateTransitionObj obj = getTransientInstance();
		if (obj == null) return null;
		pm.makeTransactional(obj);
		int curr = currentState(obj);
		if (curr != TRANSIENT_CLEAN) {
			out.print("StateTransition: Unable to create transient-clean instance");
			out.print(" from a transient instance via makeTransactional(), state is ");
			out.println(states[curr]);
			printSituation();
			return null;
		}
		return obj;
	}

	public StateTransitionObj getTransientDirtyInstance() {
		StateTransitionObj obj = getTransientCleanInstance();
		if (obj == null) return null;
		StateTransitionObj pcobj = (StateTransitionObj) obj;
		pcobj.writeField(23);
		int curr = currentState(obj);
		if (curr != TRANSIENT_DIRTY) {
			out.print("StateTransition: Unable to create transient-dirty instance");
			out.print(" from a transient-clean instance via modifying a field, state is ");
			out.println(states[curr]);
			printSituation();
			return null;
		}
		return obj;
	}

	public StateTransitionObj getPersistentNewDeletedInstance() {
		StateTransitionObj obj = getPersistentNewInstance();
		if (obj == null) return null;
		pm.deletePersistent(obj);   // should transition to persistent-new-deleted
		int curr = currentState(obj);
		if (curr != PERSISTENT_NEW_DELETED) {
			out.print("StateTransition: Unable to create persistent-new-deleted instance");
			out.print(" from a persistent-new instance via deletePersistent, state is ");
			out.println(states[curr]);
			printSituation();
			return null;
		}
		return obj;
	}

	public StateTransitionObj getPersistentDeletedInstance() {
		StateTransitionObj obj = getHollowInstance();
		if (obj == null) return null;
		pm.deletePersistent(obj);
		int curr = currentState(obj);
		if (curr != PERSISTENT_DELETED) {
			out.print("StateTransition: Unable to create persistent-deleted instance");
			out.print(" from a persistent instance via deletePersistent(), state is ");
			out.println(states[curr]);
			printSituation();
			return null;
		}
		return obj;
	}

	public StateTransitionObj getPersistentNontransactionalInstance() {
		StateTransitionObj obj = getHollowInstance();
		if (obj == null) return null;
		pm.makeNontransactional(obj);
		int curr = currentState(obj);
		if (curr != PERSISTENT_NONTRANSACTIONAL && curr != HOLLOW) {
			out.print("StateTransition: Unable to create persistent-nontransactional instance");
			out.print(" from a persistent-clean instance via makeNontransactional(), state is ");
			out.println(states[curr]);
			printSituation();
			return null;
		}
		return null;
	}

}
