/*
 * Copyright 2001 Sun Microsystems, Inc. All rights reserved.
 * Use is subject to license terms.
 *
 */

package com.sun.jdotck.models.company;

import java.io.Serializable;
import java.util.Collections;
import java.util.Set;
import java.util.HashSet;
import java.util.Date;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;

/**
 *This class represents information about a company.
 */
public class Company implements Serializable {
	private String name;
	private Date founded;
	private Address address;
	private transient HashSet departments; // element type is Department

	/** This is the JDO-required no-args constructor
	 */
	private Company() {
	}

	/**
	 *Initialize the <code>Company</code> instance.
	 */
	public Company(String name, Date founded, Address addr) {
		this.name = name;
		this.founded = founded;
		address = addr;
		departments = new HashSet();
	}

	/**
	 *Get the name of the company.
	 *@return The name of the company.
	 */
	public String getName() {
		return name;
	}

	/**
	 *Set the  name of the company.
	 *@param name The value to use for the name of the company.
	 */
	public void setName(String name) {
		this.name = name;
	}

	/**
	 *Get the date that the company was founded.
	 *@return The date the company was founded.
	 */
	public Date getFounded() {
		return founded;
	}

	/**
	 *Set the date that the company was founded.
	 *@param founded The date to set that the company was founded.
	 */
	public void setFounded(Date founded) {
		this.founded = founded;
	}

	/**
	 *Get the address of the company.
	 *@return The primary address of the company.
	 */
	public Address getAddress() {
		return address;
	}

	/**
	 *Set the primary address for the company.
	 *@param address The address to set for the company.
	 */
	public void setAddress(Address address) {
		this.address = address;
	}

	/**
	 *Get the departments contained in the company.
	 *@return An unmodifiable <code>Set</code> that contains all the
	 *<code>Department</code>s of the company.
	 */
	public Set getDepartments() {
		return Collections.unmodifiableSet(departments);
	}

	/**
	 *Add a <code>Department</code> instance to the company.
	 *@param dept The <code>Department</code> instance to add.
	 */
	public void addDepartment(Department dept) {
		departments.add(dept);
	}

	/**
	 *Remove a <code>Department</code> instance from the company.
	 *@param dept The <code>Department</code> instance to remove.
	 */
	public void removeDepartment(Department dept) {
		departments.remove(dept);
	}

	/**
	 *Initialize the set of <code>Department</code>s in the company to the parameter.
	 *@param departments The set of <code>Department</code>s for the company.
	 */
	public void setDepartments(HashSet departments) {
		this.departments = departments;
	}

	private void writeObject(ObjectOutputStream out)
			throws IOException {
		out.writeObject(name);
		out.writeObject(founded);
		out.writeObject(address);
	}

	private void readObject(ObjectInputStream in)
			throws IOException, ClassNotFoundException {
		name = (String) in.readObject();
		founded = (Date) in.readObject();
		address = (Address) in.readObject();
	}

	/**
	 *The class to be used as the application identifier
	 *for the <code>Company</code> class. It consists of both the company
	 *name and the date that the company was founded.
	 */
	public static class Oid implements Serializable {
		/**
		 *This field is part of the identifier and should match in name and type
		 *with a field in the <code>Company</code> class.
		 */
		public String name;
		/**
		 *This field is part of the identifier and should match in name and type
		 *with a field in the <code>Company</code> class.
		 */

		/**
		 *The required null constructor.
		 */
		public Oid() {
		}

		/**
		 *Initialize the identifier.
		 *@param name The name of the company.
		 */
		public Oid(String name) {
			this.name = name;
		}

		public boolean equals(java.lang.Object obj) {
			if (obj == null ||
					!this.getClass().equals(obj.getClass()))
				return (false);
			Oid o = (Oid) obj;
			if ((this.name != o.name) &&
					(this.name == null || !this.name.equals(o.name)))
				return (false);
			return (true);
		}

		public int hashCode() {
			int hashCode = 0;
			if (name != null) hashCode += name.hashCode();
			return (hashCode);
		}

	}

}

