/*
 * Copyright 2001 Sun Microsystems, Inc. All rights reserved.
 * Use is subject to license terms.
 *
 */

package com.sun.jdotck.models.company;

import java.io.Serializable;
import java.util.Collections;
import java.util.Set;
import java.util.HashSet;

/**
 *This class represents a department within a company.
 */

public class Department implements Serializable {
	private long deptid;
	private String name;
	private Company company;

	private transient HashSet employees; // element type is Employee

	/** This is the JDO-required no-args constructor
	 */
	private Department() {
	}

	/**
	 *Initialize a <code>Department</code> instance.
	 *@param deptid The department id.
	 *@param name The name of the department.
	 *@param company The company that the department is associated with.
	 */
	public Department(long deptid, String name, Company company) {
		this.deptid = deptid;
		this.name = name;
		this.company = company;
		employees = new HashSet();
	}

	/**
	 *Get the department id.
	 *@return The department id.
	 */
	public long getDeptid() {
		return deptid;
	}

	/**
	 *Set the department id.
	 *@param deptid The department id.
	 */
	public void setDeptid(long deptid) {
		this.deptid = deptid;
	}

	/**
	 *Get the name of the department.
	 *@return The name of the department.
	 */
	public String getName() {
		return name;
	}

	/**
	 *Set the name of the department.
	 *@param name The name to set for the department.
	 */
	public void setName(String name) {
		this.name = name;
	}

	/**
	 *Get the company associated with the department.
	 *@return The company.
	 */
	public Company getCompany() {
		return company;
	}

	/**
	 *Set the company for the department.
	 *@param company The company to associate with the department.
	 */
	public void setCompany(Company company) {
		this.company = company;
	}

	/**
	 *Get the employees in the department as an unmodifiable set.
	 *@return The set of employees in the department, as an unmodifiable set.
	 */
	public Set getEmployees() {
		return Collections.unmodifiableSet(employees);
	}

	/**
	 *Add an employee to the department.
	 *@param emp The employee to add to the department.
	 */
	public void addEmployee(Employee emp) {
		employees.add(emp);
	}

	/**
	 *Remove an employee from the department.
	 *@param emp The employee to remove from the department.
	 */
	public void removeEmployee(Employee emp) {
		employees.remove(emp);
	}

	/**
	 *Set the employees to be in this department.
	 *@param employees The set of employees for this department.
	 */
	public void setEmployees(HashSet employees) {
		this.employees = employees;
	}

	private void writeObject(java.io.ObjectOutputStream out)
			throws java.io.IOException {
		out.writeLong(deptid);
		out.writeObject(name);
		out.writeObject(company);
	}

	private void readObject(java.io.ObjectInputStream in)
			throws java.io.IOException, ClassNotFoundException {
		deptid = in.readLong();
		name = (String) in.readObject();
		company = (Company) in.readObject();
	}

	/**
	 *The application identity class associated with the <code>Department</code> class.
	 */
	public static class Oid implements Serializable {
		/**
		 *This field represents the application identifier field
		 *for the <code>Department</code> class.
		 *It must match in name and type with the field in the <code>Department</code> class.
		 */
		public long deptid;

		/**
		 *The required null constructor.
		 */
		public Oid() {
			deptid = 0;
		}

		public Oid(long deptid) {
			this.deptid = deptid;
		}

		public boolean equals(java.lang.Object obj) {
			if (obj == null || !this.getClass().equals(obj.getClass()))
				return (false);
			Oid o = (Oid) obj;
			if (this.deptid != o.deptid) return (false);
			return (true);
		}

		public int hashCode() {
			return ((int) deptid);
		}

	}

}

