/*
 * Copyright 2001 Sun Microsystems, Inc. All rights reserved.
 * Use is subject to license terms.
 *
 */

package com.sun.jdotck.models.company;

import java.util.Collections;
import java.util.Set;
import java.util.HashSet;
import java.util.Date;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;

/**
 *This class represents an employee.
 */

public class Employee extends Person {
	private Date hiredate;
	private double weeklyhours;
	private char discriminator;
	private Insurance insurance;
	private Department department;
	private Employee manager;
	private Employee mentor;
	private Employee protege;
	private Employee hradvisor;
	private transient HashSet reviewedProjects; // element-type is Project
	private transient HashSet projects;         // element-type is Project
	private transient HashSet team;             // element-type is Employee
	private transient HashSet hradvisees;       // element-type is Employee

	/** This is the JDO-required no-args constructor
	 */
	protected Employee() {
	}

	/**
	 *Initialize an <code>Employee</code> instance.
	 *@param personid The identifier for the person.
	 *@param firstname The first name of the employee.
	 *@param lastname The last name of the employee.
	 *@param address The address of the employee.
	 *@param birthdate The birth date of the employee.
	 *@param hiredate The date that the employee was hired.
	 */
	public Employee(long personid,
					String firstname,
					String lastname,
					Address address,
					Date birthdate,
					Date hiredate) {
		super(personid, firstname, lastname, address, birthdate);
		this.hiredate = hiredate;
		reviewedProjects = new HashSet();
		projects = new HashSet();
		team = new HashSet();
		hradvisees = new HashSet();
	}

	/**
	 *Get the date that the employee was hired.
	 *@return The date the employee was hired.
	 */
	public Date getHiredate() {
		return hiredate;
	}

	/**
	 *Set the date that the employee was hired.
	 *@param hiredate The date the employee was hired.
	 */
	public void setHiredate(Date hiredate) {
		this.hiredate = hiredate;
	}

	/**
	 *Get the weekly hours of the employee.
	 *@return The number of hours per week that the employee works.
	 */
	public double getWeeklyhours() {
		return weeklyhours;
	}

	/**
	 *Set the number of hours per week that the employee works.
	 *@param weeklyhours The number of hours per week that the employee works.
	 */
	public void setWeeklyhours(double weeklyhours) {
		this.weeklyhours = weeklyhours;
	}

	/**
	 *Get the discriminator of the employee.
	 *@return The discriminator of the employee.
	 */
	public char getDiscriminator() {
		return discriminator;
	}

	/**
	 *Set the discriminator of the employee.
	 *param discriminator The discriminator of the employee.
	 */
	public void setDiscriminator(char discriminator) {
		this.discriminator = discriminator;
	}

	/**
	 *Get the reviewed projects.
	 *@return The reviewed projects as an unmodifiable set.
	 */
	public Set getReviewedProjects() {
		return Collections.unmodifiableSet(reviewedProjects);
	}

	/**
	 *Add a reviewed project.
	 *@param project A reviewed project.
	 */
	public void addReviewedProjects(Project project) {
		reviewedProjects.add(project);
	}

	/**
	 *Remove a reviewed project.
	 *@param project A reviewed project.
	 */
	public void removeReviewedProject(Project project) {
		reviewedProjects.remove(project);
	}

	/**
	 *Set the reviewed projects for the employee.
	 *@param reviewedProjects The set of reviewed projects.
	 */
	public void setReviewedProjects(HashSet reviewedProjects) {
		this.reviewedProjects = reviewedProjects;
	}

	/**
	 *Get the employee's projects.
	 *@return The employee's projects are returned as an unmodifiable set.
	 */
	public Set getProjects() {
		return Collections.unmodifiableSet(projects);
	}

	/**
	 *Add a project for the employee.
	 *@param project The project.
	 */
	public void addProject(Project project) {
		projects.add(project);
	}

	/**
	 *Remove a project from an employee's set of projects.
	 *@param project The project.
	 */
	public void removeProject(Project project) {
		projects.remove(project);
	}

	/**
	 *Set the projects for the employee.
	 *@param projects The set of projects of the employee.
	 */
	public void setProjects(HashSet projects) {
		this.projects = projects;
	}

	/**
	 *Get the insurance of the employee.
	 *@return The employee's insurance.
	 */
	public Insurance getInsurance() {
		return insurance;
	}

	/**
	 *Set the insurance object for the employee.
	 *@param insurance The insurance object to associate with the employee.
	 */
	public void setInsurance(Insurance insurance) {
		this.insurance = insurance;
	}

	/**
	 *Get the employee's department.
	 *@return The department associated with the employee.
	 */
	public Department getDepartment() {
		return department;
	}

	/**
	 *Set the employee's department.
	 *@param department The department.
	 */
	public void setDepartment(Department department) {
		this.department = department;
	}

	/**
	 *Get the employee's manager.
	 *@return The employee's manager.
	 */
	public Employee getManager() {
		return manager;
	}

	/**
	 *Get the employee's team.
	 *@return The set of <code>Employee</code>s on this employee's team, returned as an unmodifiable set.
	 */
	public Set getTeam() {
		return Collections.unmodifiableSet(team);
	}

	/**
	 *Add an <code>Employee</code> to this employee's team.
	 *This method sets both sides of the relationship, modifying
	 *this employees team to include parameter emp and modifying
	 *emp to set its manager attribute to this object.
	 *@param emp The <code>Employee</code> to add to the team.
	 */
	public void addToTeam(Employee emp) {
		team.add(emp);
		emp.manager = this;
	}

	/**
	 *Remove an <code>Employee</code> from this employee's team.
	 *This method will also set the <code>emp</code> manager to null.
	 *@param emp The <code>Employee</code> to remove from the team.
	 */
	public void removeFromTeam(Employee emp) {
		team.remove(emp);
		emp.manager = null;
	}

	/**
	 *Set the mentor for this employee and also set the inverse protege relationship.
	 *@param mentor The mentor for this employee.
	 */
	public void setMentor(Employee mentor) {
		this.mentor = mentor;
		mentor.protege = this;
	}

	/**
	 *Get the mentor for this employee.
	 *@return The mentor.
	 */
	public Employee getMentor() {
		return mentor;
	}

	/**
	 *Get the protege of this employee.
	 *@return The protege of this employee.
	 */
	public Employee getProtege() {
		return protege;
	}

	/**
	 *Get the HR advisor for the employee.
	 *@return The HR advisor.
	 */
	public Employee getHradvisor() {
		return hradvisor;
	}

	/**
	 *Get the HR advisees of this HR advisor.
	 *@return An unmodifiable <code>Set</code> containing the <code>Employee</code>s
	 *that are HR advisees of this employee.
	 */
	public Set getHradvisees() {
		return Collections.unmodifiableSet(hradvisees);
	}

	/**
	 *Add an <code>Employee</code> as an advisee of this HR advisor.
	 *This method also sets the <code>emp</code> hradvisor to reference this object.
	 *In other words, both sides of the relationship are set.
	 *param emp The employee to add as an advisee.
	 */
	public void addAdvisee(Employee emp) {
		hradvisees.add(emp);
		emp.hradvisor = this;
	}

	/**
	 *Remove an <code>Employee</code> as an advisee of this HR advisor.
	 *This method also sets the <code>emp</code> hradvisor to null.
	 *In other words, both sides of the relationship are set.
	 *@param emp The employee to add as an HR advisee.
	 */
	public void removeAdvisee(Employee emp) {
		hradvisees.remove(emp);
		emp.hradvisor = null;
	}

	protected void writeObject(ObjectOutputStream out)
			throws IOException {
		super.writeObject(out);
		out.writeObject(hiredate);
		out.writeDouble(weeklyhours);
		out.writeChar(discriminator);
		out.writeObject(insurance);
		out.writeObject(department);
		out.writeObject(manager);
		out.writeObject(mentor);
		out.writeObject(protege);
		out.writeObject(hradvisor);
	}

	protected void readObject(ObjectInputStream in)
			throws IOException, ClassNotFoundException {
		super.readObject(in);
		hiredate = (Date) in.readObject();
		weeklyhours = in.readDouble();
		discriminator = in.readChar();
		insurance = (Insurance) in.readObject();
		department = (Department) in.readObject();
		manager = (Employee) in.readObject();
		mentor = (Employee) in.readObject();
		protege = (Employee) in.readObject();
		hradvisor = (Employee) in.readObject();
	}

}

