/*
 * Copyright 2001 Sun Microsystems, Inc. All rights reserved.
 * Use is subject to license terms.
 *
 */

package com.sun.jdotck.models.company;

import java.io.Serializable;
import java.util.Collections;
import java.util.Set;
import java.util.HashSet;
import java.math.BigDecimal;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;

/**
 *This class represents a project, a budgeted task with one or more employees
 *working on it.
 */
public class Project implements Serializable {
	private long projid;
	private String name;
	private BigDecimal budget;
	private transient HashSet reviewers; // element type is Employee
	private transient HashSet members; // element type is Employee

	/** This is the JDO-required no-args constructor
	 */
	private Project() {
	}

	/**
	 *Initialize a project.
	 *@param projid The project identifier.
	 *@param name The name of the project.
	 *@param budget The budget for the project.
	 */
	public Project(long projid, String name, BigDecimal budget) {
		this.projid = projid;
		this.name = name;
		this.budget = budget;
		reviewers = new HashSet();
		members = new HashSet();
	}

	/**
	 *Get the project ID.
	 *@return The project ID.
	 */
	public long getProjID() {
		return projid;
	}

	/**
	 *Set the project's ID.
	 *@param projid The project ID.
	 */
	public void setProjID(long projid) {
		this.projid = projid;
	}

	/**
	 *Get the name of the project.
	 *@return The name of the project.
	 */
	public String getName() {
		return name;
	}

	/**
	 *Set the name of the project.
	 *@param name The name of the project.
	 */
	public void setName(String name) {
		this.name = name;
	}

	/**
	 *Get the project's budget.
	 *@return The project's budget.
	 */
	public BigDecimal getBudget() {
		return budget;
	}

	/**
	 *Set the project's budget.
	 *@param budget The project's budget.
	 */
	public void setBudget(BigDecimal budget) {
		this.budget = budget;
	}

	/**
	 *Get the reviewers associated with this project.
	 */
	public Set getReviewers() {
		return Collections.unmodifiableSet(reviewers);
	}

	/**
	 *Add a reviewer to the project.
	 *@param emp The employee to add as a reviewer.
	 */
	public void addReviewer(Employee emp) {
		reviewers.add(emp);
	}

	/**
	 *Remove a reviewer from the project.
	 *@param emp The employee to remove as a reviewer of this project.
	 */
	public void removeReviewer(Employee emp) {
		reviewers.remove(emp);
	}

	/**
	 *Set the reviewers associated with this project.
	 *@param reviewers The set of reviewers to associate with this project.
	 */
	public void setReviewers(HashSet reviewers) {
		this.reviewers = reviewers;
	}

	/**
	 *Get the project members.
	 *@return The members of the project is returned as an unmodifiable set of <code>Employee</code>s.
	 */
	public Set getMembers() {
		return Collections.unmodifiableSet(members);
	}

	/**
	 *Add a new member to the project.
	 *@param emp The employee to add to the project.
	 */
	public void addMember(Employee emp) {
		members.add(emp);
	}

	/**
	 *Remove a member from the project.
	 *@param emp The employee to remove from the project.
	 */
	public void removeMember(Employee emp) {
		members.remove(emp);
	}

	/**
	 *Set the members of the project.
	 *@param employees The set of employees to be the members of this project.
	 */
	public void setMembers(HashSet employees) {
		members = employees;
	}

	private void writeObject(ObjectOutputStream out)
			throws IOException {
		out.writeLong(projid);
		out.writeObject(name);
		out.writeObject(budget);
	}

	private void readObject(ObjectInputStream in)
			throws IOException, ClassNotFoundException {
		projid = in.readLong();
		name = (String) in.readObject();
		budget = (java.math.BigDecimal) in.readObject();
	}

	/**
	 *This class is used to represent the application identity
	 *for the <code>Project</code> class.
	 */
	public static class Oid implements Serializable {
		/**
		 *This field represents the identifier for the <code>Project</code> class.
		 *It must match a field in the <code>Project</code> class in both name and type.
		 */
		public long projid;

		/**
		 *The required null constructor.
		 */
		public Oid() {
			projid = 0;
		}

		/**
		 *Initialize the application identifier with a project ID.
		 */
		public Oid(long projid) {
			this.projid = projid;
		}

		public boolean equals(java.lang.Object obj) {
			if (obj == null || !this.getClass().equals(obj.getClass()))
				return (false);
			Oid o = (Oid) obj;
			if (this.projid != o.projid) return (false);
			return (true);
		}

		public int hashCode() {
			return ((int) projid);
		}

	}

}

