/*
 * Copyright 2001 Sun Microsystems, Inc. All rights reserved.
 * Use is subject to license terms.
 *
 */


package com.sun.jdotck.query;

import java.io.PrintWriter;

import com.sun.javatest.Test;
import com.sun.javatest.Status;

import javax.jdo.*;

/**
 *<B>JavaTest Title:</B> Query Ordering Specification
 *<BR>
 *<B>JavaTest Keywords:</B> query
 *<BR>
 *<B>Assertion ID:</B> A14.3-6.
 *<BR>
 *<B>Assertion Description: </B>
 The <code>Query</code> ordering specification includes
 a list of one or more ordering declarations separated by commas.
 Each ordering declaration includes an expression whose type is one of:
 <UL>
 <LI>primitive types except <code>boolean</code></LI>
 <LI>wrapper types except <code>Boolean</code></LI>
 <LI><code>BigDecimal</code></LI>
 <LI><code>BigInteger</code></LI>
 <LI><code>String</code></LI>
 <LI><code>Date</code></LI></UL>
 followed by one of the following words: &quot;ascending&quot;; or &quot;descending&quot;;
 which indicates the ordering of the values for that expression.
 Ordering might be specified including navigation.
 The name of the field to be used in ordering via navigation through
 single-valued fields is specified by the Java language syntax of
 <code>field_name.field_name.....field_name</code>.

 *@test
 *@title Query Ordering Specification
 *@executeClass com.sun.jdotck.query.QueryOrderingSpecification
 *@executeArgs -PMFProperties $PMFProperties
 * -ApplicationIdentity $ApplicationIdentity
 * -Array $Array
 * -ArrayList $ArrayList
 * -ChangeApplicationIdentity $ChangeApplicationIdentity
 * -DatastoreIdentity $DatastoreIdentity
 * -HashMap $HashMap
 * -Hashtable $Hashtable
 * -LinkedList $LinkedList
 * -List $List
 * -Map $Map
 * -NonDurableIdentity $NonDurableIdentity
 * -NontransactionalRead $NontransactionalRead
 * -NontransactionalWrite $NontransactionalWrite
 * -NullCollection $NullCollection
 * -Optimistic $Optimistic
 * -RetainValues $RetainValues
 * -TransientTransactional $TransientTransactional
 * -TreeMap $TreeMap
 * -TreeSet $TreeSet
 * -Vector $Vector
 *@source QueryOrderingSpecification.java
 *@keywords query
 */

public class QueryOrderingSpecification extends QueryTest {

	/**
	 *The <code>main</code> is called when the class
	 *is directly executed from the command line.
	 *@param args The arguments passed to the program.
	 */
	public static void main(String[] args) {
		QueryOrderingSpecification t = new QueryOrderingSpecification();
		Status s = t.run(args, new PrintWriter(System.err), new PrintWriter(System.out));
		s.exit();
	}

	/**
	 *The <code>run</code> method is invoked by JavaTest to
	 *execute the test defined by this class.
	 *@param args The arguments provided by the test script,
	 *providing configuration information for a test.
	 *@param err A stream to which to report errors.
	 *@param out A stream to report messages.
	 *@return A <code>Status</code> object representing the outcome of the test.
	 */
	public Status run(String[] args, PrintWriter err, PrintWriter out) {
		initializeArgs(args, err, out);
		try {
			PersistenceManagerFactory pmf = getPMF();
			PersistenceManager pm = pmf.getPersistenceManager();
			checkQueryOrderingSpecifications(pm);
			pm.close();
			return Status.passed("Assertion A14.3-6 passed");
		} catch (Exception e) {
			e.printStackTrace(out);
		}
		return Status.failed("Assertion A14.3-6 failed");
	}

	void checkQueryOrderingSpecifications(PersistenceManager pm) {
		Transaction tx = pm.currentTransaction();
		try {
			tx.begin();
			assertBooleanTypes(pm);
			checkRemainingTypes(pm);
		} finally {
			tx.rollback();
		}
	}

	void assertBooleanTypes(PersistenceManager pm) {
		try {
			Query query = pm.newQuery();
			if (query == null)
				throw new RuntimeException("checkQueryOrderingSpecifications(): query instance is null");
			Class clazz = com.sun.jdotck.models.fieldtypes.AllTypes.class;
			query.setClass(clazz);
			query.setCandidates(pm.getExtent(clazz, false));

			query.setOrdering("fld_boolean");
			Object results = query.execute();
			throw new RuntimeException("Boolean types should be rejected in ordering specifications");
		} catch (JDOException e) {
			out.println("Caught expected " + e);
		}

		try {
			Query query = pm.newQuery();
			if (query == null)
				throw new RuntimeException("checkQueryOrderingSpecifications(): query instance is null");
			Class clazz = com.sun.jdotck.models.fieldtypes.AllTypes.class;
			query.setClass(clazz);
			query.setCandidates(pm.getExtent(clazz, false));

			query.setOrdering("fld_Boolean");
			Object results = query.execute();
			throw new RuntimeException("Boolean wrapper types should be rejected in ordering specifications");
		} catch (JDOException e) {
			out.println("Caught expected " + e);
		}
	}

	void checkRemainingTypes(PersistenceManager pm) {
		Query query = pm.newQuery();
		if (query == null)
			throw new RuntimeException("checkQueryOrderingSpecifications(): query instance is null");
		Class clazz = com.sun.jdotck.models.fieldtypes.AllTypes.class;
		query.setClass(clazz);
		query.setCandidates(pm.getExtent(clazz, false));

		StringBuffer buffer = new StringBuffer();
		buffer.append("  fld_byte ascending");
		buffer.append(", fld_char ascending");
		buffer.append(", fld_double ascending");
		buffer.append(", fld_float ascending");
		buffer.append(", fld_int ascending");
		buffer.append(", fld_long ascending");
		buffer.append(", fld_short ascending");
		buffer.append(", fld_Byte descending");
		buffer.append(", fld_Character descending");
		buffer.append(", fld_Double descending");
		buffer.append(", fld_Float ascending");
		buffer.append(", fld_Integer descending");
		buffer.append(", fld_Long descending");
		buffer.append(", fld_Short descending");
		buffer.append(", fld_String descending");
		buffer.append(", fld_Date descending");
		buffer.append(", fld_BigDecimal descending");
		buffer.append(", fld_BigInteger ascending");
		query.setOrdering(buffer.toString());
		Object results = query.execute();
	}
}
